/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWLClassClassificationGraph.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.owl.reasoner.structural;

import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.eodm.owl.AllValuesFromRestriction;
import org.eclipse.eodm.owl.EnumeratedClass;
import org.eclipse.eodm.owl.HasValueRestriction;
import org.eclipse.eodm.owl.Individual;
import org.eclipse.eodm.owl.IntersectionClass;
import org.eclipse.eodm.owl.MinCardinalityRestriction;
import org.eclipse.eodm.owl.OWLClass;
import org.eclipse.eodm.owl.OWLDatatypeProperty;
import org.eclipse.eodm.owl.OWLFactory;
import org.eclipse.eodm.owl.OWLObjectProperty;
import org.eclipse.eodm.owl.SomeValuesFromRestriction;
import org.eclipse.eodm.owl.UnionClass;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.RDFProperty;
import org.eclipse.eodm.rdfs.RDFSDatatype;
import org.eclipse.eodm.rdfs.RDFSFactory;
import org.eclipse.eodm.rdfs.RDFSLiteral;
import org.eclipse.eodm.rdfs.TypedLiteral;

/**
 * OWLClassClassificationGraph TODO Please add comments here.
 */
public class OWLClassClassificationGraph {
    public OWLClassClassificationGraph(RDFPropertyGraph pGraph) {
        this.pGraph = pGraph;
        this.identifier = new StrictClassIdentifier(pGraph);
        int thingID = identifier
                .getOWLClassID(ReasonerUtils.thing, false, true);
        int nothingID = identifier.getOWLClassID(ReasonerUtils.nothing, false,
                true);
        thingNode = new OWLClassNode(this, thingID);
        nothingNode = new OWLClassNode(this, nothingID);
        addNodeContainment(thingNode, nothingNode);
        new AllValuesFromLink(this, thingNode, thingNode, pGraph
                .getTopPropertyID());
        new SomeValuesFromLink(this, nothingNode, nothingNode, pGraph
                .getTopPropertyID());
    }

    public OWLClassNode getNodeByID(int classID) {
        return (OWLClassNode) classToNodeMap.get(new Integer(classID));
    }

    public OWLClassNode getNodeByClass(OWLClass c, boolean doNormalizationFirst) {
        return getNodeByID(identifier.getOWLClassID(c, doNormalizationFirst,
                false));
    }

    public RDFPropertyGraph getPGraph() {
        return pGraph;
    }

    /**
     * This method adds a new axiom c1 >= c2 to the classification graph. The
     * involved classes will be added first, and then the containment link will
     * be added. If more containment links can be inferred by this axiom, they
     * will be added too.
     * 
     * @param c1
     *            The class expression c1
     * @param c2
     *            The class expression c2
     */
    public void addAxiom(OWLClass c1, OWLClass c2) {
   
        OWLClassNode v1 = addClass(c1);

        OWLClassNode v2 = addClass(c2);

        addNodeContainment(v1, v2);

    }

    /**
     * This method adds a new OWL class expression to the classification graph.
     * First, all the possible equivalent forms of this expression will be found
     * (by applying the rules mainly on the someValuesFrom and allValuesFrom
     * elements). Then, for each of the forms, the basicAddClass method will be
     * called to add it into the classification graph.
     * 
     * @param c
     *            The class expression to be added
     * @return The node which contains the new added class expression
     */
    public OWLClassNode addClass(OWLClass c) {
            
        c = replaceEnumeratedClass(c);

        return addClass1(c);
    }

    protected OWLClassNode addClass1(OWLClass c) {
          
        OWLClass[] forms = ReasonerUtils.getAllForms(pGraph, c);

        OWLClassNode v = getNodeByClass(forms[0], false);
        if (v != null) {
            return v;
        }
        int id = identifier.getOWLClassID(forms[0], false, true);

 
        for (int i = 0; i < forms.length; i++) {       
            v = basicAddClass(forms[i], id, v);
        }
   
        return v;
    }

    protected OWLClassNode basicAddClass(OWLClass c, int cID,
            OWLClassNode position) {
        List superClassList = new ArrayList();
        superClassList.add(thingNode);
        List subClassList = new ArrayList();
        subClassList.add(nothingNode);

        if (position == null) {
            position = new OWLClassNode(this, cID);
        } else {
            position.addOWLClassID(cID);
        }
        if (c instanceof IntersectionClass) {
            Iterator iter = ((IntersectionClass) c).getOWLIntersectionOf()
                    .iterator();
            List inNodeList = new ArrayList();
            while (iter.hasNext()) {
                OWLClassNode v = addClass1((OWLClass) iter.next());
                inNodeList.add(v);
            }
            for (int i = inNodeList.size() - 1; i >= 0; i--) {
                OWLClassNode v1 = (OWLClassNode) inNodeList.get(i);
                for (int j = i - 1; j >= 0; j--) {
                    OWLClassNode v2 = (OWLClassNode) inNodeList.get(j);
                    if (isSubClassNode(v1, v2)) {
                        inNodeList.remove(j);
                        i--;
                    } else if (isSubClassNode(v2, v1)) {
                        inNodeList.remove(i);
                        break;
                    }
                }
            }
            superClassList.addAll(inNodeList);
            subClassList.addAll(getSubClassNodes(inNodeList));
            int targetID = getNextTargetID();
            for (int i = 0; i < inNodeList.size(); i++) {
                new ConjunctionLink(this, ((OWLClassNode) inNodeList.get(i))
                        .getAlternativeNode(), position.getAlternativeNode(),
                        targetID);
            }
            targetValue[targetID] = inNodeList.size();
        } else if (c instanceof UnionClass) {
            Iterator iter = ((UnionClass) c).getOWLUnionOf().iterator();
            List unNodeList = new ArrayList();
            while (iter.hasNext()) {
                OWLClassNode v = addClass1((OWLClass) iter.next());
                unNodeList.add(v);
            }
            for (int i = unNodeList.size() - 1; i >= 0; i--) {
                OWLClassNode v1 = (OWLClassNode) unNodeList.get(i);
                for (int j = i - 1; j >= 0; j--) {
                    OWLClassNode v2 = (OWLClassNode) unNodeList.get(j);
                    if (isSubClassNode(v2, v1)) {
                        unNodeList.remove(j);
                        i--;
                    } else if (isSubClassNode(v1, v2)) {
                        unNodeList.remove(i);
                        break;
                    }
                }
            }
            subClassList.addAll(unNodeList);
            superClassList.addAll(getSuperClassNodes(unNodeList));
            int targetID = getNextTargetID();
            for (int i = 0; i < unNodeList.size(); i++) {
                new DisjunctionLink(this, ((OWLClassNode) unNodeList.get(i))
                        .getAlternativeNode(), position.getAlternativeNode(),
                        targetID);
            }
            targetValue[targetID] = unNodeList.size();
        } else if (c instanceof AllValuesFromRestriction) {
            int pID = pGraph.getCreatePropertyID(((AllValuesFromRestriction) c)
                    .getOWLOnProperty());
            OWLClassNode v = addClass1((OWLClass) ((AllValuesFromRestriction) c)
                    .getOWLAllValuesFrom());
            new AllValuesFromLink(this, position.getAlternativeNode(), v
                    .getAlternativeNode(), pID);
            for (Iterator iter = getSuperClassNodes(v).iterator(); iter
                    .hasNext();) {
                Iterator iter1 = ((OWLClassNode) iter.next()).allValuesFromByLinks
                        .iterator();
                while (iter1.hasNext()) {
                    AllValuesFromLink link = (AllValuesFromLink) iter1.next();
                    if (pGraph.isSubProperty(link.propertyID, pID)) {
                        superClassList.add(link.getSource());
                    }
                }
            }
            for (Iterator iter = getSubClassNodes(v).iterator(); iter.hasNext();) {
                Iterator iter1 = ((OWLClassNode) iter.next()).allValuesFromByLinks
                        .iterator();
                while (iter1.hasNext()) {
                    AllValuesFromLink link = (AllValuesFromLink) iter1.next();
                    if (pGraph.isSubProperty(pID, link.propertyID)) {
                        subClassList.add(link.getSource());
                    }
                }
            }
        } else if (c instanceof SomeValuesFromRestriction) {
            int pID = pGraph
                    .getCreatePropertyID(((SomeValuesFromRestriction) c)
                            .getOWLOnProperty());
            OWLClassNode v = addClass1((OWLClass) ((SomeValuesFromRestriction) c)
                    .getOWLSomeValuesFrom());
            new SomeValuesFromLink(this, position.getAlternativeNode(), v
                    .getAlternativeNode(), pID);
            for (Iterator iter = getSuperClassNodes(v).iterator(); iter
                    .hasNext();) {
                Iterator iter1 = ((OWLClassNode) iter.next()).someValuesFromByLinks
                        .iterator();
                while (iter1.hasNext()) {
                    SomeValuesFromLink link = (SomeValuesFromLink) iter1.next();
                    if (pGraph.isSubProperty(pID, link.propertyID)) {
                        superClassList.add(link.getSource());
                    }
                }
            }
            for (Iterator iter = getSubClassNodes(v).iterator(); iter.hasNext();) {
                Iterator iter1 = ((OWLClassNode) iter.next()).someValuesFromByLinks
                        .iterator();
                while (iter1.hasNext()) {
                    SomeValuesFromLink link = (SomeValuesFromLink) iter1.next();
                    if (pGraph.isSubProperty(link.propertyID, pID)) {
                        subClassList.add(link.getSource());
                    }
                }
            }
        } else if (c instanceof MinCardinalityRestriction) {
            SomeValuesFromRestriction someRT = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            someRT.setOWLOnProperty(((MinCardinalityRestriction) c)
                    .getOWLOnProperty());
            someRT.setOWLSomeValuesFrom(ReasonerUtils
                    .cloneClass(ReasonerUtils.thing));
            OWLClassNode v = addClass1(someRT);
            superClassList.add(v);
        }

        for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
            addNodeContainment((OWLClassNode) iter.next(), position);
        }
        for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
            addNodeContainment(position, (OWLClassNode) iter.next());
        }
        return position.getAlternativeNode();
    }

    protected void addNodeContainment(OWLClassNode v1, OWLClassNode v2) {
        if (isSubClassNode(v2, v1)) {
            return;
        }
       
        adjustNodeList(v1, v2);

        v1 = v1.getAlternativeNode();
        v2 = v2.getAlternativeNode();
        ArrayList tempList = new ArrayList();
        List superClassList;
        List subClassList;
        if (v1 != v2) {
            
            superClassList = getSuperClassNodes(v1);
            subClassList = getSubClassNodes(v2);
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 0;
            }
            for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
                flags[((OWLClassNode) iter.next()).topologicalPosition] = 1;
            }
            for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
                for (Iterator iter1 = ((OWLClassNode) iter.next()).containmentLinks
                        .iterator(); iter1.hasNext();) {
                    ContainmentLink link = (ContainmentLink) iter1.next();
                    if (flags[link.getTarget().topologicalPosition] == 1) {
                        tempList.add(link);
                    }
                }
            }
            new ContainmentLink(this, v1, v2);
        
        } else {
            superClassList = getSuperClassNodesWithRemoveRedundancy(v1);
            subClassList = getSubClassNodesWithRemoveRedundancy(v2);
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 0;
            }
            for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
                flags[((OWLClassNode) iter.next()).topologicalPosition] = 1;
            }
            for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
                for (Iterator iter1 = ((OWLClassNode) iter.next()).containmentLinks
                        .iterator(); iter1.hasNext();) {
                    ContainmentLink link = (ContainmentLink) iter1.next();
                    if (flags[link.getTarget().topologicalPosition] == 1
                        && link.getSource() != v1 && link.getTarget() != v2) {
                        tempList.add(link);
                    }
                }
            }
        }

        for (Iterator iter = tempList.iterator(); iter.hasNext();) {
            ContainmentLink link = (ContainmentLink) iter.next();
            link.setSource(null);
            link.setTarget(null);
        }
        HashMap superMap = new HashMap();
        HashMap subMap = new HashMap();
        for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
            for (Iterator iter1 = ((OWLClassNode) iter.next()).allValuesFromByLinks
                    .iterator(); iter1.hasNext();) {
                AllValuesFromLink link = (AllValuesFromLink) iter1.next();
                Integer key = new Integer(link.getPropertyID());
                if (!superMap.containsKey(key)) {
                    superMap.put(key, new ArrayList());
                }
                ((ArrayList) superMap.get(key)).add(link.getSource());
            }
        }
        for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
            for (Iterator iter1 = ((OWLClassNode) iter.next()).allValuesFromByLinks
                    .iterator(); iter1.hasNext();) {
                AllValuesFromLink link = (AllValuesFromLink) iter1.next();
                Integer key = new Integer(link.getPropertyID());
                if (!subMap.containsKey(key)) {
                    subMap.put(key, new ArrayList());
                }
                ((ArrayList) subMap.get(key)).add(link.getSource());
            }
        }
        for (Iterator superIter = superMap.entrySet().iterator(); superIter
                .hasNext();) {
            Map.Entry superEntry = (Map.Entry) superIter.next();
            for (Iterator subIter = subMap.entrySet().iterator(); subIter
                    .hasNext();) {
                Map.Entry subEntry = (Map.Entry) subIter.next();
                if (pGraph.isSubProperty(((Integer) superEntry.getKey())
                        .intValue(), ((Integer) subEntry.getKey()).intValue())) {
                    for (Iterator iter = ((ArrayList) superEntry.getValue())
                            .iterator(); iter.hasNext();) {
                        OWLClassNode superV = (OWLClassNode) iter.next();
                        for (Iterator iter1 = ((ArrayList) subEntry.getValue())
                                .iterator(); iter1.hasNext();) {
                            OWLClassNode subV = (OWLClassNode) iter1.next();
                            addNodeContainment(superV, subV);
                        }
                    }
                }
            }
        }
        superMap.clear();
        subMap.clear();

        for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
            for (Iterator iter1 = ((OWLClassNode) iter.next()).someValuesFromByLinks
                    .iterator(); iter1.hasNext();) {
                SomeValuesFromLink link = (SomeValuesFromLink) iter1.next();
                Integer key = new Integer(link.getPropertyID());
                if (!superMap.containsKey(key)) {
                    superMap.put(key, new ArrayList());
                }
                ((ArrayList) superMap.get(key)).add(link.getSource());
            }
        }
        for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
            for (Iterator iter1 = ((OWLClassNode) iter.next()).someValuesFromByLinks
                    .iterator(); iter1.hasNext();) {
                SomeValuesFromLink link = (SomeValuesFromLink) iter1.next();
                Integer key = new Integer(link.getPropertyID());
                if (!subMap.containsKey(key)) {
                    subMap.put(key, new ArrayList());
                }
                ((ArrayList) subMap.get(key)).add(link.getSource());
            }
        }
        for (Iterator superIter = superMap.entrySet().iterator(); superIter
                .hasNext();) {
            Map.Entry superEntry = (Map.Entry) superIter.next();
            for (Iterator subIter = subMap.entrySet().iterator(); subIter
                    .hasNext();) {
                Map.Entry subEntry = (Map.Entry) subIter.next();
                if (pGraph
                        .isSubProperty(
                                ((Integer) subEntry.getKey()).intValue(),
                                ((Integer) superEntry.getKey()).intValue())) {
                    for (Iterator iter = ((ArrayList) superEntry.getValue())
                            .iterator(); iter.hasNext();) {
                        OWLClassNode superV = (OWLClassNode) iter.next();
                        for (Iterator iter1 = ((ArrayList) subEntry.getValue())
                                .iterator(); iter1.hasNext();) {
                            OWLClassNode subV = (OWLClassNode) iter1.next();
                            addNodeContainment(superV, subV);
                        }
                    }
                }
            }
        }
        
        for (Iterator iter = superClassList.iterator(); iter.hasNext();) {
            OWLClassNode v = (OWLClassNode) iter.next();
            OWLClassNode vv = findNextSubClassDisjunction(v);
            if (vv != null) {
                addNodeContainment(v, vv);
            }
        }

        for (Iterator iter = subClassList.iterator(); iter.hasNext();) {
            OWLClassNode v = (OWLClassNode) iter.next();
            OWLClassNode vv = findNextSuperClassConjunction(v);
            if (vv != null) {
                addNodeContainment(vv, v);
            }
        }
        
    }

    public boolean isSubClassNode(OWLClassNode v1, OWLClassNode v2) {
        
        if(v1==nothingNode || v2==thingNode) {
            return true;
        }
        
        int pa = v2.getAlternativeNode().topologicalPosition;
        int pb = v1.getAlternativeNode().topologicalPosition;
        if (pa > pb) {
            return false;
        }
        for (int i = pa; i <= pb; i++) {
            flags[i] = 0;
        }
        flags[pa] = 1;
        for (int i = pa; i < pb; i++) {
            if (flags[i] == 1) {
                Iterator iter = ((OWLClassNode) nodeList.get(i)).containmentLinks
                        .iterator();
                while (iter.hasNext()) {
                    flags[((ContainmentLink) iter.next()).getTarget().topologicalPosition] = 1;
                }
            }
        }
        return (flags[pb] == 1);
    }

    public OWLClassNode findNextSubClassDisjunction(OWLClassNode v) {
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = nodeList.size() - 1; i >= p; i--) {
            flags[i] = 0;
        }
        for (int i = 0; i < targetCounter.length; i++) {
            targetCounter[i] = 0;
        }
        List tempList = new ArrayList();
        flags[p] = 1;
        for (int i = p; i < nodeList.size(); i++) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                Iterator iter = v1.containmentLinks.iterator();
                while (iter.hasNext()) {
                    flags[((ContainmentLink) iter.next()).getTarget().topologicalPosition] = 1;
                }
                iter = v1.disjunctionLinks.iterator();
                while (iter.hasNext()) {
                    DisjunctionLink link = (DisjunctionLink) iter.next();
                    if (targetValue[link.getTargetID()] == (++targetCounter[link
                            .getTargetID()])) {
                        if (flags[link.getTarget().topologicalPosition] == 0) {
                            tempList.add(link.getTarget());
                        }
                    }
                }
            }
        }
        for (Iterator iter = tempList.iterator(); iter.hasNext();) {
            OWLClassNode v1 = (OWLClassNode) iter.next();
            if (flags[v1.topologicalPosition] == 0) {
                return v1;
            }
        }
        return null;
    }

    public OWLClassNode findNextSuperClassConjunction(OWLClassNode v) {
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = 0; i <= p; i++) {
            flags[i] = 0;
        }
        for (int i = 0; i < targetCounter.length; i++) {
            targetCounter[i] = 0;
        }
        List tempList = new ArrayList();
        flags[p] = 1;
        for (int i = p; i >= 0; i--) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                Iterator iter = v1.containmentByLinks.iterator();
                while (iter.hasNext()) {
                    flags[((ContainmentLink) iter.next()).getSource().topologicalPosition] = 1;
                }
                iter = v1.conjunctionLinks.iterator();
                while (iter.hasNext()) {
                    ConjunctionLink link = (ConjunctionLink) iter.next();
                    if (targetValue[link.getTargetID()] == (++targetCounter[link
                            .getTargetID()])) {
                        if (flags[link.getTarget().topologicalPosition] == 0) {
                            tempList.add(link.getTarget());
                        }
                    }
                }
            }
        }
        for (Iterator iter = tempList.iterator(); iter.hasNext();) {
            OWLClassNode v1 = (OWLClassNode) iter.next();
            if (flags[v1.topologicalPosition] == 0) {
                return v1;
            }
        }
        return null;
    }

    private List getSubClassNodesWithRemoveRedundancy(OWLClassNode v) {
        List result = new ArrayList();
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = nodeList.size() - 1; i >= p; i--) {
            flags[i] = 0;
        }
        flags[p] = 1;
        ContainmentLink[] tempArray = new ContainmentLink[nodeList.size()];
        for (int i = p; i < nodeList.size(); i++) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                result.add(v1);
                Iterator iter = v1.containmentLinks.iterator();
                if (i == p) {
                    List removeList = new ArrayList();
                    while (iter.hasNext()) {
                        ContainmentLink link = (ContainmentLink) iter.next();
                        int target = link.getTarget().topologicalPosition;
                        if (flags[target] == 1) {
                            removeList.add(link);
                        } else {
                            flags[target] = 1;
                            tempArray[target] = link;
                        }
                    }
                    for (int j = 0; j < removeList.size(); j++) {
                        ContainmentLink link = ((ContainmentLink) removeList
                                .get(j));
                        link.setSource(null);
                        link.setTarget(null);
                    }
                } else {
                    while (iter.hasNext()) {
                        int target = ((ContainmentLink) iter.next())
                                .getTarget().topologicalPosition;
                        flags[target] = 1;
                        if (tempArray[target] != null) {
                            tempArray[target].setSource(null);
                            tempArray[target].setTarget(null);
                            tempArray[target] = null;
                        }
                    }
                }
            }
        }
        return result;
    }

    private List getSuperClassNodesWithRemoveRedundancy(OWLClassNode v) {
        List result = new ArrayList();
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = 0; i <= p; i++) {
            flags[i] = 0;
        }
        flags[p] = 1;
        ContainmentLink[] tempArray = new ContainmentLink[p + 1];
        for (int i = p; i >= 0; i--) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                result.add(v1);
                Iterator iter = v1.containmentByLinks.iterator();
                if (i == p) {
                    List removeList = new ArrayList();
                    while (iter.hasNext()) {
                        ContainmentLink link = (ContainmentLink) iter.next();
                        int source = link.getSource().topologicalPosition;
                        if (flags[source] == 1) {
                            removeList.add(link);
                        } else {
                            flags[source] = 1;
                            tempArray[source] = link;
                        }
                    }
                    for (int j = 0; j < removeList.size(); j++) {
                        ContainmentLink link = ((ContainmentLink) removeList
                                .get(j));
                        link.setSource(null);
                        link.setTarget(null);
                    }
                } else {
                    while (iter.hasNext()) {
                        int source = ((ContainmentLink) iter.next())
                                .getSource().topologicalPosition;
                        flags[source] = 1;
                        if (tempArray[source] != null) {
                            tempArray[source].setSource(null);
                            tempArray[source].setTarget(null);
                            tempArray[source] = null;
                        }
                    }
                }
            }
        }
        return result;
    }

    public List getSubClassNodes(OWLClassNode v) {
        List result = new ArrayList();
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = nodeList.size() - 1; i >= p; i--) {
            flags[i] = 0;
        }
        flags[p] = 1;
        for (int i = p; i < nodeList.size(); i++) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                result.add(v1);
                Iterator iter = v1.containmentLinks.iterator();
                while (iter.hasNext()) {
                    flags[((ContainmentLink) iter.next()).getTarget().topologicalPosition] = 1;
                }
            }
        }
        return result;
    }

    public List getSuperClassNodes(OWLClassNode v) {
        List result = new ArrayList();
        int p = v.getAlternativeNode().topologicalPosition;
        for (int i = 0; i <= p; i++) {
            flags[i] = 0;
        }
        flags[p] = 1;
        for (int i = p; i >= 0; i--) {
            if (flags[i] == 1) {
                OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                result.add(v1);
                Iterator iter = v1.containmentByLinks.iterator();
                while (iter.hasNext()) {
                    flags[((ContainmentLink) iter.next()).getSource().topologicalPosition] = 1;
                }
            }
        }
        return result;
    }

    public List getSubClassNodes(Collection c) {
        List result = new ArrayList();
        if (c.size() <= 31) {
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 0;
            }
            Iterator iter = c.iterator();
            int index = 0;
            int min = nodeList.size();
            while (iter.hasNext()) {
                int p = ((OWLClassNode) iter.next()).getAlternativeNode().topologicalPosition;
                flags[p < min ? min = p : p] |= 1 << (index++);
            }
            int target = (1 << index) - 1;
            for (int i = min; i < nodeList.size(); i++) {
                if (flags[i] > 0) {
                    OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                    if (flags[i] == target) {
                        result.add(v1);
                    }
                    iter = v1.containmentLinks.iterator();
                    while (iter.hasNext()) {
                        flags[((ContainmentLink) iter.next()).getTarget().topologicalPosition] |= flags[i];
                    }
                }
            }
        } else {
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 1;
            }
            Iterator iter = c.iterator();
            int flags1[] = new int[nodeList.size()];
            while (iter.hasNext()) {
                for (int i = 0; i < flags1.length; i++) {
                    flags1[i] = 0;
                }
                int p = ((OWLClassNode) iter.next()).getAlternativeNode().topologicalPosition;
                flags1[p] = 1;
                for (int i = 0; i < flags1.length; i++) {
                    if (flags1[i] == 1) {
                        Iterator iter1 = ((OWLClassNode) nodeList.get(i)).containmentLinks
                                .iterator();
                        while (iter1.hasNext()) {
                            flags1[((ContainmentLink) iter1.next()).getTarget().topologicalPosition] = 1;
                        }
                    } else {
                        flags[i] = 0;
                    }
                }
            }
            for (int i = 0; i < nodeList.size(); i++) {
                if (flags[i] == 1) {
                    result.add(nodeList.get(i));
                }
            }
        }
        return result;
    }

    public List getSuperClassNodes(Collection c) {
        List result = new ArrayList();
        if (c.size() <= 31) {
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 0;
            }
            Iterator iter = c.iterator();
            int index = 0;
            int max = -1;
            while (iter.hasNext()) {
                int p = ((OWLClassNode) iter.next()).getAlternativeNode().topologicalPosition;
                flags[p > max ? max = p : p] |= 1 << (index++);
            }
            int target = (1 << index) - 1;
            for (int i = max; i >= 0; i--) {
                if (flags[i] > 0) {
                    OWLClassNode v1 = (OWLClassNode) nodeList.get(i);
                    if (flags[i] == target) {
                        result.add(v1);
                    }
                    iter = v1.containmentByLinks.iterator();
                    while (iter.hasNext()) {
                        flags[((ContainmentLink) iter.next()).getSource().topologicalPosition] |= flags[i];
                    }
                }
            }
        } else {
            for (int i = nodeList.size() - 1; i >= 0; i--) {
                flags[i] = 1;
            }
            Iterator iter = c.iterator();
            int flags1[] = new int[nodeList.size()];
            while (iter.hasNext()) {
                for (int i = 0; i < flags1.length; i++) {
                    flags1[i] = 0;
                }
                int p = ((OWLClassNode) iter.next()).getAlternativeNode().topologicalPosition;
                flags1[p] = 1;
                for (int i = flags1.length - 1; i >= 0; i--) {
                    if (flags1[i] == 1) {
                        Iterator iter1 = ((OWLClassNode) nodeList.get(i)).containmentByLinks
                                .iterator();
                        while (iter1.hasNext()) {
                            flags1[((ContainmentLink) iter1.next()).getTarget().topologicalPosition] = 1;
                        }
                    } else {
                        flags[i] = 0;
                    }
                }
            }
            for (int i = 0; i < nodeList.size(); i++) {
                if (flags[i] == 1) {
                    result.add(nodeList.get(i));
                }
            }
        }
        return result;
    }

    private void adjustNodeList(OWLClassNode v1, OWLClassNode v2) {
        int pa = v2.getAlternativeNode().topologicalPosition;
        int pb = v1.getAlternativeNode().topologicalPosition;
        if (pa >= pb) {
            return;
        }
        for (int i = pa; i <= pb; i++) {
            flags[i] = 0;
        }
        flags[pa] = 1;
        for (int i = pa; i <= pb; i++) {
            if (flags[i] == 1) {
                Iterator iter = ((OWLClassNode) nodeList.get(i)).containmentLinks
                        .iterator();
                while (iter.hasNext()) {
                    int p = ((ContainmentLink) iter.next()).getTarget().topologicalPosition;
                    if (p <= pb) {
                        flags[p] = 1;
                    }
                }
            }
        }
        flags[pb] |= 2;
        for (int i = pb; i >= pa; i--) {
            if ((flags[i] & 2) == 2) {
                Iterator iter = ((OWLClassNode) nodeList.get(i)).containmentByLinks
                        .iterator();
                while (iter.hasNext()) {
                    int p = ((ContainmentLink) iter.next()).getSource().topologicalPosition;
                    if (p >= pa) {
                        flags[p] |= 2;
                    }
                }
            }
        }
        OWLClassNode tempArray[] = new OWLClassNode[pb - pa + 1];
        int index = 0;
        for (int i = pa; i <= pb; i++) {
            if (flags[i] == 2) {
                tempArray[index++] = (OWLClassNode) nodeList.get(i);
            }
        }
        if (flags[pa] == 3) {
            OWLClassNode v = (OWLClassNode) nodeList.get(pa);
            for (int i = pa + 1; i <= pb; i++) {
                if (flags[i] == 3) {
                    combineNode(v, (OWLClassNode) nodeList.get(i));
                }
            }
            tempArray[index++] = v;
        }
        for (int i = pa; i <= pb; i++) {
            if (flags[i] <= 1) {
                tempArray[index++] = (OWLClassNode) nodeList.get(i);
            }
        }
        for (int i = 0; i < index; i++) {
            nodeList.set(i + pa, tempArray[i]);
            tempArray[i].topologicalPosition = i + pa;
        }
        if (index < pb - pa + 1) {
            nodeList.subList(pa + index, pb + 1).clear();
            for (int i = nodeList.size() - 1; i >= pa + index; i--) {
                ((OWLClassNode) nodeList.get(i)).topologicalPosition = i;
            }
        }
    }

    private void combineNode(OWLClassNode v1, OWLClassNode v2) {
        if (v1.graph != this || v2.graph != this) {
            throw new IllegalArgumentException();
        }
        v1 = v1.getAlternativeNode();
        v2 = v2.getAlternativeNode();
        if (v1 != v2) {
            Iterator iter;

            iter = v2.classes.iterator();
            while (iter.hasNext()) {
                Integer id = (Integer) iter.next();
                v1.classes.add(id);
                classToNodeMap.put(id, v1);
            }
            v2.classes.clear();

            iter = v2.containmentLinks.iterator();
            while (iter.hasNext()) {
                ContainmentLink link = (ContainmentLink) iter.next();
                v1.containmentLinks.add(link);
                link.source = v1;
            }

            iter = v2.containmentByLinks.iterator();
            while (iter.hasNext()) {
                ContainmentLink link = (ContainmentLink) iter.next();
                v1.containmentByLinks.add(link);
                link.target = v1;
            }

            iter = v2.conjunctionLinks.iterator();
            while (iter.hasNext()) {
                ConjunctionLink link = (ConjunctionLink) iter.next();
                v1.conjunctionLinks.add(link);
                link.source = v1;
            }

            iter = v2.disjunctionLinks.iterator();
            while (iter.hasNext()) {
                DisjunctionLink link = (DisjunctionLink) iter.next();
                v1.disjunctionLinks.add(link);
                link.source = v1;
            }

            iter = v2.allValuesFromByLinks.iterator();
            while (iter.hasNext()) {
                AllValuesFromLink link = (AllValuesFromLink) iter.next();
                v1.allValuesFromByLinks.add(link);
                link.target = v1;
            }

            iter = v2.someValuesFromByLinks.iterator();
            while (iter.hasNext()) {
                SomeValuesFromLink link = (SomeValuesFromLink) iter.next();
                v1.someValuesFromByLinks.add(link);
                link.target = v1;
            }

            v2.alternativeNode = v1;
            if (v2 == thingNode) {
                thingNode = v1;
            }
            if (v2 == nothingNode) {
                nothingNode = v1;
            }
        }
    }

    private synchronized int getNextTargetID() {
        if (nextTargetID >= targetValue.length) {
            int newLength = nextTargetID * 3 / 2;
            int newArray[] = new int[newLength];
            System.arraycopy(targetValue, 0, newArray, 0, targetValue.length);
            targetValue = newArray;

            targetCounter = new int[newLength];
        }
        return nextTargetID++;
    }

    void addNode(OWLClassNode v) {
        v.topologicalPosition = nodeList.size();
        nodeList.add(v);
        if (nodeList.size() > flags.length) {
            flags = new int[(int) (nodeList.size() * 3 / 2)];
        }
    }

    public OWLClass getOWLClass(int classID) {
        return restoreEnumClass(identifier.getOWLClass(classID));
    }

    protected OWLClass replaceEnumeratedClass(OWLClass c) {
        if (c instanceof IntersectionClass) {
            IntersectionClass c1 = OWLFactory.eINSTANCE
                    .createIntersectionClass();
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            for (int i = 0; i < l.size(); i++) {
                c1.getOWLIntersectionOf().add(
                        replaceEnumeratedClass((OWLClass) l.get(i)));
            }
            return c1;
        } else if (c instanceof UnionClass) {
            UnionClass c1 = OWLFactory.eINSTANCE.createUnionClass();
            List l = ((UnionClass) c).getOWLUnionOf();
            for (int i = 0; i < l.size(); i++) {
                c1.getOWLUnionOf().add(
                        replaceEnumeratedClass((OWLClass) l.get(i)));
            }
            return c1;
        } else if (c instanceof AllValuesFromRestriction) {
            AllValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createAllValuesFromRestriction();
            OWLObjectProperty p = OWLFactory.eINSTANCE
                    .createOWLObjectProperty();
            p.setNamespace(((AllValuesFromRestriction) c).getOWLOnProperty()
                    .getNamespace());
            p.setLocalName(((AllValuesFromRestriction) c).getOWLOnProperty()
                    .getLocalName());
            c1.setOWLOnProperty(p);
            c1
                    .setOWLAllValuesFrom(replaceEnumeratedClass((OWLClass) ((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom()));
            return c1;
        } else if (c instanceof SomeValuesFromRestriction) {
            SomeValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            OWLObjectProperty p = OWLFactory.eINSTANCE
                    .createOWLObjectProperty();
            p.setNamespace(((SomeValuesFromRestriction) c).getOWLOnProperty()
                    .getNamespace());
            p.setLocalName(((SomeValuesFromRestriction) c).getOWLOnProperty()
                    .getLocalName());
            c1.setOWLOnProperty(p);
            c1
                    .setOWLSomeValuesFrom(replaceEnumeratedClass((OWLClass) ((SomeValuesFromRestriction) c)
                            .getOWLSomeValuesFrom()));
            return c1;
        } else if (c instanceof EnumeratedClass) {
            return registerEnumClass((EnumeratedClass) c);
        } else if (c instanceof HasValueRestriction) {
            SomeValuesFromRestriction c2 = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            OWLObjectProperty p = OWLFactory.eINSTANCE
                    .createOWLObjectProperty();
            p.setNamespace(((HasValueRestriction) c).getOWLOnProperty()
                    .getNamespace());
            p.setLocalName(((HasValueRestriction) c).getOWLOnProperty()
                    .getLocalName());
            c2.setOWLOnProperty(p);
            if (((HasValueRestriction) c).getOWLHasValue() instanceof Individual) {
                EnumeratedClass c1 = OWLFactory.eINSTANCE
                        .createEnumeratedClass();
                c1.getOWLOneOf()
                        .add(((HasValueRestriction) c).getOWLHasValue());
                c2.setOWLSomeValuesFrom(registerEnumClass(c1));
            } else {
                OWLClass c1 = createPrimitiveClassForLiteral((RDFSLiteral) ((HasValueRestriction) c)
                        .getOWLHasValue());
                c2.setOWLSomeValuesFrom(c1);
            }
            return c2;
        } else if (c instanceof MinCardinalityRestriction) {
            SomeValuesFromRestriction someRT = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            someRT.setOWLOnProperty(((MinCardinalityRestriction) c)
                    .getOWLOnProperty());
            someRT.setOWLSomeValuesFrom(ReasonerUtils
                    .cloneClass(ReasonerUtils.thing));
            
            OWLClass c1 = OWLFactory.eINSTANCE.createOWLClass();
            c1.setNamespace(c.getNamespace());
            c1.setLocalName(c.getLocalName());
            addAxiom(someRT,c1);
            
            return c1;
        } else {
            OWLClass c1 = OWLFactory.eINSTANCE.createOWLClass();
            c1.setNamespace(c.getNamespace());
            c1.setLocalName(c.getLocalName());
            return c1;
        }
    }

    /**
     * This method will register a new enumerated class into the knowledge base,
     * and return a primitive class which is used to represent that enumerated
     * class. If a same enumerated class has already been registered, the
     * primitive class representing that class will be returned instead. All
     * possible axiom on this new added class will be added to the knowledge
     * base too.
     * 
     * @param c
     *            The enumerated class to be registered
     * @return The primitive class which is used to replace the enumerated
     *         class.
     */
    protected OWLClass registerEnumClass(EnumeratedClass c) {
        int total = c.getOWLOneOf().size();
        if (total == 0) {
            return ReasonerUtils.cloneClass(ReasonerUtils.nothing);
        }
        int newEnumClassArray[] = new int[total];
        int ii = 0;
        for (int i = 0; i < nextEnumClassID; i++) {
            enumClassCounter[i] = 0;
        }
        for (int i = 0; i < nextIndividualID; i++) {
            individualFlag[i] = 0;
        }

        Iterator iter = c.getOWLOneOf().iterator();
        while (iter.hasNext()) {
            Individual ind = (Individual) iter.next();
            String uri = ind.getURI();
            Integer index = (Integer) individualIndex.get(uri);
            if (index == null) {
                int i = getNextIndividualID();
                individualIndex.put(uri, new Integer(i));
                individualName[i] = uri;
                individuals[i] = ind;
                individualReference[i] = new LinkedList();
                individualFlag[i] = 1;
                newEnumClassArray[ii++] = i;
            } else {
                if (individualFlag[index.intValue()] == 0) {
                    individualFlag[index.intValue()] = 1;
                    newEnumClassArray[ii++] = index.intValue();
                    Iterator iter1 = individualReference[index.intValue()]
                            .iterator();
                    while (iter1.hasNext()) {
                        int enumID = ((Integer) iter1.next()).intValue();
                        enumClassCounter[enumID]++;
                    }
                }
            }
        }

        if (ii < total) {
            int[] temp = new int[ii];
            System.arraycopy(newEnumClassArray, 0, temp, 0, ii);
            newEnumClassArray = temp;
            total = ii;
        }

        int[] newSuperClasses = new int[nextEnumClassID];
        int[] newSubClasses = new int[nextEnumClassID];
        int[] newDisjoints = new int[nextEnumClassID];
        int nSuperClasses = 0, nSubClasses = 0, nDisjoints = 0;
        for (int i = 0; i < nextEnumClassID; i++) {
            if (enumClassCounter[i] == enumClass[i].length) {
                if (enumClassCounter[i] == total) {
                    return createPrimitiveClassForEnum(i);
                } else {
                    newSubClasses[nSubClasses++] = i;
                }
            } else if (enumClassCounter[i] == total) {
                newSuperClasses[nSuperClasses++] = i;
            } else if (enumClassCounter[i] == 0) {
                newDisjoints[nDisjoints++] = i;
            }
        }

        int newEnumClassID = getNextEnumClassID();
        Integer newIDObject = new Integer(newEnumClassID);
        OWLClass newEnumClass = createPrimitiveClassForEnum(newEnumClassID);
        for (int i = 0; i < total; i++) {
            individualReference[newEnumClassArray[i]].add(newIDObject);
        }
        enumClass[newEnumClassID] = newEnumClassArray;

        for (int i = 0; i < nSuperClasses; i++) {
            addAxiom(createPrimitiveClassForEnum(newSuperClasses[i]),
                    newEnumClass);
        }
        for (int i = 0; i < nSubClasses; i++) {
            addAxiom(newEnumClass,
                    createPrimitiveClassForEnum(newSubClasses[i]));
        }
        if (addDisjointsOnEnumClasses) {
            for (int i = 0; i < nDisjoints; i++) {
                IntersectionClass c1 = OWLFactory.eINSTANCE
                        .createIntersectionClass();
                c1.getOWLIntersectionOf().add(newEnumClass);
                c1.getOWLIntersectionOf().add(
                        createPrimitiveClassForEnum(newDisjoints[i]));
                addAxiom(ReasonerUtils.nothing, c1);
            }
        }
        return newEnumClass;
    }

    private OWLClass createPrimitiveClassForEnum(int index) {
        OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
        c.setNamespace(enumClassNamespace);
        c.setLocalName(String.valueOf(index));
        return c;
    }

    private OWLClass restoreEnumClass(OWLClass c) {
        if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            for (int i = 0; i < l.size(); i++) {
                l.set(i, restoreEnumClass((OWLClass) l.get(i)));
            }
            return c;
        } else if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            for (int i = 0; i < l.size(); i++) {
                l.set(i, restoreEnumClass((OWLClass) l.get(i)));
            }
            return c;
        } else if (c instanceof SomeValuesFromRestriction) {
            OWLClass c1 = restoreEnumClass((OWLClass) ((SomeValuesFromRestriction) c)
                    .getOWLSomeValuesFrom());
            if ((c1 instanceof EnumeratedClass)
                && ((EnumeratedClass) c1).getOWLOneOf().size() == 1) {
                HasValueRestriction c2 = OWLFactory.eINSTANCE
                        .createHasValueRestriction();
                OWLObjectProperty p = OWLFactory.eINSTANCE
                        .createOWLObjectProperty();
                p.setNamespace(((SomeValuesFromRestriction) c)
                        .getOWLOnProperty().getNamespace());
                p.setLocalName(((SomeValuesFromRestriction) c)
                        .getOWLOnProperty().getLocalName());
                c2.setOWLOnProperty(p);
                c2.setOWLHasValue((Individual) ((EnumeratedClass) c1)
                        .getOWLOneOf().get(0));
                return c2;
            } else if (c1.getNamespace() != null
                       && c1.getNamespace() == dataRangeNamespace) {
                HasValueRestriction c2 = OWLFactory.eINSTANCE
                        .createHasValueRestriction();
                OWLDatatypeProperty p = OWLFactory.eINSTANCE
                        .createOWLDatatypeProperty();
                p.setNamespace(((SomeValuesFromRestriction) c)
                        .getOWLOnProperty().getNamespace());
                p.setLocalName(((SomeValuesFromRestriction) c)
                        .getOWLOnProperty().getLocalName());
                c2.setOWLOnProperty(p);
                c2
                        .setOWLHasValue(literals[Integer.parseInt(c1
                                .getLocalName())]);
                return c2;
            }
            ((SomeValuesFromRestriction) c).setOWLSomeValuesFrom(c1);
            return c;
        } else if (c instanceof AllValuesFromRestriction) {
            ((AllValuesFromRestriction) c)
                    .setOWLAllValuesFrom(restoreEnumClass((OWLClass) ((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom()));
            return c;
        } else {
            if (c.getNamespace() == enumClassNamespace) {
                int index = Integer.parseInt(c.getLocalName());
                int[] enumerate = enumClass[index];
                EnumeratedClass c1 = OWLFactory.eINSTANCE
                        .createEnumeratedClass();
                for (int i = 0; i < enumerate.length; i++) {
                    c1.getOWLOneOf().add(individuals[enumerate[i]]);
                }
                return c1;
            } else {
                return c;
            }
        }
    }

    private synchronized int getNextIndividualID() {
        if (nextIndividualID >= individualName.length) {
            int newLength = nextIndividualID * 3 / 2;
            String[] newName = new String[newLength];
            System.arraycopy(individualName, 0, newName, 0,
                    individualName.length);
            individualName = newName;
            Individual[] newIndividual = new Individual[newLength];
            System.arraycopy(individuals, 0, newIndividual, 0,
                    individuals.length);
            individuals = newIndividual;
            LinkedList[] newReference = new LinkedList[newLength];
            System.arraycopy(individualReference, 0, newReference, 0,
                    individualReference.length);
            individualReference = newReference;
            int[] newFlag = new int[newLength];
            System.arraycopy(individualFlag, 0, newFlag, 0,
                    individualFlag.length);
            individualFlag = newFlag;
        }
        return nextIndividualID++;
    }

    private synchronized int getNextEnumClassID() {
        if (nextEnumClassID >= enumClass.length) {
            int newLength = nextEnumClassID * 3 / 2;
            int[][] newEnumClass = new int[newLength][];
            System.arraycopy(enumClass, 0, newEnumClass, 0, enumClass.length);
            enumClass = newEnumClass;
            int[] newCounter = new int[newLength];
            System.arraycopy(enumClassCounter, 0, newCounter, 0,
                    enumClassCounter.length);
            enumClassCounter = newCounter;
        }
        return nextEnumClassID++;
    }

    private OWLClass createPrimitiveClassForLiteral(RDFSLiteral l) {
        StringBuffer sb = new StringBuffer();
        if (l instanceof TypedLiteral) {
            sb.append(((TypedLiteral) l).getDatatype().getURI());
        } else {
            sb.append(" ");
        }
        sb.append(l.getLexicalForm());
        Integer id = (Integer) literalMap.get(sb.toString());
        if (id == null) {
            id = new Integer(getNextLiteralID());
            literals[id.intValue()] = l;
            literalMap.put(sb.toString(), id);
        }
        OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
        c.setNamespace(dataRangeNamespace);
        c.setLocalName(String.valueOf(id.intValue()));
        return c;
    }

    private synchronized int getNextLiteralID() {
        if (nextLiteralID >= literals.length) {
            int newLength = nextLiteralID * 3 / 2;
            RDFSLiteral[] newLiterals = new RDFSLiteral[newLength];
            System.arraycopy(literals, 0, newLiterals, 0, literals.length);
            literals = newLiterals;
        }
        return nextLiteralID++;
    }

    RDFPropertyGraph pGraph;

    StrictClassIdentifier identifier;

    HashMap classToNodeMap = new HashMap();

    OWLClassNode thingNode, nothingNode;

    private ArrayList nodeList = new ArrayList();

    private int flags[] = new int[10];

    private int nextTargetID = 0;

    private int targetValue[] = new int[10];

    private int targetCounter[] = new int[10];

    private static Namespace enumClassNamespace;

    private static Namespace dataRangeNamespace;

    static {
        enumClassNamespace = RDFSFactory.eINSTANCE.createNamespace();
        enumClassNamespace.setName("enum");
        enumClassNamespace.setURI("http://org.eclipse.odm/EnumClass#");
        dataRangeNamespace = RDFSFactory.eINSTANCE.createNamespace();
        dataRangeNamespace.setName("dataRange");
        dataRangeNamespace.setURI("http://org.eclipse.odm/DataRange#");
    }

    private HashMap individualIndex = new HashMap();

    private Individual individuals[] = new Individual[10];

    private String individualName[] = new String[10];

    private LinkedList individualReference[] = new LinkedList[10];

    private int individualFlag[] = new int[10];

    private int nextIndividualID = 0;

    private int enumClass[][] = new int[10][];

    private int enumClassCounter[] = new int[10];

    private int nextEnumClassID = 0;

    private RDFSLiteral literals[] = new RDFSLiteral[10];

    private HashMap literalMap = new HashMap();

    private int nextLiteralID = 0;

    private boolean addDisjointsOnEnumClasses = false;

    public boolean isAddDisjointsOnEnumClasses() {
        return addDisjointsOnEnumClasses;
    }

    public void setAddDisjointsOnEnumClasses(boolean addDisjointsOnEnumClasses) {
        this.addDisjointsOnEnumClasses = addDisjointsOnEnumClasses;
    }

    public void show() {
        int[][] containments = new int[nodeList.size()][nodeList.size()];
        for (int i = 0; i < nodeList.size(); i++) {
            System.out.println("Node No." + i + ":");
            System.out.println("-----------------------------------------");
            OWLClassNode node = (OWLClassNode) nodeList.get(i);
            Iterator iter = node.getOWLClassIDSet().iterator();
            while (iter.hasNext()) {
                ReasonerUtils.show(getOWLClass(((Integer) iter.next())
                        .intValue()), 0);
                System.out.println("-----------------------------------------");
            }
            iter = node.containmentLinks.iterator();
            while (iter.hasNext()) {
                containments[i][((ContainmentLink) iter.next()).getTarget().topologicalPosition]++;
            }
        }
        MyNumberFormat format = new MyNumberFormat(3);
        for (int i = 0; i < nodeList.size(); i++) {
            System.out.print(format.format(i) + "|");
            for (int j = 0; j < nodeList.size(); j++) {
                System.out.print(format.format(containments[i][j]));
            }
            System.out.println();
        }
        System.out.print("   +");
        for (int i = 0; i < nodeList.size(); i++) {
            System.out.print("---");
        }
        System.out.println("--");
        System.out.print("    ");
        for (int i = 0; i < nodeList.size(); i++) {
            System.out.print(format.format(i));
        }
        System.out.println();
    }

    public static void main(String args[]) throws Exception {
        RDFPropertyGraph pGraph = new RDFPropertyGraph();
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(ReasonerUtils.test);
        r.setLocalName("R");
        RDFProperty s = RDFSFactory.eINSTANCE.createRDFProperty();
        s.setNamespace(ReasonerUtils.test);
        s.setLocalName("S");
        RDFProperty t = RDFSFactory.eINSTANCE.createRDFProperty();
        t.setNamespace(ReasonerUtils.test);
        t.setLocalName("T");
        pGraph.addSubPropertyAxiom(r, s);
        pGraph.addSubPropertyAxiom(s, t);
        OWLClassClassificationGraph graph = new OWLClassClassificationGraph(
                pGraph);

        // testEnumClass(graph);
        // testAxioms(graph);
        // testCounterExample1(graph);
        testHasValue(graph);
    }

    public static void testHasValue(OWLClassClassificationGraph graph) {
        RDFSDatatype intType = RDFSFactory.eINSTANCE.createRDFSDatatype();
        intType.setNamespace(ReasonerUtils.xsd);
        intType.setLocalName("int");

        Individual i1 = OWLFactory.eINSTANCE.createIndividual();
        i1.setNamespace(ReasonerUtils.test);
        i1.setLocalName("individual1");
        Individual i2 = OWLFactory.eINSTANCE.createIndividual();
        i2.setNamespace(ReasonerUtils.test);
        i2.setLocalName("individual2");
        Individual i3 = OWLFactory.eINSTANCE.createIndividual();
        i3.setNamespace(ReasonerUtils.test);
        i3.setLocalName("individual3");

        TypedLiteral l1 = RDFSFactory.eINSTANCE.createTypedLiteral();
        l1.setDatatype(intType);
        l1.setLexicalForm("1");
        TypedLiteral l2 = RDFSFactory.eINSTANCE.createTypedLiteral();
        l2.setDatatype(intType);
        l2.setLexicalForm("2");
        TypedLiteral l3 = RDFSFactory.eINSTANCE.createTypedLiteral();
        l3.setDatatype(intType);
        l3.setLexicalForm("3");

        OWLObjectProperty r = OWLFactory.eINSTANCE.createOWLObjectProperty();
        r.setNamespace(ReasonerUtils.test);
        r.setLocalName("R");
        OWLDatatypeProperty s = OWLFactory.eINSTANCE
                .createOWLDatatypeProperty();
        s.setNamespace(ReasonerUtils.test);
        s.setLocalName("S");
        OWLDatatypeProperty t = OWLFactory.eINSTANCE
                .createOWLDatatypeProperty();
        t.setNamespace(ReasonerUtils.test);
        t.setLocalName("T");

        HasValueRestriction r1 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        r1.setOWLOnProperty(r);
        r1.setOWLHasValue(i1);
        HasValueRestriction r2 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        r2.setOWLOnProperty(r);
        r2.setOWLHasValue(i2);
        HasValueRestriction r3 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        r3.setOWLOnProperty(r);
        r3.setOWLHasValue(i3);
        HasValueRestriction s1 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        s1.setOWLOnProperty(s);
        s1.setOWLHasValue(l1);
        HasValueRestriction s2 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        s2.setOWLOnProperty(s);
        s2.setOWLHasValue(l2);
        HasValueRestriction s3 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        s3.setOWLOnProperty(s);
        s3.setOWLHasValue(l3);
        HasValueRestriction t1 = OWLFactory.eINSTANCE
                .createHasValueRestriction();
        t1.setOWLOnProperty(t);
        t1.setOWLHasValue(l1);

        graph.addClass(r1);
        graph.addClass(r2);
        graph.addClass(r3);
        graph.addClass(s1);
        graph.addClass(s2);
        graph.addClass(s3);
        graph.addClass(t1);

        graph.show();
    }

    public static void testEnumClass(OWLClassClassificationGraph graph) {
        Individual i1 = OWLFactory.eINSTANCE.createIndividual();
        i1.setNamespace(ReasonerUtils.test);
        i1.setLocalName("i1");
        Individual i2 = OWLFactory.eINSTANCE.createIndividual();
        i2.setNamespace(ReasonerUtils.test);
        i2.setLocalName("i2");
        Individual i3 = OWLFactory.eINSTANCE.createIndividual();
        i3.setNamespace(ReasonerUtils.test);
        i3.setLocalName("i3");
        Individual i4 = OWLFactory.eINSTANCE.createIndividual();
        i4.setNamespace(ReasonerUtils.test);
        i4.setLocalName("i4");
        Individual i5 = OWLFactory.eINSTANCE.createIndividual();
        i5.setNamespace(ReasonerUtils.test);
        i5.setLocalName("i5");

        EnumeratedClass ec1 = OWLFactory.eINSTANCE.createEnumeratedClass();
        ec1.getOWLOneOf().add(i1);
        ec1.getOWLOneOf().add(i2);

        EnumeratedClass ec2 = OWLFactory.eINSTANCE.createEnumeratedClass();
        ec2.getOWLOneOf().add(i1);
        ec2.getOWLOneOf().add(i3);

        EnumeratedClass ec3 = OWLFactory.eINSTANCE.createEnumeratedClass();
        ec3.getOWLOneOf().add(i1);
        ec3.getOWLOneOf().add(i2);
        ec3.getOWLOneOf().add(i3);

        EnumeratedClass ec4 = OWLFactory.eINSTANCE.createEnumeratedClass();
        ec4.getOWLOneOf().add(i1);

        EnumeratedClass ec5 = OWLFactory.eINSTANCE.createEnumeratedClass();
        ec5.getOWLOneOf().add(i4);
        ec5.getOWLOneOf().add(i5);

        graph.addClass(ec1);
        graph.addClass(ec2);
        graph.addClass(ec3);
        graph.addClass(ec4);
        graph.addClass(ec5);

        graph.show();
    }

    public static void testAxioms(OWLClassClassificationGraph graph) {
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(ReasonerUtils.test);
        r.setLocalName("R");

        OWLClass a = OWLFactory.eINSTANCE.createOWLClass();
        a.setNamespace(ReasonerUtils.test);
        a.setLocalName("A");
        OWLClass b = OWLFactory.eINSTANCE.createOWLClass();
        b.setNamespace(ReasonerUtils.test);
        b.setLocalName("B");
        OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
        c.setNamespace(ReasonerUtils.test);
        c.setLocalName("C");
        OWLClass d = OWLFactory.eINSTANCE.createOWLClass();
        d.setNamespace(ReasonerUtils.test);
        d.setLocalName("D");
        OWLClass e = OWLFactory.eINSTANCE.createOWLClass();
        e.setNamespace(ReasonerUtils.test);
        e.setLocalName("E");
        OWLClass f = OWLFactory.eINSTANCE.createOWLClass();
        f.setNamespace(ReasonerUtils.test);
        f.setLocalName("F");
        OWLClass g = OWLFactory.eINSTANCE.createOWLClass();
        g.setNamespace(ReasonerUtils.test);
        g.setLocalName("G");
        OWLClass h = OWLFactory.eINSTANCE.createOWLClass();
        h.setNamespace(ReasonerUtils.test);
        h.setLocalName("H");

        SomeValuesFromRestriction someRB = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRB.setOWLOnProperty(r);
        someRB.setOWLSomeValuesFrom(b);
        SomeValuesFromRestriction someRE = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRE.setOWLOnProperty(r);
        someRE.setOWLSomeValuesFrom(e);
        SomeValuesFromRestriction someRF = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRF.setOWLOnProperty(r);
        someRF.setOWLSomeValuesFrom(f);
        UnionClass bOrC = OWLFactory.eINSTANCE.createUnionClass();
        bOrC.getOWLUnionOf().add(b);
        bOrC.getOWLUnionOf().add(c);
        SomeValuesFromRestriction someRBOrC = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRBOrC.setOWLOnProperty(r);
        someRBOrC.setOWLSomeValuesFrom(bOrC);

        SomeValuesFromRestriction someRSomeRE = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRSomeRE.setOWLOnProperty(r);
        someRSomeRE.setOWLSomeValuesFrom(someRE);
        SomeValuesFromRestriction someRSomeRF = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRSomeRF.setOWLOnProperty(r);
        someRSomeRF.setOWLSomeValuesFrom(someRF);
        SomeValuesFromRestriction someRSomeRSomeRE = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRSomeRSomeRE.setOWLOnProperty(r);
        someRSomeRSomeRE.setOWLSomeValuesFrom(someRSomeRE);
        SomeValuesFromRestriction someRSomeRSomeRF = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRSomeRSomeRF.setOWLOnProperty(r);
        someRSomeRSomeRF.setOWLSomeValuesFrom(someRSomeRF);

        graph.addAxiom(someRF, someRB);
        graph.addAxiom(someRB, f);
        graph.addAxiom(e, someRBOrC);
        graph.addAxiom(someRBOrC, e);
        graph.addAxiom(g, someRSomeRSomeRE);
        graph.addAxiom(someRSomeRSomeRE, g);
        graph.addAxiom(h, someRSomeRSomeRF);
        graph.addAxiom(someRSomeRSomeRF, h);

        graph.show();
    }

    public static void testCounterExample1(OWLClassClassificationGraph graph) {
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(ReasonerUtils.test);
        r.setLocalName("R");

        OWLClass a = OWLFactory.eINSTANCE.createOWLClass();
        a.setNamespace(ReasonerUtils.test);
        a.setLocalName("A");
        OWLClass b = OWLFactory.eINSTANCE.createOWLClass();
        b.setNamespace(ReasonerUtils.test);
        b.setLocalName("B");
        OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
        c.setNamespace(ReasonerUtils.test);
        c.setLocalName("C");
        OWLClass d = OWLFactory.eINSTANCE.createOWLClass();
        d.setNamespace(ReasonerUtils.test);
        d.setLocalName("D");
        OWLClass e = OWLFactory.eINSTANCE.createOWLClass();
        e.setNamespace(ReasonerUtils.test);
        e.setLocalName("E");
        OWLClass f = OWLFactory.eINSTANCE.createOWLClass();
        f.setNamespace(ReasonerUtils.test);
        f.setLocalName("F");
        OWLClass g = OWLFactory.eINSTANCE.createOWLClass();
        g.setNamespace(ReasonerUtils.test);
        g.setLocalName("G");
        OWLClass h = OWLFactory.eINSTANCE.createOWLClass();
        h.setNamespace(ReasonerUtils.test);
        h.setLocalName("H");

        AllValuesFromRestriction allRA = OWLFactory.eINSTANCE
                .createAllValuesFromRestriction();
        allRA.setOWLOnProperty(r);
        allRA.setOWLAllValuesFrom(a);
        SomeValuesFromRestriction someRB = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRB.setOWLOnProperty(r);
        someRB.setOWLSomeValuesFrom(b);
        IntersectionClass aAndB = OWLFactory.eINSTANCE
                .createIntersectionClass();
        aAndB.getOWLIntersectionOf().add(a);
        aAndB.getOWLIntersectionOf().add(b);
        SomeValuesFromRestriction someRAAndB = OWLFactory.eINSTANCE
                .createSomeValuesFromRestriction();
        someRAAndB.setOWLOnProperty(r);
        someRAAndB.setOWLSomeValuesFrom(aAndB);

        graph.addAxiom(allRA, c);
        graph.addAxiom(someRB, c);
        graph.addClass(someRAAndB);

        graph.show();
    }

    public static void testCounterExample2(OWLClassClassificationGraph graph) {
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(ReasonerUtils.test);
        r.setLocalName("R");

        OWLClass a = OWLFactory.eINSTANCE.createOWLClass();
        a.setNamespace(ReasonerUtils.test);
        a.setLocalName("A");
        OWLClass b = OWLFactory.eINSTANCE.createOWLClass();
        b.setNamespace(ReasonerUtils.test);
        b.setLocalName("B");
        OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
        c.setNamespace(ReasonerUtils.test);
        c.setLocalName("C");
        OWLClass d = OWLFactory.eINSTANCE.createOWLClass();
        d.setNamespace(ReasonerUtils.test);
        d.setLocalName("D");
        OWLClass e = OWLFactory.eINSTANCE.createOWLClass();
        e.setNamespace(ReasonerUtils.test);
        e.setLocalName("E");
        OWLClass f = OWLFactory.eINSTANCE.createOWLClass();
        f.setNamespace(ReasonerUtils.test);
        f.setLocalName("F");
        OWLClass g = OWLFactory.eINSTANCE.createOWLClass();
        g.setNamespace(ReasonerUtils.test);
        g.setLocalName("G");
        OWLClass h = OWLFactory.eINSTANCE.createOWLClass();
        h.setNamespace(ReasonerUtils.test);
        h.setLocalName("H");

        IntersectionClass aAndB = OWLFactory.eINSTANCE
                .createIntersectionClass();
        aAndB.getOWLIntersectionOf().add(a);
        aAndB.getOWLIntersectionOf().add(b);
        IntersectionClass bAndC = OWLFactory.eINSTANCE
                .createIntersectionClass();
        bAndC.getOWLIntersectionOf().add(b);
        bAndC.getOWLIntersectionOf().add(c);
        IntersectionClass cAndA = OWLFactory.eINSTANCE
                .createIntersectionClass();
        cAndA.getOWLIntersectionOf().add(c);
        cAndA.getOWLIntersectionOf().add(a);
        UnionClass dOrEOrF = OWLFactory.eINSTANCE.createUnionClass();

        dOrEOrF.getOWLUnionOf().add(d);
        dOrEOrF.getOWLUnionOf().add(e);
        dOrEOrF.getOWLUnionOf().add(f);
        UnionClass aOrB = OWLFactory.eINSTANCE.createUnionClass();
        aOrB.getOWLUnionOf().add(a);
        aOrB.getOWLUnionOf().add(b);
        UnionClass bOrC = OWLFactory.eINSTANCE.createUnionClass();
        bOrC.getOWLUnionOf().add(b);
        bOrC.getOWLUnionOf().add(c);
        UnionClass cOrA = OWLFactory.eINSTANCE.createUnionClass();
        cOrA.getOWLUnionOf().add(c);
        cOrA.getOWLUnionOf().add(a);

        graph.addAxiom(d, aAndB);
        graph.addAxiom(aAndB, d);
        graph.addAxiom(e, bAndC);
        graph.addAxiom(bAndC, e);
        graph.addAxiom(f, cAndA);
        graph.addAxiom(cAndA, f);
        graph.addClass(dOrEOrF);
        graph.addAxiom(aOrB, g);
        graph.addAxiom(bOrC, g);
        graph.addAxiom(cOrA, g);

        graph.show();
    }

    public static void testAddRandomClass(OWLClassClassificationGraph graph) {
        OWLClass c = ReasonerUtils.createRandomClass(10);
        ReasonerUtils.show(c, 0);
        graph.addClass(c);
        graph.show();
    }

    public static void testAddInputClass(OWLClassClassificationGraph graph)
            throws IOException, ParseException {
        OWLClass c = ReasonerUtils.parseClass(new InputStreamReader(System.in));
        graph.addClass(c);
        graph.show();
    }

    private static class MyNumberFormat {
        public MyNumberFormat(int length) {
            this.length = length;
        }

        public String format(int number) {
            String s = String.valueOf(number);
            if (s.length() < length) {
                StringBuffer sb = new StringBuffer();
                for (int i = s.length(); i < length; i++) {
                    sb.append(' ');
                }
                sb.append(s);
                return sb.toString();
            } else {
                return s;
            }
        }

        private int length;
    }
}
