/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWLStructuralSubsumptionReasoner.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.owl.reasoner.structural;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.eclipse.eodm.owl.AllValuesFromRestriction;
import org.eclipse.eodm.owl.CardinalityRestriction;
import org.eclipse.eodm.owl.ComplementClass;
import org.eclipse.eodm.owl.EnumeratedClass;
import org.eclipse.eodm.owl.HasValueRestriction;
import org.eclipse.eodm.owl.IntersectionClass;
import org.eclipse.eodm.owl.MaxCardinalityRestriction;
import org.eclipse.eodm.owl.MinCardinalityRestriction;
import org.eclipse.eodm.owl.OWLClass;
import org.eclipse.eodm.owl.OWLDataRange;
import org.eclipse.eodm.owl.OWLFactory;
import org.eclipse.eodm.owl.OWLObjectProperty;
import org.eclipse.eodm.owl.OWLOntology;
import org.eclipse.eodm.owl.Property;
import org.eclipse.eodm.owl.SomeValuesFromRestriction;
import org.eclipse.eodm.owl.UnionClass;
import org.eclipse.eodm.owl.reasoner.OWLTaxonomyReasoner;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.RDFSClass;
import org.eclipse.eodm.rdfs.RDFSDatatype;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.vocabulary.RDFS;

public class OWLStructuralSubsumptionReasoner implements OWLTaxonomyReasoner {

    /**
     * Initialize reasoner
     * 
     * @param ontology
     */
    public void initialize(OWLOntology ontology) {
        this.ontology = ontology;

        classes.clear();
        properties.clear();

        navigateOntology(ontology);

        for (Iterator iter = ontology.getOWLImports().iterator(); iter
                .hasNext();) {
            navigateOntology((OWLOntology) iter.next());
        }

        propertyGraph = createPropertyHierarchy(properties);

        classGraph = new OWLClassClassificationGraph(propertyGraph);

        for (Iterator iter = properties.iterator(); iter.hasNext();) {
            addDomainAndRangeAxioms((Property) iter.next());
        }

        // classify every class in ontology
        for (Iterator iter = classes.iterator(); iter.hasNext();) {
            OWLClass clazz = (OWLClass) iter.next();

            // classify owl class when necessary
            classifyOWLClass(clazz);


            // add axioms
            for (Iterator iterC = clazz.getSubClass().iterator(); iterC
                    .hasNext();) {
                classGraph.addAxiom(clazz, (OWLClass) iterC.next());
            }

            // add equivalent classes
            for (Iterator iterC = clazz.getOWLEquivalentClass().iterator(); iterC
                    .hasNext();) {
                OWLClass c = (OWLClass) iterC.next();
                classGraph.addAxiom(clazz, c);
                classGraph.addAxiom(c, clazz);
            }
        

            // add disjointWith
            for (Iterator iterC = clazz.getOWLDisjointWith().iterator(); iterC
                    .hasNext();) {
                IntersectionClass inter = OWLFactory.eINSTANCE
                        .createIntersectionClass();
                inter.getOWLIntersectionOf().add(clazz);
                inter.getOWLIntersectionOf().add(iterC.next());
                classGraph.addAxiom(ReasonerUtils.nothing, inter);
            }            

        }
    }

    /**
     * Get all Descendant Classes of given OWLCLass. owl:Thing, owl:Nothing and
     * anonymous class will not be returned
     * 
     * @param c
     * @return list of Descendant classes
     */
    public List getDescendantClasses(OWLClass c) {
        Vector result = new Vector();

        OWLClassNode v = classGraph.addClass(c);
        List l = classGraph.getSubClassNodes(v);
        for (Iterator iter = l.iterator(); iter.hasNext();) {
            OWLClassNode v1 = (OWLClassNode) iter.next();

            Iterator it = v1.getOWLClassIDSet().iterator();
            while (it.hasNext()) {
                OWLClass clazz = classGraph.getOWLClass(((Integer) it.next())
                        .intValue());
                if (isNamedObject(clazz) && !isBuildInClass(clazz)) {
                    Object obj = getObjectByUri(clazz.getURI());
                    if (obj != null) {
                        result.add(obj);
                    }
                }
            }

        }

        return result;
    }

    /**
     * Get all Ancestor Classes of given OWLCLass. owl:Thing, owl:Nothing and
     * anonymous classes will not be returned
     * 
     * @param c
     * @return list of Ancestor classes
     */
    public List getAncestorClasses(OWLClass c) {
        Vector result = new Vector();

        OWLClassNode v = classGraph.addClass(c);
        List l = classGraph.getSuperClassNodes(v);
        for (Iterator iter = l.iterator(); iter.hasNext();) {
            OWLClassNode v1 = (OWLClassNode) iter.next();

            Iterator it = v1.getOWLClassIDSet().iterator();
            while (it.hasNext()) {
                OWLClass clazz = classGraph.getOWLClass(((Integer) it.next())
                        .intValue());
                if (isNamedObject(clazz) && !isBuildInClass(clazz)) {
                    Object obj = getObjectByUri(clazz.getURI());
                    if (obj != null) {
                        result.add(obj);
                    }
                }
            }

        }

        return result;
    }

    /**
     * Get all Descendant Properties of given property
     * 
     * @param property
     * @return list of Descendant Properties
     */
    public List getDescendantProperties(Property property) {
        List l = propertyGraph.getSubProperties(property);
        Vector result = new Vector();
        for (Iterator iter = l.iterator(); iter.hasNext();) {
            Object p = getObjectByUri((String) iter.next());
            if (p != null) {
                result.add(p);
            }
        }
        return result;
    }

    /**
     * Get all Ancestor Properties of given property
     * 
     * @param property
     * @return list of Ancestor properties
     */
    public List getAncestorProperties(Property property) {
        List l = propertyGraph.getSuperProperties(property);
        Vector result = new Vector();
        for (Iterator iter = l.iterator(); iter.hasNext();) {
            Object p = getObjectByUri((String) iter.next());
            if (p != null) {
                result.add(p);
            }
        }
        return result;
    }

    public boolean isSubClassOf(OWLClass c1, OWLClass c2) {
        OWLClassNode v1 = classGraph.addClass(c1);
        OWLClassNode v2 = classGraph.addClass(c2);
        return classGraph.isSubClassNode(v1, v2);
    }

    public boolean isSuperClassOf(OWLClass c1, OWLClass c2) {
        OWLClassNode v1 = classGraph.addClass(c1);
        OWLClassNode v2 = classGraph.addClass(c2);
        return classGraph.isSubClassNode(v2, v1);
    }

    /**
     * Test Whether p1 is subPropertyOf p2
     * 
     * @param p1
     * @param p2
     * @return true if p1 is subPropertyOf p2
     */
    public boolean isSubPropertyOf(Property p1, Property p2) {
        return propertyGraph.isSubProperty(p1, p2);
    }

    /**
     * Test Whether p1 is superPropertyOf p2
     * 
     * @param p1
     * @param p2
     * @return true if p1 is superPropertyOf p2
     */
    public boolean isSuperPropertyOf(Property p1, Property p2) {
        return propertyGraph.isSubProperty(p2, p1);
    }

    protected void navigateOntology(OWLOntology ontology) {

        // navigating the whole resource list
        for (Iterator iter = ontology.getContains().iterator(); iter.hasNext();) {
            RDFSResource resource = (RDFSResource) iter.next();
            if (resource instanceof OWLClass) {
                classes.add(resource);
            } else if (resource instanceof Property) {
                properties.add(resource);
            }
        }
    }

    protected RDFPropertyGraph createPropertyHierarchy(Vector props) {
        RDFPropertyGraph pGraph = new RDFPropertyGraph();

        for (Iterator iter = props.iterator(); iter.hasNext();) {
            Property p = (Property) iter.next();

            // adding p to obj map
            addObject(p.getURI(), p);

            for (Iterator iterP = p.getRDFSSubPropertyOf().iterator(); iterP
                    .hasNext();) {
                pGraph.addSubPropertyAxiom(p, (Property) iterP.next());
            }
        }

        return pGraph;
    }

    protected void addDomainAndRangeAxioms(Property p) {
        if (p instanceof OWLObjectProperty) {
      
            // clone p to p1
            Property p1 = OWLFactory.eINSTANCE.createProperty();
            p1.setNamespace(p.getNamespace());
            p1.setLocalName(p.getLocalName());

            SomeValuesFromRestriction someRT = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();

            someRT.setOWLOnProperty(p1);
            someRT.setOWLSomeValuesFrom(ReasonerUtils
                    .cloneClass(ReasonerUtils.thing));

            for (Iterator iter = p.getRDFSDomain().iterator(); iter.hasNext();) {
                OWLClass domain = (OWLClass) iter.next();
                if(!domain.getURI().equals(ReasonerUtils.THING)){
                    classGraph.addAxiom(domain, someRT);
                }
            }
            
            for (Iterator iter = p.getRDFSRange().iterator(); iter.hasNext();) {
                OWLClass range = (OWLClass) iter.next();
                if(!range.getURI().equals(ReasonerUtils.THING)) {
                    AllValuesFromRestriction c = OWLFactory.eINSTANCE
                            .createAllValuesFromRestriction();
                    c.setOWLOnProperty(p1);
                    c.setOWLAllValuesFrom(range);
                    classGraph.addAxiom(c, ReasonerUtils
                            .cloneClass(ReasonerUtils.thing));
                }
            }

        }
        

    }

    /**
     * Determine whether an owl class is primitive
     * 
     * @param c
     *            OWLClass
     * @return true if c is primitive
     */
    protected boolean isPrimitiveClass(OWLClass c) {
        return (getOWLClassType(c) == OWL_PRIMITIVE);
    }

    protected void classifyOWLClass(OWLClass clazz) {

        if (isNamedObject(clazz)) {
            // adding name clazz to obj map
            addObject(clazz.getURI(), clazz);
            if (!isPrimitiveClass(clazz)) {
                OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
                c.setNamespace(clazz.getNamespace());
                c.setLocalName(clazz.getLocalName());
                classGraph.addAxiom(c, clazz);
                classGraph.addAxiom(clazz, c);
            }
        }

        if (clazz instanceof HasValueRestriction) {
            RDFSResource val = ((HasValueRestriction) clazz).getOWLHasValue();
            for (Iterator iter = val.getRDFType().iterator(); iter.hasNext();) {
                EnumeratedClass c = OWLFactory.eINSTANCE
                        .createEnumeratedClass();
                c.getOWLOneOf().add(val);
                classGraph.addAxiom((OWLClass) iter.next(), c);
            }
        } else if (clazz instanceof EnumeratedClass) {
            for (Iterator iter = ((EnumeratedClass) clazz).getOWLOneOf()
                    .iterator(); iter.hasNext();) {
                RDFSResource val = (RDFSResource) iter.next();
                for (Iterator iterC = val.getRDFType().iterator(); iterC
                        .hasNext();) {
                    EnumeratedClass c1 = OWLFactory.eINSTANCE
                            .createEnumeratedClass();
                    c1.getOWLOneOf().add(val);

                    OWLClass c2 = (OWLClass) iterC.next();
                    if (c2 != clazz && !isBuildInClass(c2)) {
                        classGraph.addAxiom(c2, c1);
                    }

                }
            }
        }
    }

    /**
     * Determine whether an owl class is a named class
     * 
     * @param c
     * @return true if c is a named class
     */
    protected boolean isNamedObject(RDFSResource c) {
        Namespace namespace = c.getNamespace();
        return !(namespace == null || namespace.getURI().equals(RDFS.BNODE_NAMESPACE));
    }

    /**
     * Determine wheter an owl class is a build-in class. Currently only decide
     * whether it is owl:Thing or owl:Nothing
     * 
     * @param c
     * @return true if c is owl:Thing or owl:Nothing
     */
    protected boolean isBuildInClass(OWLClass c) {
        String uri = c.getURI();
        return (uri.equals(THING) || uri.equals(NOTHING));
    }

    protected OWLClassClassificationGraph classGraph = null;

    protected RDFPropertyGraph propertyGraph = null;

    protected Vector classes = new Vector();

    protected Vector properties = new Vector();

    protected static String NOTHING = "http://www.w3.org/2002/07/owl#Nothing";

    protected static String THING = "http://www.w3.org/2002/07/owl#Thing";

    // owl class type
    protected static final int OWL_PRIMITIVE = -1;

    protected static final int OWL_UNION = 0;

    protected static final int OWL_INTERSECTION = 1;

    protected static final int OWL_SOME_VALUE_FORM = 2;

    protected static final int OWL_ALL_VALUE_FROM = 3;

    protected static final int OWL_HAS_VALUE = 4;

    protected static final int OWL_CARDINARITY = 5;

    protected static final int OWL_MAX_CARDINARITY = 6;

    protected static final int OWL_MIN_CARDINARITY = 7;

    protected static final int OWL_COMPLEMENT = 8;

    protected static final int OWL_ONE_OF = 9;

    protected static final int OWL_DATA_RANGE = 10;

    protected static final int RDFS_DATATYPE = 11;

    protected static Class[] nonPrimitiveClass = { UnionClass.class,
            IntersectionClass.class, SomeValuesFromRestriction.class,
            AllValuesFromRestriction.class, HasValueRestriction.class,
            CardinalityRestriction.class, MaxCardinalityRestriction.class,
            MinCardinalityRestriction.class, ComplementClass.class,
            EnumeratedClass.class, OWLDataRange.class, RDFSDatatype.class };

    protected static int getOWLClassType(RDFSClass c) {
        for (int i = 0; i < nonPrimitiveClass.length; i++) {
            if (nonPrimitiveClass[i].isInstance(c))
                return i;
        }

        return OWL_PRIMITIVE;
    }

    protected OWLOntology ontology = null;

    /**
     * HashMap of URI -> Ontology Objects if it is support by model, it can be
     * abandoned
     */
    protected HashMap objectMap = new HashMap();

    protected Object getObjectByUri(String uri) {
        return objectMap.get(uri);
    }

    protected void addObject(String key, RDFSResource obj) {
        objectMap.put(key, obj);
    }
}
