/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: ReasonerUtils.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.owl.reasoner.structural;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.eclipse.eodm.owl.AllValuesFromRestriction;
import org.eclipse.eodm.owl.EnumeratedClass;
import org.eclipse.eodm.owl.HasValueRestriction;
import org.eclipse.eodm.owl.Individual;
import org.eclipse.eodm.owl.IntersectionClass;
import org.eclipse.eodm.owl.OWLClass;
import org.eclipse.eodm.owl.OWLFactory;
import org.eclipse.eodm.owl.OWLObjectProperty;
import org.eclipse.eodm.owl.OWLRestriction;
import org.eclipse.eodm.owl.SomeValuesFromRestriction;
import org.eclipse.eodm.owl.UnionClass;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.RDFProperty;
import org.eclipse.eodm.rdfs.RDFSFactory;
import org.eclipse.eodm.rdfs.RDFSLiteral;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.TypedLiteral;

/**
 * 
 * @author RobertLu
 * @version $Revision: 1.1.1.1 $
 */

public class ReasonerUtils {

    public static Namespace rdf, rdfs, xsd, owl, test;

    public static OWLClass thing, nothing;

    public static Random random = new Random(System.currentTimeMillis());

    public static final String THING = "http://www.w3.org/2002/07/owl#Thing";
    public static final String NOTHING = "http://www.w3.org/2002/07/owl#Nothing";
    
    static {
        rdf = RDFSFactory.eINSTANCE.createNamespace();
        rdf.setName("rdf");
        rdf.setURI("http://www.w3.org/1999/02/22-rdf-syntax-ns#");
        rdfs = RDFSFactory.eINSTANCE.createNamespace();
        rdfs.setName("rdfs");
        rdfs.setURI("http://www.w3.org/2000/01/rdf-schema#");
        xsd = RDFSFactory.eINSTANCE.createNamespace();
        xsd.setName("xsd");
        xsd.setURI("http://www.w3.org/2001/XMLSchema#");
        owl = RDFSFactory.eINSTANCE.createNamespace();
        owl.setName("owl");
        owl.setURI("http://www.w3.org/2002/07/owl#");
        test = RDFSFactory.eINSTANCE.createNamespace();
        test.setName("test");
        test.setURI("http://test.org/test#");
        thing = OWLFactory.eINSTANCE.createOWLClass();
        thing.setNamespace(owl);
        thing.setLocalName("Thing");
        nothing = OWLFactory.eINSTANCE.createOWLClass();
        nothing.setNamespace(owl);
        nothing.setLocalName("Nothing");
    }

    private static int[][] sum = new int[100][100];

    static {
        for (int i = 0; i < 100; i++) {
            sum[0][i] = 1;
        }
        for (int i = 1; i < 100; i++) {
            for (int j = i; j >= 1; j--) {
                sum[i][j] = sum[i - j][j];
                if (j < i) {
                    sum[i][j] += sum[i][j + 1];
                }
            }
            for (int j = i + 1; j < 100; j++) {
                sum[i][j] = 0;
            }
        }
    }

    /**
     * This method will clone the given class expression recursively and create
     * a copy of the expression. Except for namespace objects, the returned
     * class expression will not share any references with the given class
     * expression.
     * 
     * @param c
     *            The class expression to be cloned.
     * @return The clone of the given class expression.
     */
    public static OWLClass cloneClass(OWLClass c) {
        if (c instanceof IntersectionClass) {
            IntersectionClass c1 = OWLFactory.eINSTANCE
                    .createIntersectionClass();
            Iterator iter = ((IntersectionClass) c).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                c1.getOWLIntersectionOf().add(
                        cloneClass((OWLClass) iter.next()));
            }
            return c1;
        } else if (c instanceof UnionClass) {
            UnionClass c1 = OWLFactory.eINSTANCE.createUnionClass();
            Iterator iter = ((UnionClass) c).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                c1.getOWLUnionOf().add(cloneClass((OWLClass) iter.next()));
            }
            return c1;
        } else if (c instanceof AllValuesFromRestriction) {
            AllValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createAllValuesFromRestriction();
            RDFProperty p = ((AllValuesFromRestriction) c).getOWLOnProperty();
            OWLObjectProperty p1 = OWLFactory.eINSTANCE
                    .createOWLObjectProperty();
            p1.setNamespace(p.getNamespace());
            p1.setLocalName(p.getLocalName());
            c1.setOWLOnProperty(p1);
            c1
                    .setOWLAllValuesFrom(cloneClass((OWLClass) (((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom())));
            return c1;
        } else if (c instanceof SomeValuesFromRestriction) {
            SomeValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            RDFProperty p = ((SomeValuesFromRestriction) c).getOWLOnProperty();
            OWLObjectProperty p1 = OWLFactory.eINSTANCE
                    .createOWLObjectProperty();
            p1.setNamespace(p.getNamespace());
            p1.setLocalName(p.getLocalName());
            c1.setOWLOnProperty(p1);
            c1
                    .setOWLSomeValuesFrom(cloneClass((OWLClass) (((SomeValuesFromRestriction) c)
                            .getOWLSomeValuesFrom())));
            return c1;
        } else {
            OWLClass c1 = OWLFactory.eINSTANCE.createOWLClass();
            c1.setNamespace(c.getNamespace());
            c1.setLocalName(c.getLocalName());
            return c1;
        }
    }

    /**
     * This method transforms the inputted class expression into DNF. If the
     * class expression contains SomeValuesFrom or AllValuesFrom restrictions,
     * the expressions inside the restriction will be transformed too.
     * <p>
     * The inputted class expression will be cloned through the process so the
     * class expression itself will not be modified. The returned class
     * expression will not share any references with the inputted class
     * expression.
     * <p>
     * Note that the basic optimize has already been performed on the result
     * expression.
     * 
     * @param c
     *            The class expression to be transformed to DNF.
     * @param reverseInRestriction
     *            This parameter controls if the expression inside an on
     *            property restriction should be transformed to same format of
     *            this method or not. The recommended value is true.
     * @param toMinimum
     *            If toMinimum is set to true, forward intersection rules and
     *            backward union rules will be applied. Otherwise, backward
     *            intersection rules and forward union rules will be applied.
     * @return The class expression after transform.
     */
    public static OWLClass basicToDNF(RDFPropertyGraph pGraph, OWLClass c,
            boolean reverseInRestriction, boolean toMinimum) {
        if (c instanceof IntersectionClass) {
            BasicClassIdentifier identifier = new BasicClassIdentifier(pGraph);
            List children = new ArrayList();
            Iterator iter = ((IntersectionClass) c).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                OWLClass child = basicToCNF(pGraph, (OWLClass) iter.next(),
                        !reverseInRestriction, toMinimum);
                if (child instanceof IntersectionClass) {
                    Iterator iter1 = ((IntersectionClass) child)
                            .getOWLIntersectionOf().iterator();
                    while (iter1.hasNext()) {
                        OWLClass child1 = (OWLClass) iter1.next();
                        List children1 = new ArrayList();
                        if (child1 instanceof UnionClass) {
                            Iterator iter2 = ((UnionClass) child1)
                                    .getOWLUnionOf().iterator();
                            while (iter2.hasNext()) {
                                children1.add(new Integer(identifier
                                        .getOWLClassID((OWLClass) iter2.next(),
                                                false, true)));
                            }
                        } else {
                            children1.add(new Integer(identifier.getOWLClassID(
                                    child1, false, true)));
                        }
                        children.add(children1);
                    }
                } else if (child instanceof UnionClass) {
                    List children1 = new ArrayList();
                    Iterator iter1 = ((UnionClass) child).getOWLUnionOf()
                            .iterator();
                    while (iter1.hasNext()) {
                        children1.add(new Integer(identifier.getOWLClassID(
                                (OWLClass) iter1.next(), false, true)));
                    }
                    children.add(children1);
                } else {
                    List children1 = new ArrayList();
                    children1.add(new Integer(identifier.getOWLClassID(child,
                            false, true)));
                    children.add(children1);
                }
            }
            List answer = new Transposer(children, identifier.getClassCount())
                    .transpose();
            return optimize(pGraph, identifier.buildUnionClass(answer), 2);
        } else if (c instanceof UnionClass) {
            List children = new ArrayList();
            Iterator iter = ((UnionClass) c).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                OWLClass child = basicToDNF(pGraph, (OWLClass) iter.next(),
                        reverseInRestriction, toMinimum);
                if (child instanceof UnionClass) {
                    Iterator iter1 = ((UnionClass) child).getOWLUnionOf()
                            .iterator();
                    while (iter1.hasNext()) {
                        children.add(iter1.next());
                    }
                } else {
                    children.add(child);
                }
            }
            if (children.size() == 0) {
                return cloneClass(nothing);
            }
            if (children.size() == 1) {
                return (OWLClass) children.get(0);
            }
            UnionClass c1 = OWLFactory.eINSTANCE.createUnionClass();
            c1.getOWLUnionOf().addAll(children);
            return optimize(pGraph, c1, 1);
        } else if (c instanceof AllValuesFromRestriction) {
            AllValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createAllValuesFromRestriction();
            c1.setOWLOnProperty(((AllValuesFromRestriction) c)
                    .getOWLOnProperty());
            if (reverseInRestriction) {
                c1.setOWLAllValuesFrom(toCNF(pGraph,
                        (OWLClass) ((AllValuesFromRestriction) c)
                                .getOWLAllValuesFrom(), toMinimum));
            } else {
                c1.setOWLAllValuesFrom(toDNF(pGraph,
                        (OWLClass) ((AllValuesFromRestriction) c)
                                .getOWLAllValuesFrom(), toMinimum));
            }
            return optimize(pGraph, c1, 1);
        } else if (c instanceof SomeValuesFromRestriction) {
            SomeValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            c1.setOWLOnProperty(((SomeValuesFromRestriction) c)
                    .getOWLOnProperty());
            if (reverseInRestriction) {
                c1.setOWLSomeValuesFrom(toCNF(pGraph,
                        (OWLClass) ((SomeValuesFromRestriction) c)
                                .getOWLSomeValuesFrom(), toMinimum));
            } else {
                c1.setOWLSomeValuesFrom(toDNF(pGraph,
                        (OWLClass) ((SomeValuesFromRestriction) c)
                                .getOWLSomeValuesFrom(), toMinimum));
            }
            return optimize(pGraph, c1, 1);
        } else {
            return cloneClass(c);
        }
    }

    /**
     * This method transforms the inputted class expression into CNF. If the
     * class expression contains SomeValuesFrom or AllValuesFrom restrictions,
     * the expressions inside the restriction will be transformed too.
     * <p>
     * The inputted class expression will be cloned through the process so the
     * class expression itself will not be modified. The returned class
     * expression will not share any references with the inputted class
     * expression.
     * <p>
     * Note that the basic optimize has already been performed on the result
     * expression.
     * 
     * @param c
     *            The class expression to be transformed to CNF.
     * @param reverseInRestriction
     *            This parameter controls if the expression inside an on
     *            property restriction should be transformed to same format of
     *            this method or not. The recommended value is true.
     * @param toMinimum
     *            If toMinimum is set to true, forward intersection rules and
     *            backward union rules will be applied. Otherwise, backward
     *            intersection rules and forward union rules will be applied.
     * @return The class expression after transform.
     */
    public static OWLClass basicToCNF(RDFPropertyGraph pGraph, OWLClass c,
            boolean reverseInRestriction, boolean toMinimum) {
        if (c instanceof UnionClass) {
            BasicClassIdentifier identifier = new BasicClassIdentifier(pGraph);
            List children = new ArrayList();
            Iterator iter = ((UnionClass) c).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                OWLClass child = basicToDNF(pGraph, (OWLClass) iter.next(),
                        !reverseInRestriction, toMinimum);
                if (child instanceof UnionClass) {
                    Iterator iter1 = ((UnionClass) child).getOWLUnionOf()
                            .iterator();
                    while (iter1.hasNext()) {
                        OWLClass child1 = (OWLClass) iter1.next();
                        List children1 = new ArrayList();
                        if (child1 instanceof IntersectionClass) {
                            Iterator iter2 = ((IntersectionClass) child1)
                                    .getOWLIntersectionOf().iterator();
                            while (iter2.hasNext()) {
                                children1.add(new Integer(identifier
                                        .getOWLClassID((OWLClass) iter2.next(),
                                                false, true)));
                            }
                        } else {
                            children1.add(new Integer(identifier.getOWLClassID(
                                    child1, false, true)));
                        }
                        children.add(children1);
                    }
                } else if (child instanceof IntersectionClass) {
                    List children1 = new ArrayList();
                    Iterator iter1 = ((IntersectionClass) child)
                            .getOWLIntersectionOf().iterator();
                    while (iter1.hasNext()) {
                        children1.add(new Integer(identifier.getOWLClassID(
                                (OWLClass) iter1.next(), false, true)));
                    }
                    children.add(children1);
                } else {
                    List children1 = new ArrayList();
                    children1.add(new Integer(identifier.getOWLClassID(child,
                            false, true)));
                    children.add(children1);
                }
            }
            List answer = new Transposer(children, identifier.getClassCount())
                    .transpose();
            return optimize(pGraph, identifier.buildIntersectionClass(answer),
                    2);
        } else if (c instanceof IntersectionClass) {
            List children = new ArrayList();
            Iterator iter = ((IntersectionClass) c).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                OWLClass child = basicToCNF(pGraph, (OWLClass) iter.next(),
                        reverseInRestriction, toMinimum);
                if (child instanceof IntersectionClass) {
                    Iterator iter1 = ((IntersectionClass) child)
                            .getOWLIntersectionOf().iterator();
                    while (iter1.hasNext()) {
                        children.add(iter1.next());
                    }
                } else {
                    children.add(child);
                }
            }
            if (children.size() == 0) {
                return cloneClass(thing);
            }
            if (children.size() == 1) {
                return (OWLClass) children.get(0);
            }
            IntersectionClass c1 = OWLFactory.eINSTANCE
                    .createIntersectionClass();
            c1.getOWLIntersectionOf().addAll(children);
            return optimize(pGraph, c1, 1);
        } else if (c instanceof SomeValuesFromRestriction) {
            SomeValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createSomeValuesFromRestriction();
            c1.setOWLOnProperty(((SomeValuesFromRestriction) c)
                    .getOWLOnProperty());
            if (reverseInRestriction) {
                c1.setOWLSomeValuesFrom(toDNF(pGraph,
                        (OWLClass) ((SomeValuesFromRestriction) c)
                                .getOWLSomeValuesFrom(), toMinimum));
            } else {
                c1.setOWLSomeValuesFrom(toCNF(pGraph,
                        (OWLClass) ((SomeValuesFromRestriction) c)
                                .getOWLSomeValuesFrom(), toMinimum));
            }
            return optimize(pGraph, c1, 1);
        } else if (c instanceof AllValuesFromRestriction) {
            AllValuesFromRestriction c1 = OWLFactory.eINSTANCE
                    .createAllValuesFromRestriction();
            c1.setOWLOnProperty(((AllValuesFromRestriction) c)
                    .getOWLOnProperty());
            if (reverseInRestriction) {
                c1.setOWLAllValuesFrom(toDNF(pGraph,
                        (OWLClass) ((AllValuesFromRestriction) c)
                                .getOWLAllValuesFrom(), toMinimum));
            } else {
                c1.setOWLAllValuesFrom(toCNF(pGraph,
                        (OWLClass) ((AllValuesFromRestriction) c)
                                .getOWLAllValuesFrom(), toMinimum));
            }
            return optimize(pGraph, c1, 1);
        } else {
            return cloneClass(c);
        }
    }

    public static boolean isSubProperty(RDFPropertyGraph pGraph,
            RDFProperty p1, RDFProperty p2) {
        return pGraph.isSubProperty(p1, p2);
    }

    /**
     * This method will calculate if the class expression c1 is a sub class of
     * class expression c2. This method is called basic because the user-defined
     * axioms are not considered in this method. Only the structure information
     * of the expressions are used. Two atomic classes are considered the same
     * only when they have the same URI. The only exception is that the class
     * owl:Thing is the super class of all classes, and the class owl:Nothing is
     * the sub clsas of all the classes.
     * 
     * @param c1
     *            The first class expression to be compared.
     * @param c2
     *            The second class expression to be colmpared
     * @return Returns true if the class c1 is a sub class of class c2. This
     *         includes the possibility that c1 is actually equivalent to c2.
     *         Returns false elsewise.
     */
    public static boolean basicIsSubClass(RDFPropertyGraph pGraph, OWLClass c1,
            OWLClass c2) {
        return basicIsSubClass1(pGraph, c1, c2, new HashMap());
    }

    private static boolean basicIsSubClass1(RDFPropertyGraph pGraph,
            OWLClass c1, OWLClass c2, Map resultCache) {
        boolean flag1 = false, flag2 = false;
        ObjectPair key = new ObjectPair(c1, c2);
        if (resultCache.containsKey(key)) {
            return ((Boolean) resultCache.get(key)).booleanValue();
        }

        if (c1 instanceof UnionClass) {
            Iterator iter = ((UnionClass) c1).getOWLUnionOf().iterator();
            boolean flag = false;
            while (iter.hasNext()) {
                if (!basicIsSubClass1(pGraph, (OWLClass) iter.next(), c2,
                        resultCache)) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                resultCache.put(key, Boolean.TRUE);
                return true;
            }
        } else if (c1 instanceof IntersectionClass) {
            Iterator iter = ((IntersectionClass) c1).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                if (basicIsSubClass1(pGraph, (OWLClass) iter.next(), c2,
                        resultCache)) {
                    resultCache.put(key, Boolean.TRUE);
                    return true;
                }
            }
        } else {
            flag1 = true;
        }

        if (c2 instanceof IntersectionClass) {
            Iterator iter = ((IntersectionClass) c2).getOWLIntersectionOf()
                    .iterator();
            boolean flag = false;
            while (iter.hasNext()) {
                if (!basicIsSubClass1(pGraph, c1, (OWLClass) iter.next(),
                        resultCache)) {
                    flag = true;
                    break;
                }
            }
            if (!flag) {
                resultCache.put(key, Boolean.TRUE);
                return true;
            }
        } else if (c2 instanceof UnionClass) {
            Iterator iter = ((UnionClass) c2).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                if (basicIsSubClass1(pGraph, c1, (OWLClass) iter.next(),
                        resultCache)) {
                    resultCache.put(key, Boolean.TRUE);
                    return true;
                }
            }
        } else {
            flag2 = true;
        }

        if (!flag1 || !flag2) {
            resultCache.put(key, Boolean.FALSE);
            return false;
        }

        if (!(c1 instanceof OWLRestriction)
                && nothing.getURI().equals(c1.getURI())) {
            resultCache.put(key, Boolean.TRUE);
            return true;
        }

        if (!(c2 instanceof OWLRestriction)
                && thing.getURI().equals(c2.getURI())) {
            resultCache.put(key, Boolean.TRUE);
            return true;
        }

        if (c1 instanceof AllValuesFromRestriction
                && c2 instanceof AllValuesFromRestriction) {
            boolean result = isSubProperty(pGraph,
                    ((AllValuesFromRestriction) c2).getOWLOnProperty(),
                    ((AllValuesFromRestriction) c1).getOWLOnProperty())
                    && basicIsSubClass1(pGraph,
                            (OWLClass) ((AllValuesFromRestriction) c1)
                                    .getOWLAllValuesFrom(),
                            (OWLClass) ((AllValuesFromRestriction) c2)
                                    .getOWLAllValuesFrom(), resultCache);
            resultCache.put(key, Boolean.valueOf(result));
            return result;
        } else if (c1 instanceof SomeValuesFromRestriction
                && c2 instanceof SomeValuesFromRestriction) {
            boolean result = isSubProperty(pGraph,
                    ((SomeValuesFromRestriction) c1).getOWLOnProperty(),
                    ((SomeValuesFromRestriction) c2).getOWLOnProperty())
                    && basicIsSubClass1(pGraph,
                            (OWLClass) ((SomeValuesFromRestriction) c1)
                                    .getOWLSomeValuesFrom(),
                            (OWLClass) ((SomeValuesFromRestriction) c2)
                                    .getOWLSomeValuesFrom(), resultCache);
            resultCache.put(key, Boolean.valueOf(result));
            return result;
        } else if (!(c1 instanceof OWLRestriction)
                && !(c1 instanceof OWLRestriction)) {
            String uri1 = (c1.getNamespace() == null ? "" : c1.getNamespace()
                    .getURI())
                    + c1.getLocalName();
            String uri2 = (c2.getNamespace() == null ? "" : c2.getNamespace()
                    .getURI())
                    + c2.getLocalName();
            boolean result = uri1.equals(uri2);
            resultCache.put(key, Boolean.valueOf(result));
            return result;
        } else {
            resultCache.put(key, Boolean.FALSE);
            return false;
        }
    }

    /**
     * This method will calculate if the class expression c1 is equivalent to
     * class expression c2. This method is called basic because the user-defined
     * axioms are not considered in this method. Only the structure information
     * of the expressions are used. Two atomic classes are considered the same
     * only when they have the same URI.
     * 
     * @param c1
     *            The first class expression to be compared.
     * @param c2
     *            The second class expression to be colmpared
     * @return Returns true if the class c1 is equivalent to class c2. Returns
     *         false elsewise.
     */
    public static boolean basicIsEquivalentClass(RDFPropertyGraph pGraph,
            OWLClass c1, OWLClass c2) {
        return basicIsSubClass(pGraph, c1, c2)
                && basicIsSubClass(pGraph, c2, c1);
    }

    /**
     * This method does the following optimize on the given class expression.
     * <p>
     * Suppose the class expression A is a sub class of class expression B,
     * which can be detected by the
     * {@link org.eclipse.eodm.owl.reasoner.ReasonerUtils#basicIsSubClass(OWLClass, OWLClass) <em>basicIsSubClass</em>}
     * method.
     * 
     * <ul>
     * <li>A AND B => A</li>
     * <li>A OR B => B</li>
     * <li>If an intersection/union class expression contains only one child
     * class expression, the class expression is replaced by its only child.
     * </li>
     * <li>If an intersection class expression contains zero child class
     * expression, the class expression is replaced by owl:Thing.</li>
     * <li>If an union class expression contains zero child class expression,
     * the class expression is replaced by owl:Nothing.</li>
     * <li>The class expressions inside a SomeValuesFromRestriction or
     * AllValuesRestriction will be optimized recursively too.</li>
     * </ul>
     * <p>
     * Note that the modifies are directly applied on the given class
     * expression.
     * 
     * @param c
     *            The class expression to be optimized.
     * @param level
     *            The level of recursives where optimize should be applied. If
     *            this value is set to 0, no optimize will be performed. Set
     *            this value to -1 to disable the level restriction.
     * @return The root of the optimized class expression.
     */
    public static OWLClass optimize(RDFPropertyGraph pGraph, OWLClass c,
            int level) {
        if (level == 0) {
            return c;
        }
        if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                l.set(i, optimize(pGraph, (OWLClass) l.get(i), level - 1));
            }
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass c1 = (OWLClass) l.get(i);
                for (int j = i - 1; j >= 0; j--) {
                    OWLClass c2 = (OWLClass) l.get(j);
                    if (basicIsSubClass(pGraph, c1, c2)) {
                        l.remove(j);
                        i--;
                        continue;
                    }
                    if (basicIsSubClass(pGraph, c2, c1)) {
                        l.remove(i);
                        break;
                    }
                }
            }
            if (l.size() == 0) {
                return cloneClass(thing);
            }
            if (l.size() == 1) {
                return (OWLClass) l.get(0);
            }
            return c;
        } else if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                l.set(i, optimize(pGraph, (OWLClass) l.get(i), level - 1));
            }
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass c1 = (OWLClass) l.get(i);
                for (int j = i - 1; j >= 0; j--) {
                    OWLClass c2 = (OWLClass) l.get(j);
                    if (basicIsSubClass(pGraph, c2, c1)) {
                        l.remove(j);
                        i--;
                        continue;
                    }
                    if (basicIsSubClass(pGraph, c1, c2)) {
                        l.remove(i);
                        break;
                    }
                }
            }
            if (l.size() == 0) {
                return cloneClass(nothing);
            }
            if (l.size() == 1) {
                return (OWLClass) l.get(0);
            }
            return c;
        } else if (c instanceof AllValuesFromRestriction) {
            OWLClass c1 = optimize(pGraph,
                    (OWLClass) ((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom(), level - 1);
            //            if (basicIsSubClass(thing, c1)) {
            //                return cloneClass(thing);
            //            }
            ((AllValuesFromRestriction) c).setOWLAllValuesFrom(c1);
            return c;
        } else if (c instanceof SomeValuesFromRestriction) {
            OWLClass c1 = optimize(pGraph,
                    (OWLClass) ((SomeValuesFromRestriction) c)
                            .getOWLSomeValuesFrom(), level - 1);
            //            if (basicIsSubClass(c1, nothing)) {
            //                return cloneClass(nothing);
            //            }
            ((SomeValuesFromRestriction) c).setOWLSomeValuesFrom(c1);
            return c;
        } else {
            return c;
        }
    }

    public static void show(OWLClass c, int indent) {
        if (c instanceof IntersectionClass) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("And {");
            Iterator iter = ((IntersectionClass) c).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                show((OWLClass) iter.next(), indent + 3);
            }
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else if (c instanceof UnionClass) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("Or {");
            Iterator iter = ((UnionClass) c).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                show((OWLClass) iter.next(), indent + 3);
            }
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else if (c instanceof AllValuesFromRestriction) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("For all "
                    + ((AllValuesFromRestriction) c).getOWLOnProperty()
                            .getURI() + " {");
            show((OWLClass) ((AllValuesFromRestriction) c)
                    .getOWLAllValuesFrom(), indent + 3);
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else if (c instanceof SomeValuesFromRestriction) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("For some "
                    + ((SomeValuesFromRestriction) c).getOWLOnProperty()
                            .getURI() + " {");
            show((OWLClass) ((SomeValuesFromRestriction) c)
                    .getOWLSomeValuesFrom(), indent + 3);
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else if (c instanceof HasValueRestriction) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("Has value "
                    + ((HasValueRestriction) c).getOWLOnProperty().getURI()
                    + " {");
            RDFSResource r = ((HasValueRestriction) c).getOWLHasValue();
            if (r instanceof RDFSLiteral) {
                for (int i = 0; i < indent + 3; i++) {
                    System.out.print(' ');
                }
                if (r instanceof TypedLiteral) {
                    System.out.print(((TypedLiteral) r).getDatatype().getURI()
                            + " ");
                }
                System.out.println(((RDFSLiteral) r).getLexicalForm());
            } else {
                for (int i = 0; i < indent + 3; i++) {
                    System.out.print(' ');
                }
                System.out.println(r.getURI());
            }
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else if (c instanceof EnumeratedClass) {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("Enum {");
            Iterator iter = ((EnumeratedClass) c).getOWLOneOf().iterator();
            while (iter.hasNext()) {
                Individual ind = (Individual) iter.next();
                for (int i = 0; i < indent + 3; i++) {
                    System.out.print(' ');
                }
                System.out.println(ind.getURI());
            }
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println("}");
        } else {
            for (int i = 0; i < indent; i++) {
                System.out.print(' ');
            }
            System.out.println(c.getURI());
        }
    }

    /**
     * Appling the following two rules on the given class expresson:
     * <p>
     * Suppose the property R is a sub property of property S
     * <ul>
     * <li>(Some R.A) AND (All S.B) => (Some R.(A AND B)) AND (All S.B)</li>
     * <li>(All R.A) AND (All S.B) => (All R.(A AND B)) AND (All S.B)</li>
     * </ul>
     * <p>
     * This method supposed that the inputted class expression is already in
     * DNF.
     * 
     * @param c
     *            The class expression on which the rules are applied. Note that
     *            the original class expression will be modified directly.
     * @return Returns <b>null </b> if none of the above rules can be applies.
     *         Otherwise, the root of the class expression after applying the
     *         rules is returned.
     */
    public static OWLClass applyForwardIntersectionRules(
            RDFPropertyGraph pGraph, OWLClass c) {
        boolean flag = false;
        if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass result = applyForwardIntersectionRules(pGraph,
                        (OWLClass) l.get(i));
                if (result != null) {
                    l.set(i, result);
                    flag = true;
                }
            }
        } else if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            //            for (int i = l.size() - 1; i >= 0; i--) {
            //                OWLClass result = applyForwardIntersectionRules((OWLClass) l
            //                        .get(i));
            //                if (result != null) {
            //                    l.set(i, result);
            //                    flag = true;
            //                }
            //            }
            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof AllValuesFromRestriction) {
                    AllValuesFromRestriction c1 = (AllValuesFromRestriction) l
                            .get(i);
                    for (int j = l.size() - 1; j >= 0; j--) {
                        if (l.get(j) instanceof SomeValuesFromRestriction) {
                            SomeValuesFromRestriction c2 = (SomeValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && !basicIsSubClass(pGraph, (OWLClass) c2
                                            .getOWLSomeValuesFrom(),
                                            (OWLClass) c1.getOWLAllValuesFrom())) {
                                flag = true;
                                IntersectionClass root;
                                if (c2.getOWLSomeValuesFrom() instanceof IntersectionClass) {
                                    root = (IntersectionClass) c2
                                            .getOWLSomeValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createIntersectionClass();
                                    root.getOWLIntersectionOf().add(
                                            c2.getOWLSomeValuesFrom());
                                    c2.setOWLSomeValuesFrom(root);
                                }
                                if (c1.getOWLAllValuesFrom() instanceof IntersectionClass) {
                                    Iterator iter = ((IntersectionClass) c1
                                            .getOWLAllValuesFrom())
                                            .getOWLIntersectionOf().iterator();
                                    while (iter.hasNext()) {
                                        root.getOWLIntersectionOf().add(
                                                cloneClass((OWLClass) iter
                                                        .next()));
                                    }
                                } else {
                                    root.getOWLIntersectionOf().add(
                                            cloneClass((OWLClass) c1
                                                    .getOWLAllValuesFrom()));
                                }
                            }
                        }
                    }
                }
            }

            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof AllValuesFromRestriction) {
                    AllValuesFromRestriction c1 = (AllValuesFromRestriction) l
                            .get(i);
                    for (int j = i - 1; j >= 0; j--) {
                        if (l.get(j) instanceof AllValuesFromRestriction) {
                            AllValuesFromRestriction c2 = (AllValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && isSubProperty(pGraph, c1
                                            .getOWLOnProperty(), c2
                                            .getOWLOnProperty())) {
                                flag = true;
                                IntersectionClass root;
                                if (c2.getOWLAllValuesFrom() instanceof IntersectionClass) {
                                    root = (IntersectionClass) c2
                                            .getOWLAllValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createIntersectionClass();
                                    root.getOWLIntersectionOf().add(
                                            c2.getOWLAllValuesFrom());
                                    c2.setOWLAllValuesFrom(root);
                                }
                                if (c1.getOWLAllValuesFrom() instanceof IntersectionClass) {
                                    root.getOWLIntersectionOf().addAll(
                                            ((IntersectionClass) c1
                                                    .getOWLAllValuesFrom())
                                                    .getOWLIntersectionOf());
                                    ((IntersectionClass) c1
                                            .getOWLAllValuesFrom())
                                            .getOWLIntersectionOf().clear();
                                } else {
                                    root.getOWLIntersectionOf().add(
                                            c1.getOWLAllValuesFrom());
                                }
                                l.remove(i);
                                break;
                            }
                        }
                    }
                }
            }
            List cloneL = ((IntersectionClass) cloneClass(c))
                    .getOWLIntersectionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof AllValuesFromRestriction) {
                    AllValuesFromRestriction c1 = (AllValuesFromRestriction) cloneL
                            .get(i);
                    for (int j = l.size() - 1; j >= 0; j--) {
                        if (i == j) {
                            continue;
                        }
                        if (l.get(j) instanceof AllValuesFromRestriction) {
                            AllValuesFromRestriction c2 = (AllValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && !basicIsSubClass(pGraph, (OWLClass) c2
                                            .getOWLAllValuesFrom(),
                                            (OWLClass) c1.getOWLAllValuesFrom())) {
                                flag = true;
                                IntersectionClass root;
                                if (c2.getOWLAllValuesFrom() instanceof IntersectionClass) {
                                    root = (IntersectionClass) c2
                                            .getOWLAllValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createIntersectionClass();
                                    root.getOWLIntersectionOf().add(
                                            c2.getOWLAllValuesFrom());
                                    c2.setOWLAllValuesFrom(root);
                                }
                                if (c1.getOWLAllValuesFrom() instanceof IntersectionClass) {
                                    Iterator iter = ((IntersectionClass) c1
                                            .getOWLAllValuesFrom())
                                            .getOWLIntersectionOf().iterator();
                                    while (iter.hasNext()) {
                                        root.getOWLIntersectionOf().add(
                                                cloneClass((OWLClass) iter
                                                        .next()));
                                    }
                                } else {
                                    root.getOWLIntersectionOf().add(
                                            cloneClass((OWLClass) c1
                                                    .getOWLAllValuesFrom()));
                                }
                            }
                        }
                    }
                }
            }
        }
        return (flag ? c : null);
    }

    /**
     * Appling the following two rules on the given class expresson:
     * <p>
     * Suppose the property R is a sub property of property S
     * <ul>
     * <li>(All R.A) OR (Some S.B) => (All R.(A OR B)) OR (Some S.B)</li>
     * <li>(Some R.A) OR (Some S.B) => (Some R.(A OR B)) OR (Some S.B)</li>
     * </ul>
     * <p>
     * This method supposed that the inputted class expression is already in
     * CNF.
     * 
     * @param c
     *            The class expression on which the rules are applied. Note that
     *            the original class expression will be modified directly.
     * @return Returns <b>null </b> if none of the above rules can be applies.
     *         Otherwise, the root of the class expression after applying the
     *         rules is returned.
     */
    public static OWLClass applyForwardUnionRules(RDFPropertyGraph pGraph,
            OWLClass c) {
        boolean flag = false;
        if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass result = applyForwardUnionRules(pGraph, (OWLClass) l
                        .get(i));
                if (result != null) {
                    l.set(i, result);
                    flag = true;
                }
            }
        } else if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            //            for (int i = l.size() - 1; i >= 0; i--) {
            //                OWLClass result = applyForwardUnionRules((OWLClass) l.get(i));
            //                if (result != null) {
            //                    l.set(i, result);
            //                    flag = true;
            //                }
            //            }
            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof SomeValuesFromRestriction) {
                    SomeValuesFromRestriction c1 = (SomeValuesFromRestriction) l
                            .get(i);
                    for (int j = l.size() - 1; j >= 0; j--) {
                        if (l.get(j) instanceof AllValuesFromRestriction) {
                            AllValuesFromRestriction c2 = (AllValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && !basicIsSubClass(pGraph, (OWLClass) c1
                                            .getOWLSomeValuesFrom(),
                                            (OWLClass) c2.getOWLAllValuesFrom())) {
                                flag = true;
                                UnionClass root;
                                if (c2.getOWLAllValuesFrom() instanceof UnionClass) {
                                    root = (UnionClass) c2
                                            .getOWLAllValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createUnionClass();
                                    root.getOWLUnionOf().add(
                                            c2.getOWLAllValuesFrom());
                                    c2.setOWLAllValuesFrom(root);
                                }
                                if (c1.getOWLSomeValuesFrom() instanceof UnionClass) {
                                    Iterator iter = ((UnionClass) c1
                                            .getOWLSomeValuesFrom())
                                            .getOWLUnionOf().iterator();
                                    while (iter.hasNext()) {
                                        root.getOWLUnionOf().add(
                                                cloneClass((OWLClass) iter
                                                        .next()));
                                    }
                                } else {
                                    root.getOWLUnionOf().add(
                                            cloneClass((OWLClass) c1
                                                    .getOWLSomeValuesFrom()));
                                }
                            }
                        }
                    }
                }
            }

            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof SomeValuesFromRestriction) {
                    SomeValuesFromRestriction c1 = (SomeValuesFromRestriction) l
                            .get(i);
                    for (int j = i - 1; j >= 0; j--) {
                        if (l.get(j) instanceof SomeValuesFromRestriction) {
                            SomeValuesFromRestriction c2 = (SomeValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && isSubProperty(pGraph, c1
                                            .getOWLOnProperty(), c2
                                            .getOWLOnProperty())) {
                                flag = true;
                                UnionClass root;
                                if (c2.getOWLSomeValuesFrom() instanceof UnionClass) {
                                    root = (UnionClass) c2
                                            .getOWLSomeValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createUnionClass();
                                    root.getOWLUnionOf().add(
                                            c2.getOWLSomeValuesFrom());
                                    c2.setOWLSomeValuesFrom(root);
                                }
                                if (c1.getOWLSomeValuesFrom() instanceof UnionClass) {
                                    root.getOWLUnionOf().addAll(
                                            ((UnionClass) c1
                                                    .getOWLSomeValuesFrom())
                                                    .getOWLUnionOf());
                                    ((UnionClass) c1.getOWLSomeValuesFrom())
                                            .getOWLUnionOf().clear();
                                } else {
                                    root.getOWLUnionOf().add(
                                            c1.getOWLSomeValuesFrom());
                                }
                                l.remove(i);
                                break;
                            }
                        }
                    }
                }
            }
            List cloneL = ((UnionClass) cloneClass(c)).getOWLUnionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                if (l.get(i) instanceof SomeValuesFromRestriction) {
                    SomeValuesFromRestriction c1 = (SomeValuesFromRestriction) cloneL
                            .get(i);
                    for (int j = l.size() - 1; j >= 0; j--) {
                        if (i == j) {
                            continue;
                        }
                        if (l.get(j) instanceof SomeValuesFromRestriction) {
                            SomeValuesFromRestriction c2 = (SomeValuesFromRestriction) l
                                    .get(j);
                            if (isSubProperty(pGraph, c2.getOWLOnProperty(), c1
                                    .getOWLOnProperty())
                                    && !basicIsSubClass(pGraph, (OWLClass) c1
                                            .getOWLSomeValuesFrom(),
                                            (OWLClass) c2
                                                    .getOWLSomeValuesFrom())) {
                                flag = true;
                                UnionClass root;
                                if (c2.getOWLSomeValuesFrom() instanceof UnionClass) {
                                    root = (UnionClass) c2
                                            .getOWLSomeValuesFrom();
                                } else {
                                    root = OWLFactory.eINSTANCE
                                            .createUnionClass();
                                    root.getOWLUnionOf().add(
                                            c2.getOWLSomeValuesFrom());
                                    c2.setOWLSomeValuesFrom(root);
                                }
                                if (c1.getOWLSomeValuesFrom() instanceof UnionClass) {
                                    Iterator iter = ((UnionClass) c1
                                            .getOWLSomeValuesFrom())
                                            .getOWLUnionOf().iterator();
                                    while (iter.hasNext()) {
                                        root.getOWLUnionOf().add(
                                                cloneClass((OWLClass) iter
                                                        .next()));
                                    }
                                } else {
                                    root.getOWLUnionOf().add(
                                            cloneClass((OWLClass) c1
                                                    .getOWLSomeValuesFrom()));
                                }
                            }
                        }
                    }
                }
            }
        }
        return (flag ? c : null);
    }

    /**
     * Appling the following two rules on the given class expresson:
     * <p>
     * Suppose the property R is a sub property of property S
     * <ul>
     * <li>All R.(A AND B) => (All R.A) AND (All R.B)</li>
     * <li>(Some R.(A AND B)) AND (All S.B) => (Some R.A) AND (All S.B)</li>
     * </ul>
     * <p>
     * This method supposed that the inputted class expression is already in
     * DNF.
     * 
     * @param c
     *            The class expression on which the rules are applied. Note that
     *            the original class expression will be modified directly.
     * @return Returns <b>null </b> if none of the above rules can be applies.
     *         Otherwise, the root of the class expression after applying the
     *         rules is returned.
     */
    public static OWLClass applyBackwardIntersectionRules(
            RDFPropertyGraph pGraph, OWLClass c) {
        boolean flag = false;
        if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass result = applyBackwardIntersectionRules(pGraph,
                        (OWLClass) l.get(i));
                if (result != null) {
                    l.set(i, result);
                    flag = true;
                }
            }
        } else {
            if (c instanceof IntersectionClass) {
                List l = ((IntersectionClass) c).getOWLIntersectionOf();
                for (int i = l.size() - 1; i >= 0; i--) {
                    if (l.get(i) instanceof AllValuesFromRestriction
                            && ((AllValuesFromRestriction) l.get(i))
                                    .getOWLAllValuesFrom() instanceof IntersectionClass) {
                        flag = true;
                        RDFProperty p = ((AllValuesFromRestriction) l.get(i))
                                .getOWLOnProperty();
                        Iterator iter = ((IntersectionClass) ((AllValuesFromRestriction) l
                                .get(i)).getOWLAllValuesFrom())
                                .getOWLIntersectionOf().iterator();
                        while (iter.hasNext()) {
                            AllValuesFromRestriction c1 = OWLFactory.eINSTANCE
                                    .createAllValuesFromRestriction();
                            c1.setOWLOnProperty(p);
                            c1.setOWLAllValuesFrom((OWLClass) iter.next());
                            l.add(c1);
                        }
                        l.remove(i);
                    }
                }
            } else if (c instanceof AllValuesFromRestriction) {
                if (((AllValuesFromRestriction) c).getOWLAllValuesFrom() instanceof IntersectionClass) {
                    flag = true;
                    IntersectionClass c1 = OWLFactory.eINSTANCE
                            .createIntersectionClass();
                    RDFProperty p = ((AllValuesFromRestriction) c)
                            .getOWLOnProperty();
                    Iterator iter = ((IntersectionClass) ((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom()).getOWLIntersectionOf()
                            .iterator();
                    while (iter.hasNext()) {
                        AllValuesFromRestriction c2 = OWLFactory.eINSTANCE
                                .createAllValuesFromRestriction();
                        c2.setOWLOnProperty(p);
                        c2.setOWLAllValuesFrom((OWLClass) iter.next());
                        c1.getOWLIntersectionOf().add(c2);
                    }
                    c = c1;
                }
            }

            if (c instanceof IntersectionClass) {
                List l = ((IntersectionClass) c).getOWLIntersectionOf();
                for (int i = l.size() - 1; i >= 0; i--) {
                    if (l.get(i) instanceof AllValuesFromRestriction) {
                        AllValuesFromRestriction c1 = (AllValuesFromRestriction) l
                                .get(i);
                        for (int j = l.size() - 1; j >= 0; j--) {
                            if (l.get(j) instanceof SomeValuesFromRestriction) {
                                SomeValuesFromRestriction c2 = (SomeValuesFromRestriction) l
                                        .get(j);
                                if (isSubProperty(pGraph,
                                        c2.getOWLOnProperty(), c1
                                                .getOWLOnProperty())) {
                                    OWLClass c3 = getIntersectionQuotient(
                                            pGraph,
                                            (OWLClass) c2
                                                    .getOWLSomeValuesFrom(),
                                            (OWLClass) c1.getOWLAllValuesFrom(),
                                            false);
                                    if (c3 != null) {
                                        flag = true;
                                        c2.setOWLSomeValuesFrom(c3);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return (flag ? c : null);
    }

    /**
     * Appling the following two rules on the given class expresson:
     * <p>
     * Suppose the property R is a sub property of property S
     * <ul>
     * <li>Some R.(A OR B) => (Some R.B) OR (Some R.B)</li>
     * <li>(All R.(A OR B)) OR (Some S.B) => (All R.A) OR (Some S.B)</li>
     * </ul>
     * <p>
     * This method supposed that the inputted class expression is already in
     * CNF.
     * 
     * @param c
     *            The class expression on which the rules are applied. Note that
     *            the original class expression will be modified directly.
     * @return Returns <b>null </b> if none of the above rules can be applies.
     *         Otherwise, the root of the class expression after applying the
     *         rules is returned.
     */
    public static OWLClass applyBackwardUnionRules(RDFPropertyGraph pGraph,
            OWLClass c) {
        boolean flag = false;
        if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            for (int i = l.size() - 1; i >= 0; i--) {
                OWLClass result = applyBackwardUnionRules(pGraph, (OWLClass) l
                        .get(i));
                if (result != null) {
                    l.set(i, result);
                    flag = true;
                }
            }
        } else {
            if (c instanceof UnionClass) {
                List l = ((UnionClass) c).getOWLUnionOf();
                for (int i = l.size() - 1; i >= 0; i--) {
                    if (l.get(i) instanceof SomeValuesFromRestriction
                            && ((SomeValuesFromRestriction) l.get(i))
                                    .getOWLSomeValuesFrom() instanceof UnionClass) {
                        flag = true;
                        RDFProperty p = ((SomeValuesFromRestriction) l.get(i))
                                .getOWLOnProperty();
                        Iterator iter = ((UnionClass) ((SomeValuesFromRestriction) l
                                .get(i)).getOWLSomeValuesFrom())
                                .getOWLUnionOf().iterator();
                        while (iter.hasNext()) {
                            SomeValuesFromRestriction c1 = OWLFactory.eINSTANCE
                                    .createSomeValuesFromRestriction();
                            c1.setOWLOnProperty(p);
                            c1.setOWLSomeValuesFrom((OWLClass) iter.next());
                            l.add(c1);
                        }
                        l.remove(i);
                    }
                }
            } else if (c instanceof SomeValuesFromRestriction) {
                if (((SomeValuesFromRestriction) c).getOWLSomeValuesFrom() instanceof UnionClass) {
                    flag = true;
                    UnionClass c1 = OWLFactory.eINSTANCE.createUnionClass();
                    RDFProperty p = ((SomeValuesFromRestriction) c)
                            .getOWLOnProperty();
                    Iterator iter = ((UnionClass) ((SomeValuesFromRestriction) c)
                            .getOWLSomeValuesFrom()).getOWLUnionOf().iterator();
                    while (iter.hasNext()) {
                        SomeValuesFromRestriction c2 = OWLFactory.eINSTANCE
                                .createSomeValuesFromRestriction();
                        c2.setOWLOnProperty(p);
                        c2.setOWLSomeValuesFrom((OWLClass) iter.next());
                        c1.getOWLUnionOf().add(c2);
                    }
                    c = c1;
                }
            }

            if (c instanceof UnionClass) {
                List l = ((UnionClass) c).getOWLUnionOf();
                for (int i = l.size() - 1; i >= 0; i--) {
                    if (l.get(i) instanceof SomeValuesFromRestriction) {
                        SomeValuesFromRestriction c1 = (SomeValuesFromRestriction) l
                                .get(i);
                        for (int j = l.size() - 1; j >= 0; j--) {
                            if (l.get(j) instanceof AllValuesFromRestriction) {
                                AllValuesFromRestriction c2 = (AllValuesFromRestriction) l
                                        .get(j);
                                if (isSubProperty(pGraph,
                                        c2.getOWLOnProperty(), c1
                                                .getOWLOnProperty())) {
                                    OWLClass c3 = getUnionQuotient(
                                            pGraph,
                                            (OWLClass) c2.getOWLAllValuesFrom(),
                                            (OWLClass) c1
                                                    .getOWLSomeValuesFrom(),
                                            false);
                                    if (c3 != null) {
                                        flag = true;
                                        c2.setOWLAllValuesFrom(c3);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        return (flag ? c : null);
    }

    /**
     * This method returns a class expression c which is the most generic class
     * expression that satisfies (c AND c2) == c1.
     * <p>
     * This method will clone the inputted class expressions first, so they will
     * not be modified. The returned class expression will not share any
     * references with the inputted class expression.
     * 
     * @param c1
     *            The dividend class expression.
     * @param c2
     *            The divisor class expression.
     * @param includeSelf
     *            Specifies if the result can be the dividend itself. If this
     *            parameter is set to false and only possible result is the
     *            dividend itself, null will be returned instead of the normal
     *            result.
     * @return The result class expression. If such a class expression does not
     *         exist, <b>null </b> is returned instead.
     */
    public static OWLClass getIntersectionQuotient(RDFPropertyGraph pGraph,
            OWLClass c1, OWLClass c2, boolean includeSelf) {
        boolean flag = includeSelf;
        c1 = basicToCNF(pGraph, c1, true, true);
        c2 = basicToCNF(pGraph, c2, true, true);

        if (!basicIsSubClass(pGraph, c1, c2)) {
            return null;
        }
        if (basicIsSubClass(pGraph, thing, c1)) {
            if (includeSelf) {
                return cloneClass(thing);
            } else {
                return null;
            }
        }
        if (basicIsSubClass(pGraph, c1, nothing)) {
            if (basicIsSubClass(pGraph, c2, nothing)) {
                return cloneClass(thing);
            } else {
                if (includeSelf) {
                    return cloneClass(nothing);
                } else {
                    return null;
                }
            }
        }
        List cList = new ArrayList();
        if (c1 instanceof IntersectionClass) {
            cList.addAll(((IntersectionClass) c1).getOWLIntersectionOf());
        } else {
            cList.add(c1);
        }
        if (c2 instanceof IntersectionClass) {
            Iterator iter = ((IntersectionClass) c2).getOWLIntersectionOf()
                    .iterator();
            while (iter.hasNext()) {
                OWLClass c3 = (OWLClass) iter.next();
                for (int i = cList.size() - 1; i >= 0; i--) {
                    OWLClass c4 = (OWLClass) cList.get(i);
                    if (basicIsEquivalentClass(pGraph, c3, c4)) {
                        cList.remove(i);
                        flag = true;
                        break;
                    }
                }
            }
        } else {
            for (int i = cList.size() - 1; i >= 0; i--) {
                OWLClass c4 = (OWLClass) cList.get(i);
                if (basicIsEquivalentClass(pGraph, c2, c4)) {
                    cList.remove(i);
                    flag = true;
                    break;
                }
            }
        }
        if (!flag) {
            return null;
        }
        if (cList.size() == 0) {
            return cloneClass(thing);
        }
        if (cList.size() == 1) {
            return (OWLClass) cList.get(0);
        }
        ((IntersectionClass) c1).getOWLIntersectionOf().clear();
        ((IntersectionClass) c1).getOWLIntersectionOf().addAll(cList);
        return c1;
    }

    /**
     * This method returns a class expression c which is the most specific class
     * expression that satisfies (c OR c2) == c1.
     * <p>
     * This method will clone the inputted class expressions first, so they will
     * not be modified. The returned class expression will not share any
     * references with the inputted class expression.
     * 
     * @param c1
     *            The dividend class expression.
     * @param c2
     *            The divisor class expression.
     * @param includeSelf
     *            Specifies if the result can be the dividend itself. If this
     *            parameter is set to false and only possible result is the
     *            dividend itself, null will be returned instead of the normal
     *            result.
     * @return The result class expression. If such a class expression does not
     *         exist, <b>null </b> is returned instead.
     */
    public static OWLClass getUnionQuotient(RDFPropertyGraph pGraph,
            OWLClass c1, OWLClass c2, boolean includeSelf) {
        boolean flag = includeSelf;
        c1 = basicToDNF(pGraph, c1, true, false);
        c2 = basicToDNF(pGraph, c2, true, false);

        if (!basicIsSubClass(pGraph, c2, c1)) {
            return null;
        }
        if (basicIsSubClass(pGraph, c1, nothing)) {
            if (includeSelf) {
                return cloneClass(nothing);
            } else {
                return null;
            }
        }
        if (basicIsSubClass(pGraph, thing, c1)) {
            if (basicIsSubClass(pGraph, thing, c2)) {
                return cloneClass(nothing);
            } else {
                if (includeSelf) {
                    return cloneClass(thing);
                } else {
                    return null;
                }
            }
        }
        List cList = new ArrayList();
        if (c1 instanceof UnionClass) {
            cList.addAll(((UnionClass) c1).getOWLUnionOf());
        } else {
            cList.add(c1);
        }
        if (c2 instanceof UnionClass) {
            Iterator iter = ((UnionClass) c2).getOWLUnionOf().iterator();
            while (iter.hasNext()) {
                OWLClass c3 = (OWLClass) iter.next();
                for (int i = cList.size() - 1; i >= 0; i--) {
                    OWLClass c4 = (OWLClass) cList.get(i);
                    if (basicIsEquivalentClass(pGraph, c3, c4)) {
                        cList.remove(i);
                        flag = true;
                        break;
                    }
                }
            }
        } else {
            for (int i = cList.size() - 1; i >= 0; i--) {
                OWLClass c4 = (OWLClass) cList.get(i);
                if (basicIsEquivalentClass(pGraph, c2, c4)) {
                    cList.remove(i);
                    flag = true;
                    break;
                }
            }
        }
        if (!flag) {
            return null;
        }
        if (cList.size() == 0) {
            return cloneClass(nothing);
        }
        if (cList.size() == 1) {
            return (OWLClass) cList.get(0);
        }
        ((UnionClass) c1).getOWLUnionOf().clear();
        ((UnionClass) c1).getOWLUnionOf().addAll(cList);
        return c1;
    }

    /**
     * This method is supposed to be one of the standard normalization methods.
     * The rules will be applied on the given class expression as many as
     * possible. The final result will be in DNF.
     * <p>
     * Note that, the inputted class expression will be cloned during the
     * process, so the original expression will not be modified.
     * 
     * @param c
     *            The class expression to be normalized.
     * @param toMinimum
     *            If toMinimum is set to true, forward intersection rules and
     *            backward union rules will be applied. Otherwise, backward
     *            intersection rules and forward union rules will be applied.
     * @return The class expression after normalization.
     */
    public static OWLClass toDNF(RDFPropertyGraph pGraph, OWLClass c,
            boolean toMinimum) {
        OWLClass c1;
        boolean flag;
        int timer = 0;

        do {
            timer++;
            if (timer > 4) {
                System.err.println("to DNF("
                        + (toMinimum ? "Minimum" : "Maximum") + "): " + timer);
                show(c, 0);
            }
            flag = false;
            c = basicToCNF(pGraph, c, true, toMinimum);
            c1 = toMinimum ? applyBackwardUnionRules(pGraph, c)
                    : applyForwardUnionRules(pGraph, c);
            if (c1 != null) {
                flag = true;
                c = optimize(pGraph, c1, -1);
            }
            c = basicToDNF(pGraph, c, true, toMinimum);
            c1 = toMinimum ? applyForwardIntersectionRules(pGraph, c)
                    : applyBackwardIntersectionRules(pGraph, c);
            if (c1 != null) {
                flag = true;
                c = optimize(pGraph, c1, -1);
            }
        } while (flag);
        return c;
    }

    /**
     * This method is supposed to be one of the standard normalization methods.
     * The rules will be applied on the given class expression as many as
     * possible. The final result will be in CNF.
     * <p>
     * Note that, the inputted class expression will be cloned during the
     * process, so the original expression will not be modified.
     * 
     * @param c
     *            The class expression to be normalized.
     * @param toMinimum
     *            If toMinimum is set to true, forward intersection rules and
     *            backward union rules will be applied. Otherwise, backward
     *            intersection rules and forward union rules will be applied.
     * @return The class expression after normalization.
     */
    public static OWLClass toCNF(RDFPropertyGraph pGraph, OWLClass c,
            boolean toMinimum) {
        OWLClass c1;
        boolean flag;
        int timer = 0;

        do {
            timer++;
            if (timer > 4) {
                System.err.println("to CNF("
                        + (toMinimum ? "Minimum" : "Maximum") + "): " + timer);
                show(c, 0);
            }
            flag = false;
            c = basicToDNF(pGraph, c, true, toMinimum);
            c1 = toMinimum ? applyForwardIntersectionRules(pGraph, c)
                    : applyBackwardIntersectionRules(pGraph, c);
            if (c1 != null) {
                flag = true;
                c = optimize(pGraph, c1, -1);
            }
            c = basicToCNF(pGraph, c, true, toMinimum);
            c1 = toMinimum ? applyBackwardUnionRules(pGraph, c)
                    : applyForwardUnionRules(pGraph, c);
            if (c1 != null) {
                flag = true;
                c = optimize(pGraph, c1, -1);
            }
        } while (flag);
        return c;
    }

    public static OWLClass toDNF1(RDFPropertyGraph pGraph, OWLClass c,
            boolean toMinimum) {
        OWLClass d;
        do {
            d = c;
            c = toDNF(pGraph, c, !toMinimum);
            c = toDNF(pGraph, c, toMinimum);
        } while (!basicIsEquivalentClass(pGraph, c, d));
        return c;
    }

    public static OWLClass toCNF1(RDFPropertyGraph pGraph, OWLClass c,
            boolean toMinimum) {
        OWLClass d;
        do {
            d = c;
            c = toCNF(pGraph, c, !toMinimum);
            c = toCNF(pGraph, c, toMinimum);
        } while (!basicIsEquivalentClass(pGraph, c, d));
        return c;
    }

    public static String getClassUniqueName(OWLClass c) {
        if (c instanceof IntersectionClass) {
            List l = ((IntersectionClass) c).getOWLIntersectionOf();
            String[] temp = new String[l.size()];
            for (int i = 0; i < l.size(); i++) {
                temp[i] = getClassUniqueName((OWLClass) l.get(i));
            }
            Arrays.sort(temp);
            StringBuffer sb = new StringBuffer();
            sb.append("I {");
            for (int i = 0; i < temp.length; i++) {
                sb.append(" ");
                sb.append(temp[i]);
            }
            sb.append(" }");
            return sb.toString();
        } else if (c instanceof UnionClass) {
            List l = ((UnionClass) c).getOWLUnionOf();
            String[] temp = new String[l.size()];
            for (int i = 0; i < l.size(); i++) {
                temp[i] = getClassUniqueName((OWLClass) l.get(i));
            }
            Arrays.sort(temp);
            StringBuffer sb = new StringBuffer();
            sb.append("U {");
            for (int i = 0; i < temp.length; i++) {
                sb.append(" ");
                sb.append(temp[i]);
            }
            sb.append(" }");
            return sb.toString();
        } else if (c instanceof AllValuesFromRestriction) {
            return "A "
                    + ((AllValuesFromRestriction) c).getOWLOnProperty()
                            .getURI()
                    + " "
                    + getClassUniqueName((OWLClass) ((AllValuesFromRestriction) c)
                            .getOWLAllValuesFrom());
        } else if (c instanceof SomeValuesFromRestriction) {
            return "S "
                    + ((SomeValuesFromRestriction) c).getOWLOnProperty()
                            .getURI()
                    + " "
                    + getClassUniqueName((OWLClass) ((SomeValuesFromRestriction) c)
                            .getOWLSomeValuesFrom());
        } else {
            return "P " + c.getURI();
        }
    }

    public static void showAll(RDFPropertyGraph pGraph, OWLClass c) {
        System.out.println("Original:");
        show(c, 0);
        System.out.println();
        int num = getFormCount();
        for (int i = 0; i < num; i++) {
            System.out.println(getFormName(i) + ":");
            show(getForm(pGraph, c, i), 0);
            System.out.println();
        }
    }

    public static void checkAll(RDFPropertyGraph pGraph, OWLClass c) {
        boolean errorFlag[] = new boolean[4];
        System.out.println("Original:");
        show(c, 0);
        System.out.println();

        OWLClass cc[] = getAllForms(pGraph, c);
        int num = cc.length;
        for (int i = 0; i < num; i++) {
            System.out.println(getFormName(i) + ":");
            show(cc[i], 0);
            System.out.println();
        }

        OWLClass dd[] = new OWLClass[4];
        for (int i = 0; i < num; i++) {
            System.err.println("Checking step " + i);
            for (int j = 0; j < num; j++) {
                dd[j] = getForm(pGraph, cc[j], i);
                errorFlag[j] = false;
            }
            for (int j = 0; j < num; j++) {
                for (int k = j + 1; k < num; k++) {
                    if (!basicIsEquivalentClass(pGraph, dd[j], dd[k])) {
                        System.err.println("Error detected: "
                                + "i = " + i + " j1 = " + j + " j2 = " + k);
                        errorFlag[j] = true;
                        errorFlag[k] = true;
                    }
                }
            }
            for (int j = 0; j < num; j++) {
                if (errorFlag[j]) {
                    System.err.println("Expression " + j + " is:");
                    show(dd[j], 0);
                    System.out.println();
                }
            }
        }
        System.err.println("Checking finished.");
    }

    public static RDFProperty getRandomProperty() {
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(test);
        r.setLocalName(String.valueOf((char) ('R' + random.nextInt(3))));
        return r;
    }

    public static OWLClass createRandomClass(int size) {
        if (size == 1) {
            if (random.nextInt(16) == 0) {
                if (random.nextBoolean()) {
                    return cloneClass(thing);
                } else {
                    return cloneClass(nothing);
                }
            }
            int index = random.nextInt(10);
            OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
            c.setNamespace(test);
            c.setLocalName(String.valueOf((char) ('A' + index)));
            return c;
        }
        if (size == 2 || random.nextInt(10) < 4) {
            if (random.nextBoolean()) {
                AllValuesFromRestriction c = OWLFactory.eINSTANCE
                        .createAllValuesFromRestriction();
                c.setOWLOnProperty(getRandomProperty());
                c.setOWLAllValuesFrom(createRandomClass(size - 1));
                return c;
            } else {
                SomeValuesFromRestriction c = OWLFactory.eINSTANCE
                        .createSomeValuesFromRestriction();
                c.setOWLOnProperty(getRandomProperty());
                c.setOWLSomeValuesFrom(createRandomClass(size - 1));
                return c;
            }
        } else {
            OWLClass c;
            List l;
            if (random.nextBoolean()) {
                c = OWLFactory.eINSTANCE.createIntersectionClass();
                l = ((IntersectionClass) c).getOWLIntersectionOf();
            } else {
                c = OWLFactory.eINSTANCE.createUnionClass();
                l = ((UnionClass) c).getOWLUnionOf();
            }
            size--;
            int r = random.nextInt(sum[size][1] - 1);
            int min = 1;
            while (size > 0) {
                for (int i = min; i <= size; i++) {
                    if (r < sum[size][i] - sum[size][i + 1]) {
                        l.add(createRandomClass(i));
                        size -= i;
                        min = i;
                        break;
                    } else {
                        r -= sum[size][i] - sum[size][i + 1];
                    }
                }
            }
            return c;
        }
    }

    public static void showList(List list) {
        for (int i = 0; i < list.size(); i++) {
            List l = (List) list.get(i);
            for (int j = 0; j < l.size(); j++) {
                System.out.print(l.get(j));
                System.out.print(" ");
            }
            System.out.println();
        }
    }

    private static OWLClass parseClass1(ReaderTokenizer tokenizer,
            boolean inList) throws IOException, ParseException {
        String token = tokenizer.nextToken();
        if ("Or".equals(token)) {
            token = tokenizer.nextToken();
            if (!"{".equals(token)) {
                throw new ParseException((token == null ? "<EOF>" : token)
                        + " encountered.  '{' expected.");
            }
            UnionClass c = OWLFactory.eINSTANCE.createUnionClass();
            for (;;) {
                OWLClass c1 = parseClass1(tokenizer, true);
                if (c1 == null) {
                    break;
                }
                c.getOWLUnionOf().add(c1);
            }
            return c;
        } else if ("And".equals(token)) {
            token = tokenizer.nextToken();
            if (!"{".equals(token)) {
                throw new ParseException((token == null ? "<EOF>" : token)
                        + " encountered.  '{' expected.");
            }
            IntersectionClass c = OWLFactory.eINSTANCE
                    .createIntersectionClass();
            for (;;) {
                OWLClass c1 = parseClass1(tokenizer, true);
                if (c1 == null) {
                    break;
                }
                c.getOWLIntersectionOf().add(c1);
            }
            return c;
        } else if ("For".equals(token)) {
            token = tokenizer.nextToken();
            if ("some".equals(token)) {
                SomeValuesFromRestriction c = OWLFactory.eINSTANCE
                        .createSomeValuesFromRestriction();
                token = tokenizer.nextToken();
                if (token == null || token.indexOf('#') == -1) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "<PROPERTY_URI> expected.");
                }
                int index = token.indexOf('#') + 1;
                String nsURI = token.substring(0, index);
                String localName = token.substring(index);
                Namespace namespace;
                if (nsURI.equals(test.getURI())) {
                    namespace = test;
                } else if (nsURI.equals(rdf.getURI())) {
                    namespace = rdf;
                } else if (nsURI.equals(rdfs.getURI())) {
                    namespace = rdfs;
                } else if (nsURI.equals(owl.getURI())) {
                    namespace = owl;
                } else if (nsURI.equals(xsd.getURI())) {
                    namespace = xsd;
                } else {
                    namespace = RDFSFactory.eINSTANCE.createNamespace();
                    namespace.setName("unknown");
                    namespace.setURI(nsURI);
                }
                RDFProperty p = RDFSFactory.eINSTANCE.createRDFProperty();
                p.setNamespace(namespace);
                p.setLocalName(localName);
                c.setOWLOnProperty(p);
                token = tokenizer.nextToken();
                if (!"{".equals(token)) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "'{' expected.");
                }
                OWLClass c1 = parseClass1(tokenizer, false);
                c.setOWLSomeValuesFrom(c1);
                token = tokenizer.nextToken();
                if (!"}".equals(token)) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "'}' expected.");
                }
                return c;
            } else if ("all".equals(token)) {
                AllValuesFromRestriction c = OWLFactory.eINSTANCE
                        .createAllValuesFromRestriction();
                token = tokenizer.nextToken();
                if (token == null || token.indexOf('#') == -1) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "<PROPERTY_URI> expected.");
                }
                int index = token.indexOf('#') + 1;
                String nsURI = token.substring(0, index);
                String localName = token.substring(index);
                Namespace namespace;
                if (nsURI.equals(test.getURI())) {
                    namespace = test;
                } else if (nsURI.equals(rdf.getURI())) {
                    namespace = rdf;
                } else if (nsURI.equals(rdfs.getURI())) {
                    namespace = rdfs;
                } else if (nsURI.equals(owl.getURI())) {
                    namespace = owl;
                } else if (nsURI.equals(xsd.getURI())) {
                    namespace = xsd;
                } else {
                    namespace = RDFSFactory.eINSTANCE.createNamespace();
                    namespace.setName("unknown");
                    namespace.setURI(nsURI);
                }
                RDFProperty p = RDFSFactory.eINSTANCE.createRDFProperty();
                p.setNamespace(namespace);
                p.setLocalName(localName);
                c.setOWLOnProperty(p);
                token = tokenizer.nextToken();
                if (!"{".equals(token)) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "'{' expected.");
                }
                OWLClass c1 = parseClass1(tokenizer, false);
                c.setOWLAllValuesFrom(c1);
                token = tokenizer.nextToken();
                if (!"}".equals(token)) {
                    throw new ParseException((token == null ? "<EOF>" : token)
                            + " encountered.  " + "'}' expected.");
                }
                return c;
            } else {
                throw new ParseException((token == null ? "<EOF>" : token)
                        + " encountered.  " + "'some', 'all' expected.");
            }
        } else if (token != null && token.indexOf('#') != -1) {
            int index = token.indexOf('#') + 1;
            String nsURI = token.substring(0, index);
            String localName = token.substring(index);
            Namespace namespace;
            if (nsURI.equals(test.getURI())) {
                namespace = test;
            } else if (nsURI.equals(rdf.getURI())) {
                namespace = rdf;
            } else if (nsURI.equals(rdfs.getURI())) {
                namespace = rdfs;
            } else if (nsURI.equals(owl.getURI())) {
                namespace = owl;
            } else if (nsURI.equals(xsd.getURI())) {
                namespace = xsd;
            } else {
                namespace = RDFSFactory.eINSTANCE.createNamespace();
                namespace.setName("unknown");
                namespace.setURI(nsURI);
            }
            OWLClass c = OWLFactory.eINSTANCE.createOWLClass();
            c.setNamespace(namespace);
            c.setLocalName(localName);
            return c;
        } else if ("}".equals(token) && inList) {
            return null;
        } else {
            throw new ParseException((token == null ? "<EOF>" : token)
                    + " encountered.  "
                    + "'Or', 'And', 'For', <CLASS_URI> expected.");
        }
    }

    public static OWLClass parseClass(Reader reader) throws IOException,
            ParseException {
        ReaderTokenizer tokenizer = new ReaderTokenizer(reader);
        return parseClass1(tokenizer, false);
    }

    public static OWLClass[] getAllForms(RDFPropertyGraph pGraph, OWLClass c) {
        OWLClass[] forms = new OWLClass[4];
        forms[0] = toDNF1(pGraph, c, true);
        forms[1] = toCNF1(pGraph, c, true);
        forms[2] = toDNF(pGraph, forms[0], false);
        forms[3] = toCNF(pGraph, forms[1], false);
        //        forms[0] = toDNF(pGraph, c, true);
        //        forms[1] = basicToCNF(pGraph, forms[0], true, true);
        //        forms[2] = toDNF(pGraph, c, false);
        //        forms[3] = basicToCNF(pGraph, forms[2], true, false);
        return forms;
    }

    public static int getFormCount() {
        return 4;
    }

    public static OWLClass getForm(RDFPropertyGraph pGraph, OWLClass c, int form) {
        switch (form) {
        case 0:
            return toDNF1(pGraph, c, true);
        case 1:
            return toCNF1(pGraph, c, true);
        case 2:
            return toDNF1(pGraph, c, false);
        case 3:
            return toCNF1(pGraph, c, false);
        //        case 0:
        //            return toDNF(pGraph, c, true);
        //        case 1:
        //            return toCNF(pGraph, c, true);
        //        case 2:
        //            return toDNF(pGraph, c, false);
        //        case 3:
        //            return toCNF(pGraph, c, false);
        default:
            throw new IllegalArgumentException();
        }
    }

    public static String getFormName(int form) {
        switch (form) {
        case 0:
            return "Minimal DNF";
        case 1:
            return "Minimal CNF";
        case 2:
            return "Maximal DNF";
        case 3:
            return "Maximal CNF";
        default:
            throw new IllegalArgumentException();
        }
    }

    public static void testInputClassNormalization(RDFPropertyGraph pGraph)
            throws ParseException {
        OWLClass c;
        try {
            c = parseClass(new InputStreamReader(System.in));
            showAll(pGraph, c);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static void testInputClassNormalizationCheck(RDFPropertyGraph pGraph)
            throws ParseException {
        OWLClass c;
        try {
            c = parseClass(new InputStreamReader(System.in));
            checkAll(pGraph, c);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public static void testRandomClassNormalization(RDFPropertyGraph pGraph,
            int size) {
        showAll(pGraph, createRandomClass(size));
    }

    public static void testRandomClassNormalizationCheck(
            RDFPropertyGraph pGraph, int size) {
        checkAll(pGraph, createRandomClass(size));
    }

    public static void testTransposer() {
        Integer i0 = new Integer(0);
        Integer i1 = new Integer(1);
        Integer i2 = new Integer(2);
        Integer i3 = new Integer(3);
        Integer i4 = new Integer(4);
        Integer i5 = new Integer(5);

        ArrayList s1 = new ArrayList();
        s1.add(i0);
        s1.add(i1);
        s1.add(i2);
        ArrayList s2 = new ArrayList();
        s2.add(i2);
        s2.add(i3);
        ArrayList s3 = new ArrayList();
        s3.add(i2);
        s3.add(i5);
        s3.add(i4);
        ArrayList s4 = new ArrayList();
        s4.add(i3);
        s4.add(i4);
        s4.add(i5);

        ArrayList ss = new ArrayList();
        ss.add(s1);
        ss.add(s2);
        ss.add(s3);
        ss.add(s4);

        Transposer tt = new Transposer(ss, 6);
        List answer = tt.transpose();
        showList(answer);
    }

    public static void main(String[] args) throws Exception {
        RDFPropertyGraph pGraph = new RDFPropertyGraph();
        RDFProperty r = RDFSFactory.eINSTANCE.createRDFProperty();
        r.setNamespace(test);
        r.setLocalName("R");
        RDFProperty s = RDFSFactory.eINSTANCE.createRDFProperty();
        s.setNamespace(test);
        s.setLocalName("S");
        RDFProperty t = RDFSFactory.eINSTANCE.createRDFProperty();
        t.setNamespace(test);
        t.setLocalName("T");
        pGraph.addSubPropertyAxiom(r, s);
        pGraph.addSubPropertyAxiom(s, t);

        testRandomClassNormalizationCheck(pGraph, 30);
        //        testInputClassNormalizationCheck(pGraph);
        //        OWLClass c = parseClass(new InputStreamReader(System.in));
        //        toDNF(pGraph, c, true);
        //        show(applyForwardIntersectionRules(pGraph, c), 0);
    }
}