/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWLXMLSaverImpl.java,v 1.2 2006/06/29 03:25:19 gxie Exp $
 */

package org.eclipse.eodm.owl.resource;

import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.eclipse.emf.common.util.EList;


import org.eclipse.eodm.rdfs.*;
import org.eclipse.eodm.rdfs.vocabulary.RDF;
import org.eclipse.eodm.rdfs.vocabulary.RDFS;
import org.eclipse.eodm.owl.*;
import org.eclipse.eodm.owl.vocabulary.OWL;

/**
 * OWLXMLSaverImpl
 */
public class OWLXMLSaverImpl {
    public static final String OWL_PREFIX = "owl:";

    public static final String RDF_PREFIX = "rdf:";

    public static final String RDFS_PREFIX = "rdfs:";

    //the output writer
    private Writer writer = null;

    //the ontology to be saved
    private OWLOntology ontology = null;

    public OWLXMLSaverImpl(Writer writer, OWLOntology ontology) {
        super();
        this.writer = writer;
        this.ontology = ontology;
    }

    private static final int PLUS = 1;

    private static final int STAY = 0;

    private static final int REDUCTION = -1;

    private int indentLevel = 1;

    private void writeIndents(int oper) throws IOException {
        if (oper == REDUCTION)
            indentLevel--;
        for (int i = indentLevel; i > 0; i--)
            writer.write("\t");
        if (oper == PLUS)
            indentLevel++;
    }

    /* 1. All unused elements in the ontology */
    private List unusedElements = new Vector();

    /* 2. All elements used for serialization */
    private Map usedElements = new HashMap();

    /**
     * Save all the resources.
     * 
     * @throws IOException
     */
    public void save() throws IOException {
        //replace all invalid characters
        EList listOnto=ontology.getContains();
        int size =ontology.getContains().size();
        for (int i=0; i<size;i++){
            
            if (listOnto.get(i) instanceof RDFSLiteral)
            {
                if (((RDFSLiteral)listOnto.get(i)).getLexicalForm()!=null)
                    ((RDFSLiteral)listOnto.get(i)).setLexicalForm(replaceKeywords(((RDFSLiteral)listOnto.get(i)).getLexicalForm()));
                continue;
            }
            if ( listOnto.get(i) instanceof RDFSResource){
                if (((RDFSResource)listOnto.get(i)).getLocalName()!=null){
                    ((RDFSResource)listOnto.get(i)).setLocalName(replaceKeywords(((RDFSResource)listOnto.get(i)).getLocalName()));
                    ((RDFSResource)listOnto.get(i)).setLocalName(replaceSpace(((RDFSResource)listOnto.get(i)).getLocalName()));
                }
            }
        }


        //1. Ontology Header.
        writeOntoHeader(ontology, false, false);
        //2. All typed elements with URI.
        for (Iterator iterator = ontology.getContains().iterator(); iterator
                .hasNext();) {
            Object object = iterator.next();
            if (object == null)
                System.out.println("null pointer");
            if (!(object instanceof RDFSLiteral)) {
                RDFSResource resource = (RDFSResource) object;
                if (!isAnonymousElement(resource))
                    saveResource(resource);
                else
                    unusedElements.add(resource);
            }
        }
        //3. Elements left unused.
        for (int i = 0, n = unusedElements.size(); i < n; i++) {
            RDFSResource resource = (RDFSResource) unusedElements.get(i);
            if (!usedElements.containsKey(resource.getURI()))
                saveResource(resource);
        }
    }

    /**
     * Save the resource : Class / ObjectProperty / DatatypeProperty /
     * Individual Statement / Container
     * 
     * @param resource
     * @throws IOException
     */
    private void saveResource(RDFSResource resource) throws IOException {
        indentLevel = 1;
        if (resource instanceof OWLClass)
            writeClass((OWLClass) resource, false, false);
        else if (resource instanceof OWLObjectProperty)
            writeObjProperty((OWLObjectProperty) resource, false, false);
        else if (resource instanceof OWLDatatypeProperty)
            writeDatatypeProperty((OWLDatatypeProperty) resource, false, false);
        else if (resource instanceof OWLAnnotationProperty)
            writeAnnoProperty((OWLAnnotationProperty) resource, false, false);
        else if (resource instanceof OWLOntologyProperty)
            writeOntoProperty((OWLOntologyProperty) resource, false, false);
        else if (resource instanceof Individual)
            writeIndividual((Individual) resource, false, false);
        else if (resource instanceof RDFSContainer)
            writeContainer((RDFSContainer) resource, false, false);
        else if (resource instanceof RDFStatement)
            writeStatement((RDFStatement) resource, false, false);
        else if (resource instanceof OWLDataRange)
            writeDataRange((OWLDataRange) resource, false, false);
    }

    /** ************************* write method ********************************* */
    private void writeOntoProperty(OWLOntologyProperty ontoProperty,
            boolean isReferenced, boolean referencedAsNode) throws IOException {
        startElement(ontoProperty, OWL_PREFIX + OWL.ONTOLOGY_PROPERTY, isReferenced, referencedAsNode);
        endElement(OWL_PREFIX + OWL.ONTOLOGY_PROPERTY);
    }

    private void writeAnnoProperty(OWLAnnotationProperty annoProperty,
            boolean isReferenced, boolean referencedAsNode) throws IOException {
        startElement(annoProperty, OWL_PREFIX + OWL.ANNOTATION_PROPERTY, isReferenced, referencedAsNode);
        endElement(OWL_PREFIX + OWL.ANNOTATION_PROPERTY);
    }

    //owl:Ontology
    private void writeOntoHeader(OWLOntology onto, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(onto, OWL_PREFIX + OWL.ONTOLOGY, isReferenced, referencedAsNode);
        if (!isReferenced)
            writeOntoAttrs(onto);
        endElement(OWL_PREFIX + OWL.ONTOLOGY);
    }

    //owl:priorVersion owl:backwardCompatibleWith owl:imcompatibleWith
    //owl:versionInfo owl:imports
    private void writeOntoAttrs(OWLOntology onto) throws IOException {
        if (onto.getOWLPriorVersion() != null) {
            Iterator iter = onto.getOWLPriorVersion().iterator();
            while (iter.hasNext()) {
                writeIndents(PLUS);
                writer.write("<owl:priorVersion>\n");
                writeOntoHeader((OWLOntology) (iter.next()), true, true);
                writeIndents(REDUCTION);
                writer.write("</owl:priorVersion>\n");
            }
        }
        if (onto.getOWLBackwardCompatibleWith() != null) {
            Iterator iter = onto.getOWLBackwardCompatibleWith().iterator();
            while (iter.hasNext()) {
                writeIndents(PLUS);
                writer.write("<owl:backwardCompatibleWith>\n");
                writeOntoHeader((OWLOntology) (iter.next()), true, true);
                writeIndents(REDUCTION);
                writer.write("</owl:backwardCompatibleWith>\n");
            }
        }
        if (onto.getOWLIncompatibleWith() != null) {
            Iterator iter = onto.getOWLIncompatibleWith().iterator();
            while (iter.hasNext()) {
                writeIndents(PLUS);
                writer.write("<owl:incompatibleWith>\n");
                writeOntoHeader((OWLOntology) (iter.next()), true, true);
                writeIndents(REDUCTION);
                writer.write("</owl:incompatibleWith>\n");
            }
        }
        if (onto.getOWLVersionInfo() != null) {
            Iterator iter = onto.getOWLVersionInfo().iterator();
            while (iter.hasNext()) {
                writeAttribute(OWL_PREFIX + OWL.VERSION_INFO,
                        (RDFSResource) iter.next());
            }
        }
        writeAttributes(OWL_PREFIX + OWL.IMPORTS, onto.getOWLImports());
        writeRDFSResourceAttrs(ontology);
    }

    //owl:Class owl:Restriction
    private void writeClass(OWLClass cls, boolean isReferenced, boolean referencedAsNode) throws IOException {
        usedElements.put(cls.getURI(), cls);

        if (isReferenced && !isAnonymousElement(cls)) {
            //if the class is referenced by other resources, the parameter isReferenced should be set to true
            startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
            endElement(OWL_PREFIX + OWL.CLASS);
            return;
        }
        
        //6.OWLRestriction
        if (cls instanceof OWLRestriction) {
            writeRestriction((OWLRestriction) cls, isReferenced, referencedAsNode);
            return;
        }

        //5.EnumeratedClass
        if (cls instanceof EnumeratedClass)
            writeEnumeratedCls((EnumeratedClass) cls, isReferenced, referencedAsNode);
        //4.ComplementClass
        else if (cls instanceof ComplementClass)
            writeComplementCls((ComplementClass) cls, isReferenced, referencedAsNode);
        //3.UnionClass
        else if (cls instanceof UnionClass)
            writeUnionCls((UnionClass) cls, isReferenced, referencedAsNode);
        //2.IntersectionClass
        else if (cls instanceof IntersectionClass)
            writeIntersectionCls((IntersectionClass) cls, isReferenced, referencedAsNode);
        //1.OWLClass with URI
        else {
            startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
            if (!isReferenced || isAnonymousElement(cls))
                writeClsAttrs(cls);
            endElement(OWL_PREFIX + OWL.CLASS);
            //DeprecatedClass
            if ( cls.getDeprecated() != null && cls.getDeprecated().booleanValue() ) {
                startElement(cls, OWL_PREFIX + OWL.DEPRECATED_CLASS, isReferenced, referencedAsNode);
                endElement(OWL_PREFIX + OWL.DEPRECATED_CLASS);
            }
        }
    }

    //owl:equivalentClass owl:disjointWith rdfs:subClassOf
    private void writeClsAttrs(OWLClass cls) throws IOException {
        writeAttributes(OWL_PREFIX + OWL.EQUIVALENT_CLASS, cls
                .getOWLEquivalentClass());
        writeAttributes(OWL_PREFIX + OWL.DISJOINT_WITH, cls
                .getOWLDisjointWith());
        writeAttributes(RDFS_PREFIX + RDFS.P_SUBCLASSOF, cls
                .getRDFSSubClassOf());
        writeRDFSResourceAttrs(cls);
    }

    //owl:Restriction
    private void writeRestriction(OWLRestriction restriction, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        if ( !isReferenced && !isAnonymousElement(restriction) ) {
            startElement(restriction, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
            writeIndents(PLUS);
            writer.write("<owl:equivalentClass>\n");
            writeIndents(PLUS);
            writer.write("<owl:Restriction>\n");
        } else
        {
            writeIndents(PLUS);
        	if ( restriction.getLocalName() == null || restriction.getLocalName().length() == 0 )
                writer.write("<owl:Restriction>\n");
        	else
        	{
        	    if (restriction.getNamespace()!=null && restriction.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE ))
        	    {
        	        writer.write("<owl:Restriction rdf:nodeID=\"" + restriction.getLocalName() + "\">");
        	    }
        	    else if (restriction.getNamespace()==null)
        	        writer.write("<owl:Restriction rdf:nodeID=\"" + restriction.getLocalName() + "\">");
        	    else
        	    {        	        
        	        String abbreviation = null;
        	        boolean isDefined = false;
        	        Namespace ns=null;
        	        Iterator iter = ontology.getOwnedNamespace().iterator();
        	        while (iter.hasNext()) {
        	            ns = (Namespace) iter.next();
        	            if (ns.getURI().equals(restriction.getNamespace().getURI())) {
        	                isDefined = true;
        	                abbreviation = ns.getName();
        	                break;
        	            }
        	        }
        	        
        	        if (isDefined ) {
        	            if (ns.getURI().indexOf("%")==-1)
        	                writer.write( "<owl:Restriction rdf:about=\"&" + abbreviation + ";" + restriction.getLocalName()+"\">");
        	            else
        	                writer.write ("<owl:Restriction rdf:about=\""+restriction.getURI()+"\">");
        	        } 
        	        else
        	            writer.write ("<owl:Restriction rdf:about=\""+restriction.getURI()+"\">");

        	        }        	    
        	}
        	    //writer.write("<owl:Restriction rdf:nodeID=\"" + restriction.getLocalName() + "\">");
        }
        
        //owl:onProperty
        writeAttribute(OWL_PREFIX + OWL.ON_PROPERTY, restriction
                .getOWLOnProperty());
        //owl:allValuesFrom
        if (restriction instanceof AllValuesFromRestriction)
            writeAttribute(OWL_PREFIX + OWL.ALL_VALUES_FROM,
                    ((AllValuesFromRestriction) restriction)
                            .getOWLAllValuesFrom());
        //owl:someValuesFrom
        else if (restriction instanceof SomeValuesFromRestriction)
            writeAttribute(OWL_PREFIX + OWL.SOME_VALUES_FROM,
                    ((SomeValuesFromRestriction) restriction)
                            .getOWLSomeValuesFrom());
        //owl:hasValue
        else if (restriction instanceof HasValueRestriction)
            writeAttribute(OWL_PREFIX + OWL.HAS_VALUE,
                    ((HasValueRestriction) restriction).getOWLHasValue());
        //owl:maxCardinality
        else if (restriction instanceof MaxCardinalityRestriction)
            writeAttribute(OWL_PREFIX + OWL.MAX_CARDINALITY,
                    ((MaxCardinalityRestriction) restriction)
                            .getOWLMaxCardinality());
        //owl:minCardinality
        else if (restriction instanceof MinCardinalityRestriction)
            writeAttribute(OWL_PREFIX + OWL.MIN_CARDINALITY,
                    ((MinCardinalityRestriction) restriction)
                            .getOWLMinCardinality());
        //owl:cardinality
        else if (restriction instanceof CardinalityRestriction)
            writeAttribute(OWL_PREFIX + OWL.CARDINALITY,
                    ((CardinalityRestriction) restriction).getOWLCardinality());
        writeClsAttrs(restriction);
        
        endElement(OWL_PREFIX + OWL.RESTRICTION);

        if ( !isReferenced && !isAnonymousElement(restriction) ) {
            writeIndents(REDUCTION);
            writer.write("</owl:equivalentClass>\n");
            endElement(OWL_PREFIX + OWL.CLASS);
        }
    }

    //owl:oneOf
    private void writeEnumeratedCls(EnumeratedClass cls, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
        EList list = cls.getOWLOneOf();
        writeIndents(PLUS);
        writer.write("<owl:oneOf rdf:parseType=\"Collection\">\n");
        for (int i = 0; i < list.size(); i++)
            writeIndividual((Individual) list.get(i), true, true);
        writeIndents(REDUCTION);
        writer.write("</owl:oneOf>\n");
        writeClsAttrs(cls);
        endElement(OWL_PREFIX + OWL.CLASS);
    }

    //owl:intersectionOf
    private void writeIntersectionCls(IntersectionClass cls, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
        writeIndents(PLUS);
        EList list = cls.getOWLIntersectionOf();
        writer.write("<owl:intersectionOf rdf:parseType=\"Collection\">\n");
        for (int i = 0; i < list.size(); i++)
            writeClass((OWLClass) list.get(i), true, true);
        writeIndents(REDUCTION);
        writer.write("</owl:intersectionOf>\n");
        writeClsAttrs(cls);
        endElement(OWL_PREFIX + OWL.CLASS);
    }

    //owl:unionOf
    private void writeUnionCls(UnionClass cls, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
        writeIndents(PLUS);
        EList list = cls.getOWLUnionOf();
        writer.write("<owl:unionOf rdf:parseType=\"Collection\">\n");
        for (int i = 0; i < list.size(); i++)
            writeClass((OWLClass) list.get(i), true, true);
        writeIndents(REDUCTION);
        writer.write("</owl:unionOf>\n");
        writeClsAttrs(cls);
        endElement(OWL_PREFIX + OWL.CLASS);
    }

    //owl:complementOf
    private void writeComplementCls(ComplementClass cls, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(cls, OWL_PREFIX + OWL.CLASS, isReferenced, referencedAsNode);
        writeAttribute(OWL_PREFIX + OWL.COMPLEMENT_OF, cls.getOWLComplementOf());
        writeClsAttrs(cls);
        endElement(OWL_PREFIX + OWL.CLASS);
    }

    //owl:ObjectProperty
    private void writeObjProperty(OWLObjectProperty property, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(property, OWL_PREFIX + OWL.OBJECT_PROPERTY, isReferenced, referencedAsNode);
        //owl:inverseOf
        if (property.getOWLInverseOf() != null)
            writeAttribute(OWL_PREFIX + OWL.INVERSE_OF, property
                    .getOWLInverseOf());
        //owl:equivalentProperty
        writeAttributes(OWL_PREFIX + OWL.EQUIVALENT_PROPERTY, property
                .getOWLEquivalentProperty());

        if (property.getTransitive() != null
            && property.getTransitive().booleanValue()) {
            writeIndents(STAY);
            writer.write("<rdf:type"
                         + " rdf:resource= \"" + OWL.TRANSITIVE_PROPERTY_URI
                         + "\"/>\n");
        }
        if (property.getFunctional() != null
            && property.getFunctional().booleanValue()) {
            writeIndents(STAY);
            writer.write("<rdf:type"
                         + " rdf:resource= \"" + OWL.FUNCTIONAL_PROPERTY_URI
                         + "\"/>\n");
        }
        if (property.getSymmetric() != null
            && property.getSymmetric().booleanValue()) {
            writeIndents(STAY);
            writer.write("<rdf:type"
                         + " rdf:resource= \"" + OWL.SYMMETRIC_PROPERTY_URI
                         + "\"/>\n");
        }
        if (property.getInverseFunctional() != null
            && property.getInverseFunctional().booleanValue()) {
            writeIndents(STAY);
            writer.write("<rdf:type"
                         + " rdf:resource= \""
                         + OWL.INVERSE_FUNCTIONAL_PROPERTY_URI + "\"/>\n");
        }
        writeRDFPropertyAttrs(property);
        endElement(OWL_PREFIX + OWL.OBJECT_PROPERTY);
        //DeprecatedProperty
        if ( property.getDeprecated() != null && property.getDeprecated().booleanValue() ) {
            startElement(property, OWL_PREFIX + OWL.DEPRECATED_PROPERTY, isReferenced, referencedAsNode);
            endElement(OWL_PREFIX + OWL.DEPRECATED_PROPERTY);
        }
    }

    //owl:DatatypeProperty
    private void writeDatatypeProperty(OWLDatatypeProperty property,
            boolean isReferenced, boolean referencedAsNode) throws IOException {
        //deal with datatype property
        startElement(property, OWL_PREFIX + OWL.DATATYPE_PROPERTY, isReferenced, referencedAsNode);
        //owl:equivalentProperty
        if (property.getOWLEquivalentProperty() != null)
            writeAttributes(OWL_PREFIX + OWL.EQUIVALENT_PROPERTY, property
                    .getOWLEquivalentProperty());
        if (property.getFunctional() != null
            && property.getFunctional().booleanValue()) {
            writeIndents(STAY);
            writer.write("<rdf:type"
                         + " rdf:resource= \"" + OWL.FUNCTIONAL_PROPERTY_URI
                         + "\"/>\n");
        }
        writeRDFPropertyAttrs(property);
        endElement(OWL_PREFIX + OWL.DATATYPE_PROPERTY);
        //DeprecatedProperty
        if ( property.getDeprecated() != null && property.getDeprecated().booleanValue() ) {
            startElement(property, OWL_PREFIX + OWL.DEPRECATED_PROPERTY, isReferenced, referencedAsNode);
            endElement(OWL_PREFIX + OWL.DEPRECATED_PROPERTY);
        }
    }

    //rdfs:subPropertyOf rdfs:range rdfs:domain
    private void writeRDFPropertyAttrs(RDFProperty property) throws IOException {
        writeAttributes(RDFS_PREFIX + RDFS.P_SUBPROPERTYOF, property
                .getRDFSSubPropertyOf());
        writeAttributes(RDFS_PREFIX + RDFS.P_DOMAIN, property.getRDFSDomain());
        writeAttributes(RDFS_PREFIX + RDFS.P_RANGE, property.getRDFSRange());
        writeRDFSResourceAttrs(property);
    }

    //rdfs:isDefinedBy rdfs:seeAlso rdf:value rdf:type rdfs:comment rdf:li
    // rdfs:label
    private void writeRDFSResourceAttrs(RDFSResource resource)
            throws IOException {
        writeAttributes(RDFS_PREFIX + RDFS.ISDEFINEDBY, resource
                .getRDFSIsDefinedBy());
        writeAttributes(RDFS_PREFIX + RDFS.SEEALSO, resource.getRDFSSeeAlso());
        writeAttributes(RDF_PREFIX + RDF.P_VALUE, resource.getRDFValue());
        writeAttributes(RDFS_PREFIX + RDFS.P_COMMENT, resource.getRDFSComment());
        writeAttributes(RDF_PREFIX + RDF.S_LI, resource.getRDFSMember());
        writeAttributes(RDFS_PREFIX + RDFS.P_LABEL, resource.getRDFSLabel());
    }

    //owl:Thing
    private void writeIndividual(Individual inv, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        String type = null;
        String uri = null;
        for (int i = 0; i < inv.getRDFType().size(); i++) {
            RDFSClass cls = (RDFSClass) inv.getRDFType().get(i);
            if (!cls.getURI().equals(OWL.THING_URI)) {
                type = cls.getNamespace().getName() + ":" + cls.getLocalName();
                uri = cls.getURI();
                break;
            }
        }
        if (type == null) {
            type = OWL_PREFIX + OWL.THING;
            uri = OWL.THING_URI;
        }

        startElement(inv, type, isReferenced, referencedAsNode);
        if (!isReferenced) {
            //owl:differentFrom
            writeAttributes(OWL_PREFIX + OWL.DIFFERENT_FROM, inv
                    .getOWLDifferentFrom());
            //owl:sameAs
            writeAttributes(OWL_PREFIX + OWL.SAME_AS, inv.getOWLSameAs());
            for (int i = inv.getObjectSlot().size() - 1; i >= 0; i--) {
                ObjectSlot slot = (ObjectSlot) inv.getObjectSlot().get(i);
                OWLObjectProperty property = slot.getProperty();
                writeAttributes(property.getNamespace().getName()
                                + ":" + property.getLocalName(), slot
                        .getContent());
            }
            for (int i = inv.getDatatypeSlot().size() - 1; i >= 0; i--) {
                DatatypeSlot slot = (DatatypeSlot) inv.getDatatypeSlot().get(i);
                OWLDatatypeProperty property = (OWLDatatypeProperty) slot
                        .getProperty();
                writeAttributes(property.getNamespace().getName()
                                + ":" + property.getLocalName(), slot
                        .getContent());
            }
            for (int i = 0; i < inv.getRDFType().size(); i++) {
                RDFSClass cls = (RDFSClass) inv.getRDFType().get(i);
                if (cls.getURI().equals(OWL.THING_URI)
                    || cls.getURI().equals(uri))
                    continue;
                writeAttribute(RDF_PREFIX + RDF.P_TYPE, cls);
            }
            writeRDFSResourceAttrs(inv);
        }
        endElement(type);
    }

    //owl:DataRange owl:oneOf
    private void writeDataRange(OWLDataRange range, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        usedElements.put(range.getURI(), range);
        startElement(range, OWL_PREFIX + OWL.DATA_RANGE, isReferenced, referencedAsNode);
        writeIndents(PLUS);
        writer.write("<owl:oneOf>\n");
        writeRDFList(range.getOWLOneOf());
        writeIndents(REDUCTION);
        writer.write("</owl:oneOf>\n");
        writeAttributes(RDFS_PREFIX + RDFS.P_SUBCLASSOF, range
                .getRDFSSubClassOf());
        writeRDFSResourceAttrs(range);
        endElement(OWL_PREFIX + OWL.DATA_RANGE);
    }

    //rdf:List rdf:first rdf:rest
    private void writeRDFList(EList list) throws IOException {
        for (int i = 0; i < list.size(); i++) {
            writeIndents(PLUS);
            writer.write("<rdf:List>\n");
            writeAttribute("rdf:first", (RDFSLiteral) list.get(i));
            if (i == list.size() - 1) {
            	writeIndents(STAY);
                writer.write("<rdf:rest rdf:resource=\"&rdf;nil\"/>\n");
            } else {
                writeIndents(PLUS);
                writer.write("<rdf:rest>\n");
            }
        }
        for (int i = list.size()-1; i >= 0; i--) {
        	if (i < list.size() - 1) {
	            writeIndents(REDUCTION);
	            writer.write("</rdf:rest>\n");
        	}
            writeIndents(REDUCTION);
            writer.write("</rdf:List>\n");
        }
    }

    //rdf:Statement
    private void writeStatement(RDFStatement statement, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        startElement(statement, RDF_PREFIX + RDF.C_STATEMENT, false, referencedAsNode);
        //rdf:subject
        writeAttribute(RDF_PREFIX + RDF.P_SUBJECT, statement.getRDFSubject());
        //rdf:predicate
        writeAttribute(RDF_PREFIX + RDF.P_PREDICATE, statement
                .getRDFPredicate());
        //rdf:object
        writeAttribute(RDF_PREFIX + RDF.P_OBJECT, statement.getRDFObject());
        writeRDFSResourceAttrs(statement);
        endElement(RDF_PREFIX + RDF.C_STATEMENT);
    }

    private void writeContainer(RDFSContainer container, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        //rdf:Alt
        if (container instanceof RDFAlt) {
            startElement(container, RDF_PREFIX + RDF.C_ALT, isReferenced, referencedAsNode);
            writeRDFSResourceAttrs(container);
            endElement(RDF_PREFIX + RDF.C_ALT);
            return;
        }
        //rdf:Bag
        if (container instanceof RDFBag) {
            startElement(container, RDF_PREFIX + RDF.C_BAG, isReferenced, referencedAsNode);
            writeRDFSResourceAttrs(container);
            endElement(RDF_PREFIX + RDF.C_BAG);
            return;
        }
        //rdf:Seq
        if (container instanceof RDFSeq) {
            startElement(container, RDF_PREFIX + RDF.C_SEQ, isReferenced, referencedAsNode);
            writeRDFSResourceAttrs(container);
            endElement(RDF_PREFIX + RDF.C_SEQ);
            return;
        }
    }

    /** ************************************************************************* */

    private void startElement(RDFSResource resource, String tag, boolean isReferenced, boolean referencedAsNode)
            throws IOException {
        writeIndents(PLUS);
        writer.write("<" + tag + " " + getIDAttr(resource, isReferenced, referencedAsNode) + ">\n");
    }

    private void endElement(String tag) throws IOException {
        writeIndents(REDUCTION);
        writer.write("</" + tag + ">\n");
    }

    private boolean isAnonymousElement(RDFSResource resource) {
        if (resource instanceof RDFSLiteral)
            return false;
        if (resource instanceof OWLRestriction)
        	return true;
        if (resource instanceof OWLClass && (resource.getLocalName()==null || resource.getLocalName().length()==0))
            return true;
        return resource.getNamespace() != null
               && resource.getNamespace().getURI() != null
               && resource.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE);
    }

    private String getIDAttr(RDFSResource resource, boolean isReferenced, boolean referencedAsNode) {
              
        if (isAnonymousElement(resource))
            if ( resource.getLocalName() == null || resource.getLocalName().length() == 0 )
        		return "";
        	else
        		return "rdf:nodeID=\"" + resource.getLocalName() + "\"";
        if (resource instanceof OWLClass || resource instanceof OWLDataRange){
            if ( resource.getLocalName() == null || resource.getLocalName().length() == 0 )
        		return "";
        }
        if (resource instanceof OWLOntology) {
            if (!isReferenced)
                return "rdf:about=\"" + resource.getURI() + "\"";
            else if ( referencedAsNode )
                return "rdf:about=\"" + resource.getURI() + "\"";
            else
            	return "rdf:resource=\"" + resource.getURI() + "\"";
        }
        String key = isReferenced ? ( referencedAsNode ? "rdf:about" : "rdf:resource" ) : "rdf:about";
        String value = null;
        String abbreviation = null;
        boolean isDefined = false;
        Namespace ns=null;
        Iterator iter = ontology.getOwnedNamespace().iterator();
        while (iter.hasNext()) {
            ns = (Namespace) iter.next();
            if (ns.getURI().equals(resource.getNamespace().getURI())) {
                isDefined = true;
                abbreviation = ns.getName();
                break;
            }
        }
        if (isDefined ) {
            if (ns.getURI().indexOf("%")==-1)
                value = "&" + abbreviation + ";" + resource.getLocalName();
            else
                value = resource.getURI();
        } else
            value = resource.getURI();
        return key + "=\"" + value + "\"";
    }
    
    private void writeAttribute(String name, RDFSResource value)
            throws IOException {
        if (value instanceof RDFSLiteral) {
            writeIndents(STAY);
            if (value instanceof RDFXMLLiteral)
                writer.write("<"
                             + name
                             + " rdf:parseType=\"Literal\">"
                             + ((RDFSLiteral) value)
                                     .getLexicalForm());
            else if (value instanceof TypedLiteral)
                writer.write("<"
                             + name
                             + " rdf:datatype=\""
                             + ((TypedLiteral) value).getDatatype().getURI()
                             + "\">"
                             + ((RDFSLiteral) value)
                                     .getLexicalForm());
            else if (value instanceof PlainLiteral) {
                String lang = ((PlainLiteral) value).getLanguage();
                if (lang != null && lang.length() > 0)
                    writer.write("<"
                                 + name
                                 + " xml:lang=\""
                                 + lang
                                 + "\">"
                                 + ((RDFSLiteral) value)
                                         .getLexicalForm());
                else
                    writer.write("<"
                                 + name
                                 + ">"
                                 + ((RDFSLiteral) value)
                                         .getLexicalForm());
            } else
                writer.write("<"
                             + name
                             + ">"
                             + ((RDFSLiteral) value)
                                     .getLexicalForm());
            writer.write("</" + name + ">\n");
            return;
        }

        if (!((value instanceof OWLClass || value instanceof OWLDataRange)&& (value.getLocalName() ==null || value.getLocalName().length() ==0)))
	        if (!isAnonymousElement(value)|| isAnonymousElement(value) && usedElements.containsKey(value.getURI())) {
	            writeIndents(STAY);
	            writer.write("<" + name + " " + getIDAttr(value, true, false) + "/>\n");
	            return;
	        }

        writeIndents(PLUS);
        writer.write("<" + name + ">\n");
        if (value instanceof OWLClass)
            writeClass((OWLClass) value, true, true);
        else if (value instanceof OWLObjectProperty)
            writeObjProperty((OWLObjectProperty) value, true, true);
        else if (value instanceof OWLDatatypeProperty)
            writeDatatypeProperty((OWLDatatypeProperty) value, true, true);
        else if (value instanceof Individual)
            writeIndividual((Individual) value, true, true);
        else if (value instanceof RDFStatement)
            writeStatement((RDFStatement) value, true, true);
        else if (value instanceof RDFSContainer)
            writeContainer((RDFSContainer) value, true, true);
        else if (value instanceof OWLDataRange)
            writeDataRange((OWLDataRange) value, true, true);
        writeIndents(REDUCTION);
        writer.write("</" + name + ">\n");
    }

    private void writeAttributes(String name, EList values) throws IOException {
        for (Iterator iterator = values.iterator(); iterator.hasNext();)
            writeAttribute(name, (RDFSResource) iterator.next());
    }

    private String replaceKeywords(String str) {
        str = str.replaceAll("&", "&amp;");
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("'", "&apos;");
        str = str.replaceAll("\"", "&quot;");
        return str;
    }
    
    private String replaceSpace(String str){
        return str.replaceAll(" ", "%20");
    }
    
}