/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWLOntoModel.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.owl.resource.parser;

import java.util.Map;

import org.eclipse.eodm.owl.Individual;
import org.eclipse.eodm.owl.OWLAllDifferent;
import org.eclipse.eodm.owl.OWLAnnotationProperty;
import org.eclipse.eodm.owl.OWLClass;
import org.eclipse.eodm.owl.OWLDataRange;
import org.eclipse.eodm.owl.OWLDatatypeProperty;
import org.eclipse.eodm.owl.OWLObjectProperty;
import org.eclipse.eodm.owl.OWLOntology;
import org.eclipse.eodm.owl.OWLOntologyProperty;
import org.eclipse.eodm.owl.OWLRestriction;
import org.eclipse.eodm.owl.resource.parser.exception.OWLParserException;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.RDFSContainer;
import org.eclipse.eodm.rdfs.RDFSDatatype;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.RDFStatement;

/**
 * OWLOntoModel - OWL ontology model
 * <p>
 * Another memory storage manager of ODM in {URI - Object} pairs.
 */
public interface OWLOntoModel {
    /**
     * Get the included ontology uri. The URI is thought as the uique key of the
     * ontology model.
     * 
     * @return the ontology uri
     */
    String getModelURI();

    /**
     * Whether the model has been parsed completely.
     * 
     * @return true if it has been parsed
     */
    boolean isParsed();

    /**
     * Set the model to be parsed.
     *  
     */
    void setParsed(boolean parsed);

    /**
     * Set the OWL Ontology of this Ontology Model.
     * 
     * @param ontology
     *            the OWL Ontology instance
     */
    void setOntology(OWLOntology ontology);

    /**
     * Get the OWL Ontology of this Ontology Model.
     * 
     * @return the OWL Ontology
     */
    OWLOntology getOntology();

    /**
     * Add another ontology.
     * 
     * @param otherOnto
     *            other ontology
     */
    void addOtherOnto(OWLOntology otherOnto);

    /**
     * Get another ontology.
     * 
     * @param uri
     *            the ontology uri
     * @return the other ontology
     */
    OWLOntology getOtherOnto(String uri);

    /**
     * Add a Namespace.
     * 
     * @param namespace
     *            the Namespace
     */
    void addNamespace(Namespace namespace);

    /**
     * Get a Namspace defined in the Ontology Document.
     * 
     * @param uri
     *            the Namespace URI
     * @return the Namespace
     */
    Namespace getNamespace(String uri);

    /**
     * Get all the Namespaces.
     * 
     * @return all the Namespaces
     */
    Map getNamespaces();

    /**
     * Add an imported Model. owl:import ensures the current Ontology Model can
     * refer to those elements defined in the imported Ontology Models.
     * 
     * @param model
     *            the imported Ontology Model
     */
    void addImpModel(OWLOntoModel model);

    /**
     * Get an imported Ontology Model by its URI.
     * 
     * @param uri
     *            the imported Ontology Model URI
     * @return the imported Ontology Model
     */
    OWLOntoModel getImpModel(String uri);

    /**
     * Get all the imported Ontology Models
     * 
     * @return all the imported Ontology Models
     */
    Map getImpModels();

    /**
     * Add an OWLClass to the Ontology Model. The class can be one of such types
     * as follows: 1) Class identifier 2) Class expression, e.g. EnumeratedClass /
     * UnionClass / IntersectionClass / ComplementClass.
     * 
     * @param cls
     *            the OWLClass to be added
     */
    void addClass(OWLClass cls) throws OWLParserException;

    /**
     * Get an OWLClass by URI.
     * 
     * @param uri
     *            the Class URI
     * @return the OWLClass
     */
    OWLClass getClass(String uri);

    /**
     * Get all the OWL Classes.
     * 
     * @return all the Classes
     */
    Map getClasses();

    /**
     * Add an OWLRestriction to the Ontology Model. An OWLRestriction should be
     * one of the following concrete type: HasValueRestriction /
     * AllValuesFromRestriction / SomeValuesFromRestriction /
     * CardinalityRestriction / MaxCardinalityRestriction /
     * MinCardinalityRestriction
     * 
     * @param restriction
     *            the restriction to be added
     */
    void addRestriction(OWLRestriction restriction) throws OWLParserException;

    /**
     * Get an OWLRestriction by URI.
     * 
     * @param uri
     *            the Restriction URI
     * @return the OWLRestriction
     */
    OWLRestriction getRestriction(String uri);

    /**
     * Get all the Restrictions.
     * 
     * @return all the restrictions
     */
    Map getRestrictions();

    /**
     * Add an ObjectProperty to the Ontology Model.
     * 
     * @param objectProperty
     *            the ObjectProperty to be added
     */
    void addObjectProperty(OWLObjectProperty objectProperty)
            throws OWLParserException;

    /**
     * Get an OWL ObjectProperty by URI.
     * 
     * @param uri
     *            the ObjectProperty URI
     * @return the ObjectProperty
     */
    OWLObjectProperty getObjectProperty(String uri);

    /**
     * Get all the OWL Object Properties.
     * 
     * @return all the Object Properties
     */
    Map getObjectProperties();

    /**
     * Add a DatatypeProperty to the OntologyModel.
     * 
     * @param datatypeProperty
     *            the DatatypeProperty to be added
     */
    void addDatatypeProperty(OWLDatatypeProperty datatypeProperty)
            throws OWLParserException;

    /**
     * Get an OWL DatatypeProperty by URI.
     * 
     * @param uri
     *            the DatatypeProperty URI
     * @return the DatatypeProperty
     */
    OWLDatatypeProperty getDatatypeProperty(String uri);

    /**
     * Get all the OWL Datatype Properties.
     * 
     * @return all the Datatype Properties
     */
    Map getDatatypeProperties();

    /**
     * Add an Individual to the Ontology Model.
     * 
     * @param individual
     *            the individual to be added
     */
    void addIndividual(Individual individual) throws OWLParserException;

    /**
     * Get an Individual by URI.
     * 
     * @param uri
     *            the Individual URI
     * @return the Individual
     */
    Individual getIndividual(String uri);

    /**
     * Get all the OWL Individuals.
     * 
     * @return all the Individuals
     */
    Map getIndividuals();

    /**
     * Add a DataRange to the Ontology Model.
     * 
     * @param dataRange
     *            the DataRange to be added.
     */
    void addDataRange(OWLDataRange dataRange) throws OWLParserException;

    /**
     * Get a DataRange by URI.
     * 
     * @param uri
     *            the DataRange URI
     * @return the DataRange
     */
    OWLDataRange getDataRange(String uri);

    /**
     * Get all the OWL Data Ranges.
     * 
     * @return all the Data Ranges
     */
    Map getDataRanges();

    /**
     * owl:AllDifferent contains some individuals which are different from each
     * other.
     * 
     * @param allDifferent
     *            the owl:AllDifferent object
     */
    void addAllDifferent(String id, OWLAllDifferent allDifferent)
            throws OWLParserException;

    /**
     * Get the owl:AllDifferent object by id.
     * 
     * @param id
     *            the owl:AllDifferent object id
     * @return the owl:AllDifferent object
     */
    OWLAllDifferent getAllDifferent(String id);

    /**
     * Get all owl:AllDifferents.
     * 
     * @return all owl:AllDifferents.
     */
    Map getAllDifferents();

    /**
     * Add a statment to the model.
     * 
     * @param statment
     *            the statment to be added
     */
    void addStatement(RDFStatement statment);

    /**
     * Get a statement by uri.
     * 
     * @param uri
     *            the statement uri
     * @return the statement
     */
    RDFStatement getStatement(String uri);

    /**
     * Get all the statements in the model
     * 
     * @return all the statements
     */
    Map getStatements();

    /**
     * Add a datatype to the model.
     * 
     * @param datatype
     *            the datatype to be added.
     */
    void addDatatype(RDFSDatatype datatype);

    /**
     * Get a datatype by uri.
     * 
     * @param uri
     *            the datatype uri
     * @return the datatype
     */
    RDFSDatatype getDatatype(String uri);

    /**
     * Get all the datatypes.
     * 
     * @return all the datatypes
     */
    Map getDatatypes();

    /**
     * Add an annotation property
     * @param annoProperty
     */
    void addAnnoProperty(OWLAnnotationProperty annoProperty);

    /**
     * Get an annotation property by uri
     * @param uri the input uri
     * @return an annotation property with the specified uri
     */
    OWLAnnotationProperty getAnnoProperty(String uri);

    /**
     * Get all annotation properties
     * @return
     */
    Map getAnnoProperties();

    /**
     * Add an ontology property
     * @param ontoProperty
     */
    void addOntoProperty(OWLOntologyProperty ontoProperty);

    /**
     * Get an ontology property by uri
     * @param uri the input uri
     * @return an ontology property with the specified uri
     */
    OWLOntologyProperty getOntoProperty(String uri);

    /**
     * Get all ontology properties
     * @return
     */
    Map getOntoProperties();

    /**
     * Add a container to the model.
     * 
     * @param container
     *            the container to be added
     */
    void addContainer(RDFSContainer container);

    /**
     * Get a container by uri.
     * 
     * @param uri
     *            the container uri
     * @return the container
     */
    RDFSContainer getContainer(String uri);

    /**
     * Get all the containers.
     * 
     * @return all the containers
     */
    Map getContainers();

    /**
     * Get a resource with uri.
     * 
     * @param uri
     *            the resource uri
     * @return the resource
     */
    RDFSResource getResource(String uri);

    /**
     * Clear all the resources.
     * <p>
     * To clear all the resources is necessary(for preformance consideration)
     * when removing an ontology model from the mapping repository.
     */
    void clearResources();
    
    /**
     * Add a RDFSresource
     * @param resource
     */
    void addRDFSResource(RDFSResource resource);

    /**
     * Get a RDFSResource by URI
     * @param uri
     * @return resources with the specified URI
     */
    RDFSResource getRDFSResource(String uri);

    /**
     * Get all RDFSResources
     * @return
     */
    Map getRDFSResource();
}