/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWLParser.java,v 1.2 2006/07/19 10:09:19 gxie Exp $
 */

package org.eclipse.eodm.owl.resource.parser;

import java.util.List;

import org.eclipse.eodm.owl.OWLOntology;
import org.eclipse.eodm.owl.resource.parser.exception.OWLParserException;

/**
 * OWLParser - OWL Semantics and Abstract Syntax Parser.
 * <p>
 * Parsing OWL in RDF/XML into ODM (Ontology Definition Meta-Model). Semantics
 * and Abstract Syntax is described at <a
 * href="http://www.w3.org/TR/owl-semantics/">OWL S&AS </a>.
 *  
 */
public interface OWLParser {
    /**
     * Add an OWL Document to the OWL Parser, which will be parsed into an OWL
     * Ontology by calling {@link #parseOWLDocument(OWLDocument) parse}method.
     * <p>
     * If the document does not exist, the parser will pre-parse it and add the
     * resulting ontology model to the ontology mapping repository. Or the
     * parser throws an OWLParserException if the document already exists.
     * <p>
     * The Pre-parsing step is to identify all typed elements explicitly defined
     * in the document, all implicitly typed elements and axioms are not parsed
     * in the current step. Later, calling method
     * {@link #parseOWLDocument(OWLDocument) parse}will parse all the left
     * triples.
     * 
     * @param document
     *            the OWL Document to be added
     * @throws OWLParserException
     *             if the OWL Document already exists or the parsing fails
     */
    void addOWLDocument(OWLDocument document) throws OWLParserException;

    /**
     * This method will accept pre-existing ontology and URIs provided by users
     * 
     * @param document
     *            the OWL Document to be added
     * @param listOnto
     *            a list of named classes and named properties
     * @param listURIs
     *            a list of ontology URIs
     * @throws OWLParserException
     */
    void addOWLDocument(OWLDocument document, List listOnto, List listURIs)
            throws OWLParserException;

    /**
     * Add a local file to the OWL Parser.
     * 
     * @see #addOWLDocument(OWLDocument)
     * @param localFileName
     *            the local file name
     * @throws OWLParserException
     *             if the OWL Document already exists or the parsing fails
     */
    void addOWLDocument(String localFileName) throws OWLParserException;

    /**
     * Parse OWL Document in RDF/XML into an OWL Ontology Definition Meta-Model.
     * If the Document is not in the mapping repository, then add it. Or just
     * handle implicitly typed elements and axioms of the pre-parsed ontology.
     * 
     * @param document
     *            the OWL Document as the input
     * @return the OWL Ontology object
     * @throws OWLParserException
     *             if the OWL parsing fails
     */
    OWLOntology parseOWLDocument(OWLDocument document)
            throws OWLParserException;

    /**
     * Parse a local file into an OWL Ontology.
     * 
     * @see #parseOWLDocument(OWLDocument)
     * @param localFileName
     *            the local file name
     * @return the OWL Ontology object
     * @throws OWLParserException
     *             if the OWL parsing fails
     */
    OWLOntology parseOWLDocument(String localFileName)
            throws OWLParserException;

    /**
     * Clear all OWL Documents and OWL Ontologies.
     *  
     */
    void clear();
}