/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 * 
 * $Id: OWL2ECoreTransformer.java,v 1.2 2006/06/22 05:08:16 gxie Exp $
 */
package org.eclipse.eodm.owl.transformer.ecore;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAnnotation;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EEnumLiteral;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EcoreFactory;
import org.eclipse.emf.ecore.EcorePackage;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.EcoreResourceFactoryImpl;
import org.eclipse.eodm.owl.CardinalityRestriction;
import org.eclipse.eodm.owl.EnumeratedClass;
import org.eclipse.eodm.owl.HasValueRestriction;
import org.eclipse.eodm.owl.MaxCardinalityRestriction;
import org.eclipse.eodm.owl.MinCardinalityRestriction;
import org.eclipse.eodm.owl.OWLClass;
import org.eclipse.eodm.owl.OWLDataRange;
import org.eclipse.eodm.owl.OWLDatatypeProperty;
import org.eclipse.eodm.owl.OWLFactory;
import org.eclipse.eodm.owl.OWLObjectProperty;
import org.eclipse.eodm.owl.OWLOntology;
import org.eclipse.eodm.owl.OWLRestriction;
import org.eclipse.eodm.owl.reasoner.OWLTaxonomyReasoner;
import org.eclipse.eodm.owl.reasoner.structural.StructuralReasonerFactory;
import org.eclipse.eodm.owl.resource.OWLXMLResourceFactoryImpl;
import org.eclipse.eodm.owl.transformer.OWLTransformationException;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.RDFProperty;
import org.eclipse.eodm.rdfs.RDFSDatatype;
import org.eclipse.eodm.rdfs.RDFSLiteral;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.vocabulary.RDF;
import org.eclipse.eodm.rdfs.vocabulary.RDFS;
import org.eclipse.eodm.rdfs.vocabulary.XSD;


public class OWL2ECoreTransformer {

	// input owl file
	private String inputOWLFile = null;

	// output ecore file
	private String outputEcoreFile = null;

	//  owl factory
	private OWLFactory factory = null;

	//  owl ontology
	private OWLOntology ontology = null;
	
	// owl ontology temp
	//private OWLOntology ontology_temp = null;

	//  ecore factory
	private EcoreFactory ecoreFactory = null;

	// ecorepackage
	private EcorePackage ecorePackage = null;

	// ecore package
	private EPackage epackage = null;
	
	// reasoner
	private OWLTaxonomyReasoner  reasoner = null;
    
	// record log for transformer
	private BufferedWriter bw = null;
	
	private StringBuffer sb = null;

    public OWL2ECoreTransformer()
    {
        initialize(null);
    }
    
	public OWL2ECoreTransformer(String logFile) {
		// initialization
		initialize(logFile);
	}

	/*
	 * initialization
	 */
	private void initialize(String logFile) {
		// create factory
		factory = OWLFactory.eINSTANCE;
		// create ontology
		ontology = factory.createOWLOntology();
		// init ecore package
		ecoreFactory = EcoreFactory.eINSTANCE;
		ecorePackage = EcorePackage.eINSTANCE;
		epackage = ecoreFactory.createEPackage();
		
        // init reasoner
		reasoner = StructuralReasonerFactory.instance().createOWLTaxonmyReasoner();
		
        // init buffered writer
        if ( logFile != null ) 
        {
    		try {
    		    // TODO
                bw = new BufferedWriter(new FileWriter(logFile));
            } catch (IOException e) {
                e.printStackTrace();
            }
        } else
        {
            bw = null ;
        }
        
        sb = new StringBuffer();
	}

	/**
	 * Translate function Translate the input owl file to target ecore file.
	 * 
	 * @param input
	 *            owl file
	 * @throws OWLTransformationException
	 */
	public void convertOWL2Ecore(String input) throws OWLTransformationException {
		convertOWL2Ecore(input, null);
	}

	/**
	 * Translate function Translate the input owl file to target ecore file.
	 * 
	 * @param input
	 *            owl file
	 * @param output
	 *            target ecore file
	 * @throws OWLTransformationException
	 */
	public void convertOWL2Ecore(String input, String output)
			throws OWLTransformationException {

		// check files
		checkFiles(input, output);
		
		// load OWL model
		loadOWLModel();

        // create Ecore model
        createEcoreModel();
        
		// merge the imported ontology to this ontology
		mergeOnto();
		
        // preprocess ontology
		preprocessOntology(ontology);
		
		// convert ontology to ecore
		owl2Ecore(ontology);

		// save ecore model
		saveEPackage();
		
		// logging
		log(sb.toString());

	}
	
	/*
	 * convert owl to ecore
	 * 
	 * @param ontology
	 * @throws OWLTransformationException
	 */
	private void owl2Ecore(OWLOntology ontology) throws OWLTransformationException {
	    
	    reasoner.initialize(ontology);
	    
		// OWL -> Ecore

		// OWLClass -> EClass
		List classList = ontology.getResourceList(OWLClass.class);
		Iterator ri = classList.iterator();

		while (ri.hasNext()) {
			OWLClass c = (OWLClass) ri.next();

			// filter out Restriction anonymous owlclass
			if (c.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
				continue;
			
            // fitler out owl:thing
            if(c.getURI().equals(OWL_THING) && c.getSubClass().isEmpty() ) {
                continue;
            }
            
			EClass eclass = ecoreFactory.createEClass();
			eclass.setName(c.getLocalName());
			//System.out.println(c.getLocalName());
			createAnnotation(ecoreFactory, eclass, c);
			
			// log
			sb.append("OWL Class " + eclass.getName() + " transformed to EClass successfully!\n");
			
			// record disjointWith
			Iterator iter_dw = c.getOWLDisjointWith().iterator();
			while (iter_dw.hasNext()) {
			    OWLClass owlc = (OWLClass) iter_dw.next();
			    if (owlc != null) {
			        EAnnotation ea = ecoreFactory.createEAnnotation();
			        ea.setSource("The " + c.getLocalName() + " class is disjoint with " + owlc.getLocalName() + ".");
			        eclass.getEAnnotations().add(ea);
			        
			        // log
			        sb.append("  The " + c.getLocalName() + " class is disjoint with " + owlc.getLocalName() + "! OWLDisjointWith transformed to EAnnotation successfully!\n");
			    }
			}
			
			epackage.getEClassifiers().add(eclass);
			
		}
		
		// log
		sb.append("\n");
		
		// OWLClass.RDFSSubClassOf -> EClass.eSuperType
		ri = classList.iterator();


		while (ri.hasNext()) {
			OWLClass c = (OWLClass) ri.next();
			EClass cclass = (EClass) epackage.getEClassifier(c.getLocalName());

			if (cclass == null)
				continue;
			
			//System.out.println("===============================");
			//System.out.println("**** " + cclass.getName() + " ****");

			Iterator ssc = c.getRDFSSubClassOf().iterator();
			// from reasoner to get the ancestor classes of c
			//Iterator ssc = reasoner.getAncestorClasses(c).iterator();
			while (ssc.hasNext()) {
				OWLClass superClass = (OWLClass) ssc.next();
				
				// ?
		        // filter out Restriction anonymous owlclass
				if (superClass.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
					continue;
				
				if (!superClass.getLocalName().equals(c.getLocalName())
				        && !containSameName(cclass, superClass)) {
				    //System.out.println(superClass.getLocalName());
				    EClass ecs = (EClass) epackage.getEClassifier(superClass.getLocalName());
				    // TODO
				    // process inter-subclassof
				    // now the current method is to cut half subclassof from inter-subclassof to avoid deadlock
				    if (!checkInterSubClassOf(cclass, ecs)) {
				        cclass.getESuperTypes().add(ecs);
				        
				        // log
				        sb.append("RDFSSubClassOf(" + cclass.getName() + ", " + ecs.getName() + ") transformed to eSuperTypes successfully!\n");
				    }
				    // log
				    else {
				        ecs.getESuperTypes().remove(cclass);
				        
				        EAnnotation ea1 = ecoreFactory.createEAnnotation();
				        ea1.setSource("The " + cclass.getName() + " class is equivalent with " + ecs.getName() + ".");
				        cclass.getEAnnotations().add(ea1);
				        EAnnotation ea2 = ecoreFactory.createEAnnotation();
				        ea2.setSource("The " + ecs.getName() + " class is equivalent with " + cclass.getName() + ".");
				        ecs.getEAnnotations().add(ea2);
				        
				        // log
				        sb.append("OWLEquivalentClass( " + cclass.getName() + ", " + ecs.getName() + ") transformed to EAnnotation successfully!\n");
				    }
				}
			}
			//System.out.println("-------------------------------");
		}
		
		// log
		sb.append("\n");

		//RDFSDatatype -> EDataType, but XSD datatypes will not be transformed
		ri = ontology.getResourceList(RDFSDatatype.class).iterator();
		HashMap dataTypeBags = new HashMap();
		while (ri.hasNext()) {
			RDFSDatatype dt = (RDFSDatatype) (ri.next());
			if (dt.getURI() == null)
				continue;
			// skip XSD datatypes and Ecore datatypes
			if (dt.getURI().equals(XSD.NAMESPACE + XSD.sxboolean)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxfloat)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxbyte)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxint)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxlong)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxdouble)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxshort)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sxstring)
					|| dt.getURI().equals(XSD.NAMESPACE + XSD.sinteger)
					|| dt.getURI().indexOf(XSD.ECORE_NAMESPACE) > -1)
				continue;

			//			 if (dt.getLocalName().startsWith("Bag_")){
			//				 boolean stop = true;
			//			 }
			if (EnumeratedClass.class.isInstance((dt.getRDFSIsDefinedBy()
					.size() > 0 ? dt.getRDFSIsDefinedBy().get(0) : null))) {

				// OWLEnumeratedClass -> EEnum
				dataTypeBags.put((EnumeratedClass) dt.getRDFSIsDefinedBy().get(
						0), createEEnum((EnumeratedClass) dt
						.getRDFSIsDefinedBy().get(0)));
				
				// log
				// TODO

			} else {
				// Create a user-defined EDatatType
				EDataType ed = ecoreFactory.createEDataType();
				ed.setName(dt.getLocalName());
				createAnnotation(ecoreFactory, ed, dt);
				epackage.getEClassifiers().add(ed);
				
				// log
				sb.append("RDFSDatatype " + ed.getName() + " transformed to EDatatype successfully!\n");
			}
		}
		
		// log
		sb.append("\n");
		
		// OWL EnumeratedClass -> EEnum
		List enumclass = ontology.getResourceList(EnumeratedClass.class);
		// remove datatype bags
		enumclass.removeAll(dataTypeBags.keySet());

		ri = enumclass.iterator();
		while (ri.hasNext()) {
			createEEnum((EnumeratedClass) ri.next());
		}
		
		// log
		sb.append("\n");

		// OWLDatatypeProperty -> EAttribute; OWLObjectProperty -> EReference
		ri = classList.iterator();
		while (ri.hasNext()) {
			OWLClass oc = (OWLClass) ri.next();
			EClass eclass = null;
			if (!EDataType.class.isInstance(epackage.getEClassifier(oc
					.getLocalName())))
				eclass = (EClass) epackage.getEClassifier(oc.getLocalName());
			if (eclass == null)
				continue;

			Iterator pi = oc.getOwnedProperty().iterator();
			while (pi.hasNext()) {
			    RDFProperty p = (RDFProperty) (pi.next());
				RDFSResource ran = null;
				
				if (p instanceof OWLDatatypeProperty) {
				    // OWLDatatypeProperty -> EAttribute
				    OWLDatatypeProperty dp = (OWLDatatypeProperty) p;
				    
				    if (dp.getRDFSRange().iterator().hasNext() == false) {
				        // if property without range, use String for default
				        EAttribute attribute = ecoreFactory.createEAttribute();
						createAnnotation(ecoreFactory, attribute, dp);
						
						// log
						sb.append("OWLDatatypeProperty " + dp.getLocalName() + " transformed to EAttribute successfully! Default type is EString!\n");
						
						// record subPropertyOf to its annoation
						Iterator iter_spo = dp.getRDFSSubPropertyOf().iterator();
						while (iter_spo.hasNext()) {
						    RDFProperty rdfp = (RDFProperty) iter_spo.next();
						    if (rdfp != null) {
						        EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is subPropertyOf " + rdfp.getLocalName() + ".");
						        attribute.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  subPropertyOf(" + dp.getLocalName() + ", " + rdfp.getLocalName() + ") transformed to EAnnotation successfully!\n");
						    }
						}
						
						// record Functional to its annoation if dp.getFunctional() is true
						Boolean b = dp.getFunctional();
						if (b != null && b.booleanValue() == true) {
					        EAnnotation ea = ecoreFactory.createEAnnotation();
					        ea.setSource("The property is Functional.");
					        attribute.getEAnnotations().add(ea);
					        
					        // log
					        sb.append("  " + dp.getLocalName() + " is functional! Functional transformed to EAnnotation successfully!\n");
					    }
						
						attribute.setName(dp.getLocalName());
						attribute.setEType(ecorePackage.getEString());
						eclass.getEStructuralFeatures().add(attribute);
						
				    } else {
				        // property has ranges
				        // use only the first range
				        // TODO
				        ran = (RDFSResource) dp.getRDFSRange().iterator().next();
				        
				        if (ran instanceof RDFSDatatype) {
				            // RDFSDatatype
				            EAttribute attribute = ecoreFactory.createEAttribute();
							createAnnotation(ecoreFactory, attribute, dp);
							
							// log
							sb.append("OWLDatatypeProperty " + dp.getLocalName() + " transformed to EAttribute successfully!\n");
							
							// record subPropertyOf to its annoation
							Iterator iter_spo = dp.getRDFSSubPropertyOf().iterator();
							while (iter_spo.hasNext()) {
							    RDFProperty rdfp = (RDFProperty) iter_spo.next();
							    if (rdfp != null) {
							        EAnnotation ea = ecoreFactory.createEAnnotation();
							        ea.setSource("The property is subPropertyOf " + rdfp.getLocalName() + ".");
							        attribute.getEAnnotations().add(ea);
							        
							        // log
							        sb.append("  subPropertyOf(" + dp.getLocalName() + ", " + rdfp.getLocalName() + ") transformed to EAnnotation successfully!\n");
							    }
							}
							
							// record Functional to its annoation if dp.getFunctional() is true
							Boolean b = dp.getFunctional();
							if (b != null && b.booleanValue() == true) {
						        EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is Functional.");
						        attribute.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + dp.getLocalName() + " is functional! Functional transformed to EAnnotation successfully!\n");
						    }
							
							attribute.setName(dp.getLocalName());
							
							if (ran.getURI() != null) {
								//9 kind mapping: XSD<->EMF
								if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxboolean))
									attribute.setEType(ecorePackage.getEBoolean());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxfloat))
									attribute.setEType(ecorePackage.getEFloat());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxbyte))
									attribute.setEType(ecorePackage.getEByte());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxint))
									attribute.setEType(ecorePackage.getEInt());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxlong))
									attribute.setEType(ecorePackage.getELong());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxdouble))
									attribute.setEType(ecorePackage.getEDouble());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxshort))
									attribute.setEType(ecorePackage.getEShort());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sinteger))
									attribute.setEType(ecorePackage
											.getEIntegerObject());
								else if (ran.getURI().equals(
										XSD.NAMESPACE + XSD.sxstring)
										|| ran.getLocalName().endsWith(
												RDFS.C_LITERAL)
										| ran.getLocalName().endsWith(
												RDF.C_XMLLITERAL))
									attribute.setEType(ecorePackage.getEString());
								else {
									// user-defined datatype
									if (EnumeratedClass.class.isInstance((ran
											.getRDFSIsDefinedBy().size() > 0 ? ran
											.getRDFSIsDefinedBy().get(0) : null))) {
										// Set EEnum datatype
										attribute.setEType((EEnum) dataTypeBags
												.get(ran.getRDFSIsDefinedBy()
														.get(0)));
									} else {
										attribute
												.setEType(epackage
														.getEClassifier(ran
																.getLocalName()));
									}
									EAnnotation an = ecoreFactory
											.createEAnnotation();
									an.setSource(RDFS.C_DATATYPE);
									attribute.getEAnnotations().add(an);
								}
							} // end of "if (ran.getURI() != null)"
							
							// log
							sb.append("  The type of EAttribute " + attribute.getName() + " is " + attribute.getEAttributeType().getName() + "\n");
							
							eclass.getEStructuralFeatures().add(attribute);
							
				        } // end of "if (ran instanceof RDFSDatatype)"
				        else if (ran instanceof OWLDataRange) {
				            // OWLDataRange
				            EAttribute attribute = ecoreFactory.createEAttribute();
				            attribute.setName(dp.getLocalName());
							createAnnotation(ecoreFactory, attribute, dp);
							
							// log
							sb.append("OWLDatatypeProperty " + dp.getLocalName() + " transformed to EAttribute successfully!\n");
							
							// record subPropertyOf to its annoation
							Iterator iter_spo = dp.getRDFSSubPropertyOf().iterator();
							while (iter_spo.hasNext()) {
							    RDFProperty rdfp = (RDFProperty) iter_spo.next();
							    if (rdfp != null) {
							        EAnnotation ea = ecoreFactory.createEAnnotation();
							        ea.setSource("The property is subPropertyOf " + rdfp.getLocalName() + ".");
							        attribute.getEAnnotations().add(ea);
							        
							        // log
							        sb.append("  subPropertyOf(" + dp.getLocalName() + ", " + rdfp.getLocalName() + ") transformed to EAnnotation successfully!\n");
							    }
							}
							
							// record Functional to its annoation if dp.getFunctional() is true
							Boolean b = dp.getFunctional();
							if (b != null && b.booleanValue() == true) {
						        EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is Functional.");
						        attribute.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + dp.getLocalName() + " is functional! Functional transformed to EAnnotation successfully!\n");
						    }
							
							OWLDataRange odr = (OWLDataRange) ran;
							// create EEnum
							EEnum enumDataRange = createEEnum_OWLDataRange(odr);
							// point datatype's range to this EEnum
							attribute.setEType(enumDataRange);
							
							// log
							sb.append("  The type of EAttribute " + dp.getLocalName() + " is EEnum\n");
							
							eclass.getEStructuralFeatures().add(attribute);
							
				        } // end of "else if (ran instanceof OWLDataRange)"
				    }
				}
				else if (p instanceof OWLObjectProperty) {
				    // OWLObjectProperty -> EReference
				    OWLObjectProperty op = (OWLObjectProperty) p;
				    
				    // use only the first range
				    //ran = (RDFSResource) op.getRDFSRange().iterator().next();
				    
				    if (op.getRDFSRange().iterator().hasNext() == false) {
				        // if property without range, use Object for default
				        EReference reference = ecoreFactory.createEReference();
				        reference.setName(op.getLocalName());
						createAnnotation(ecoreFactory, reference, op);
						
						// log
						sb.append("OWLObjectProperty " + reference.getName() + " transformed to EReference successfully! Default type is EObject\n");
						
						// record subPropertyOf to its annoation
						Iterator iter_spo = op.getRDFSSubPropertyOf().iterator();
						while (iter_spo.hasNext()) {
						    RDFProperty rdfp = (RDFProperty) iter_spo.next();
						    if (rdfp != null) {
						        EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is subPropertyOf " + rdfp.getLocalName() + ".");
						        reference.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  subPropertyOf(" + op.getLocalName() + ", " + rdfp.getLocalName() + ") transformed to EAnnotation successfully!\n");
						    }
						}
						
						// record Transitive to its annoation if dp.getTransitive() is true
						Boolean b = op.getTransitive();
						if (b != null && b.booleanValue() == true) {
						    EAnnotation ea = ecoreFactory.createEAnnotation();
					        ea.setSource("The property is Transitive.");
					        reference.getEAnnotations().add(ea);
					        
					        // log
					        sb.append("  " + op.getLocalName() + " is transitive! Transitive transformed to EAnnotation successfully!\n");
						}
						// record Symmetric to its annoation if dp.getSymmetric() is true
						b = op.getSymmetric();
						if (b != null && b.booleanValue() == true) {
						    EAnnotation ea = ecoreFactory.createEAnnotation();
					        ea.setSource("The property is Symmetric.");
					        reference.getEAnnotations().add(ea);
					        
					        // log
					        sb.append("  " + op.getLocalName() + " is symmetric! Symmetric transformed to EAnnotation successfully!\n");
						}
						// record Functional to its annoation if dp.getFunctional() is true
						b = op.getFunctional();
						if (b != null && b.booleanValue() == true) {
					        EAnnotation ea = ecoreFactory.createEAnnotation();
					        ea.setSource("The property is Functional.");
					        reference.getEAnnotations().add(ea);
					        
					        // log
					        sb.append("  " + op.getLocalName() + " is functional! Functional transformed to EAnnotation successfully!\n");
					    }
						// record InverseFunctional to its annoation if dp.getInverseFunctional() is true
						b = op.getInverseFunctional();
						if (b != null && b.booleanValue() == true) {
						    EAnnotation ea = ecoreFactory.createEAnnotation();
					        ea.setSource("The property is InverseFunctional.");
					        reference.getEAnnotations().add(ea);
					        
					        // log
					        sb.append("  " + op.getLocalName() + " is inverseFunctional! InverseFunctional transformed to EAnnotation successfully!\n");
						}
						
						reference.setEType(ecorePackage.getEObject());
						eclass.getEStructuralFeatures().add(reference);
						
				    } else {
				        // use only the first range
					    // TODO
				        ran = (RDFSResource) op.getRDFSRange().iterator().next();
				        
				        if (ran instanceof OWLClass) {
				            EReference reference = ecoreFactory.createEReference();
							reference.setName(op.getLocalName());
							createAnnotation(ecoreFactory, reference, op);
							
							// log
							sb.append("OWLObjectProperty " + reference.getName() + " transformed to EReference successfully!\n");
							
							// record subPropertyOf to its annoation
							Iterator iter_spo = op.getRDFSSubPropertyOf().iterator();
							while (iter_spo.hasNext()) {
							    RDFProperty rdfp = (RDFProperty) iter_spo.next();
							    if (rdfp != null) {
							        EAnnotation ea = ecoreFactory.createEAnnotation();
							        ea.setSource("The property is subPropertyOf " + rdfp.getLocalName() + ".");
							        reference.getEAnnotations().add(ea);
							        
							        // log
							        sb.append("  subPropertyOf(" + op.getLocalName() + ", " + rdfp.getLocalName() + ") transformed to EAnnotation successfully!\n");
							    }
							}
							
							// record Transitive to its annoation if dp.getTransitive() is true
							Boolean b = op.getTransitive();
							if (b != null && b.booleanValue() == true) {
							    EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is Transitive.");
						        reference.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + op.getLocalName() + " is transitive! Transitive transformed to EAnnotation successfully!\n");
							}
							// record Symmetric to its annoation if dp.getSymmetric() is true
							b = op.getSymmetric();
							if (b != null && b.booleanValue() == true) {
							    EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is Symmetric.");
						        reference.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + op.getLocalName() + " is symmetric! Symmetric transformed to EAnnotation successfully!\n");
							}
							// record Functional to its annoation if dp.getFunctional() is true
							b = op.getFunctional();
							if (b != null && b.booleanValue() == true) {
						        EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is Functional.");
						        reference.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + op.getLocalName() + " is functional! Functional transformed to EAnnotation successfully!\n");
						    }
							// record InverseFunctional to its annoation if dp.getInverseFunctional() is true
							b = op.getInverseFunctional();
							if (b != null && b.booleanValue() == true) {
							    EAnnotation ea = ecoreFactory.createEAnnotation();
						        ea.setSource("The property is InverseFunctional.");
						        reference.getEAnnotations().add(ea);
						        
						        // log
						        sb.append("  " + op.getLocalName() + " is inverseFunctional! InverseFunctional transformed to EAnnotation successfully!\n");
							}
							
							reference.setEType(epackage.getEClassifier(ran.getLocalName()));
							eclass.getEStructuralFeatures().add(reference);
							
				        }
				    }
				} // end of OWLObjectProperty
				
			} // end of "while (pi.hasNext())"
		} // end of "while (ri.hasNext())"
		
		// log
		sb.append("\n");
		
		// inverseOf
		ri = classList.iterator();
		while (ri.hasNext()) {
		    OWLClass oc = (OWLClass) ri.next();
			EClass eclass = null;
			if (!EDataType.class.isInstance(epackage.getEClassifier(oc
					.getLocalName())))
				eclass = (EClass) epackage.getEClassifier(oc.getLocalName());
			if (eclass == null)
				continue;
			
			Iterator pi = oc.getOwnedProperty().iterator();
			while (pi.hasNext()) {
			    RDFProperty p = (RDFProperty) (pi.next());
				RDFSResource ran = null;
				
				if (p instanceof OWLObjectProperty) {
				    OWLObjectProperty op = (OWLObjectProperty) p;
				    //ran = (RDFSResource) op.getRDFSRange().iterator().next();
				    if (op.getRDFSRange().iterator().hasNext() == true) {
				        // TODO
				        ran = (RDFSResource) op.getRDFSRange().iterator().next();
				        if (ran instanceof OWLClass) {
				            OWLObjectProperty owlop = op.getOWLInverseOf();
					        if ((ran != null) && (op != null) && (owlop != null)) {
					        	EReference er1 = null;
					            EReference er2 = null;
					        	Iterator i = epackage.getEClassifiers().iterator();
					        	while (i.hasNext()) {
					        		EClassifier classifier = (EClassifier) i.next();
					        		if (classifier instanceof EClass) {
					        			EClass e_class = (EClass) classifier;
					        			EReference er1_temp = (EReference) e_class.getEStructuralFeature(op.getLocalName());
					        			if (er1_temp != null) {
					        				er1 = er1_temp;
					        				continue;
					        			}
					        			EReference er2_temp = (EReference) e_class.getEStructuralFeature(owlop.getLocalName());
					        			if (er2_temp != null) {
					        				er2 = er2_temp;
					        				continue;
					        			}
					        		}
					        	}
					            if ((er1 != null) && (er2 != null)) {
					                er1.setEOpposite(er2);
					                er2.setEOpposite(er1);
					                
					                // log
					                sb.append("inverseOf(" + er1.getName() + ", " + er2.getName() + ") transformed to eOpposite successfully!\n");
					            }
					        }
				        }
				    }
				}
			}
		} // end of "inverseOf"
		
		// log
		sb.append("\n");
		
		// cardinality && hasValue
		ri = classList.iterator();
		while (ri.hasNext()) {
		    OWLClass owlClass = (OWLClass) ri.next();
		    // find anonymous class
		    if (!owlClass.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
		        continue;
		    
		    if (owlClass instanceof OWLRestriction) {
		        
		        // MinCardinalityRestriction
		        if (owlClass instanceof MinCardinalityRestriction) {
		            
		            MinCardinalityRestriction mincr = (MinCardinalityRestriction) owlClass;
		            
		            RDFProperty rdfp = mincr.getOWLOnProperty();
		            int rdfsl_value = Integer.valueOf(mincr.getOWLMinCardinality().getLexicalForm()).intValue();
		            
		            // get mincr's sub classes
		            Iterator iter_mincr_sub = reasoner.getDescendantClasses(mincr).iterator();
		            //System.out.println("---- " + owlClass.getLocalName() + " ----");
		            while (iter_mincr_sub.hasNext()) {
		                OWLClass oclass = (OWLClass) iter_mincr_sub.next();
		                //System.out.println(oclass.getLocalName());
		                
		                // filter out indirect super class whose name is mincr's name
		                boolean isdSuper = false;
		                Iterator iter_d_super = oclass.getRDFSSubClassOf().iterator();
		                while (iter_d_super.hasNext()) {
		                    OWLClass osuper = (OWLClass) iter_d_super.next();
		                    if (osuper.getLocalName().equals(mincr.getLocalName())) {
		                        isdSuper = true;
		                        break;
		                    }
		                }
		                if (!isdSuper)
		                    continue;
		                
		                EClass eclass = (EClass) epackage.getEClassifier(oclass.getLocalName());
		                if (eclass != null) {
		                    Iterator iter_mincr_sub_supers = reasoner.getAncestorClasses(oclass).iterator();
			                while (iter_mincr_sub_supers.hasNext()) {
			                    OWLClass oc = (OWLClass) iter_mincr_sub_supers.next();
			                    //if (oc.getLocalName().equals(oclass.getLocalName()))
			                        //continue;
			                    
			                    EClass ec = (EClass) epackage.getEClassifier(oc.getLocalName());
			                    if (ec != null) {
			                        Iterator iter_er = ec.getEReferences().iterator();
			                        while (iter_er.hasNext()) {
			                            EReference er = (EReference) iter_er.next();
			                            if (er.getName().equals(rdfp.getLocalName())) {
			                                
			                                boolean isExist = false;
			                                boolean isHalfExist = false;
			                                EReference temp = null;
			                                Iterator iter_count = eclass.getEReferences().iterator();
			                                while (iter_count.hasNext()) {
			                                    EReference er_temp = (EReference) iter_count.next();
			                                    if (er_temp.getName().equals(er.getName()) && er_temp.getLowerBound() == rdfsl_value) {
			                                        isExist = true;
			                                        break;
			                                    }
			                                    if (er_temp.getName().equals(er.getName())) {
			                                        isHalfExist = true;
			                                        temp = er_temp;
			                                        continue;
			                                    }
			                                }
			                                if (isExist == false && isHalfExist == false) {
			                                    EReference er_new = ecoreFactory.createEReference();
				                                
				                                er_new.setName(rdfp.getLocalName());
				                                er_new.setEType(er.getEReferenceType());
				                                er_new.setLowerBound(rdfsl_value);
				                                createAnnotation(ecoreFactory, er_new, rdfp);
				                                
				                                eclass.getEStructuralFeatures().add(er_new);
				                                
				                                // log
				                                sb.append("minCardinalityRestriction for OWLObjectProperty " + er_new.getName() + " transformed to ecore lowerBound successfully!\n");
				                                
			                                } else if (isExist == false && isHalfExist == true) {
			                                    temp.setLowerBound(rdfsl_value);
			                                    
			                                    // log
			                                    sb.append("minCardinalityRestriction for OWLObjectProperty " + temp.getName() + " transformed to ecore lowerBound successfully!\n");
			                                }
			                            }
			                        }
			                    }
			                }
		                }
		            }
		        }
		        
		        // MaxCardinalityRestriction
		        if (owlClass instanceof MaxCardinalityRestriction) {

		            MaxCardinalityRestriction maxcr = (MaxCardinalityRestriction) owlClass;
		            
		            RDFProperty rdfp = maxcr.getOWLOnProperty();
		            int rdfsl_value = Integer.valueOf(maxcr.getOWLMaxCardinality().getLexicalForm()).intValue();
		            
		            // get maxcr's sub classes
		            Iterator iter_maxcr_sub = reasoner.getDescendantClasses(maxcr).iterator();
		            //System.out.println("---- " + owlClass.getLocalName() + " ----");
		            while (iter_maxcr_sub.hasNext()) {
		                OWLClass oclass = (OWLClass) iter_maxcr_sub.next();
		                //System.out.println(oclass.getLocalName());
		                
		                // filter out indirect super class whose name is maxcr's name
		                boolean isdSuper = false;
		                Iterator iter_d_super = oclass.getRDFSSubClassOf().iterator();
		                while (iter_d_super.hasNext()) {
		                    OWLClass osuper = (OWLClass) iter_d_super.next();
		                    if (osuper.getLocalName().equals(maxcr.getLocalName())) {
		                        isdSuper = true;
		                        break;
		                    }
		                }
		                if (!isdSuper)
		                    continue;
		                
		                EClass eclass = (EClass) epackage.getEClassifier(oclass.getLocalName());
		                if (eclass != null) {
		                    Iterator iter_maxcr_sub_supers = reasoner.getAncestorClasses(oclass).iterator();
			                while (iter_maxcr_sub_supers.hasNext()) {
			                    OWLClass oc = (OWLClass) iter_maxcr_sub_supers.next();
			                    //if (oc.getLocalName().equals(oclass.getLocalName()))
			                        //continue;
			                    
			                    EClass ec = (EClass) epackage.getEClassifier(oc.getLocalName());
			                    if (ec != null) {
			                        Iterator iter_er = ec.getEReferences().iterator();
			                        while (iter_er.hasNext()) {
			                            EReference er = (EReference) iter_er.next();
			                            if (er.getName().equals(rdfp.getLocalName())) {
			                                
			                                boolean isExist = false;
			                                boolean isHalfExist = false;
			                                EReference temp = null;
			                                Iterator iter_count = eclass.getEReferences().iterator();
			                                while (iter_count.hasNext()) {
			                                    EReference er_temp = (EReference) iter_count.next();
			                                    if (er_temp.getName().equals(er.getName()) && er_temp.getUpperBound() == rdfsl_value) {
			                                        isExist = true;
			                                        break;
			                                    }
			                                    if (er_temp.getName().equals(er.getName())) {
			                                        isHalfExist = true;
			                                        temp = er_temp;
			                                        continue;
			                                    }
			                                }
			                                if (isExist == false && isHalfExist == false) {
			                                    EReference er_new = ecoreFactory.createEReference();
				                                
				                                er_new.setName(rdfp.getLocalName());
				                                er_new.setEType(er.getEReferenceType());
				                                
				                                er_new.setUpperBound(rdfsl_value);
				                                createAnnotation(ecoreFactory, er_new, rdfp);
				                                
				                                eclass.getEStructuralFeatures().add(er_new);
				                                
				                                
				                                // log
				                                sb.append("maxCardinalityRestriction for OWLObjectProperty " + er_new.getName() + " transformed to ecore upperBound successfully!\n");
				                                
			                                } else if (isExist == false && isHalfExist == true) {
			                                    temp.setUpperBound(rdfsl_value);
			                                    
			                                    // log
			                                    sb.append("maxCardinalityRestriction for OWLObjectProperty " + temp.getName() + " transformed to ecore upperBound successfully!\n");
			                                }
			                            }
			                        }
			                    }
			                }
		                }
		            }
		        }
		        
		        // CardinalityRestriction
		        if (owlClass instanceof CardinalityRestriction) {
		            
		            CardinalityRestriction cr = (CardinalityRestriction) owlClass;
		            
		            RDFProperty rdfp = cr.getOWLOnProperty();
		            int rdfsl_value = Integer.valueOf(cr.getOWLCardinality().getLexicalForm()).intValue();
		            
		            // get cr's sub classes
		            Iterator iter_cr_sub = reasoner.getDescendantClasses(cr).iterator();
		            //System.out.println("---- " + owlClass.getLocalName() + " ----");
		            while (iter_cr_sub.hasNext()) {
		                OWLClass oclass = (OWLClass) iter_cr_sub.next();
		                //System.out.println(oclass.getLocalName());
		                
		                // filter out indirect super class whose name is cr's name
		                boolean isdSuper = false;
		                Iterator iter_d_super = oclass.getRDFSSubClassOf().iterator();
		                while (iter_d_super.hasNext()) {
		                    OWLClass osuper = (OWLClass) iter_d_super.next();
		                    if (osuper.getLocalName().equals(cr.getLocalName())) {
		                        isdSuper = true;
		                        break;
		                    }
		                }
		                if (!isdSuper)
		                    continue;
		                
		                EClass eclass = (EClass) epackage.getEClassifier(oclass.getLocalName());
		                if (eclass != null) {
		                    Iterator iter_cr_sub_supers = reasoner.getAncestorClasses(oclass).iterator();
			                while (iter_cr_sub_supers.hasNext()) {
			                    OWLClass oc = (OWLClass) iter_cr_sub_supers.next();
			                    //if (oc.getLocalName().equals(oclass.getLocalName()))
			                        //continue;
			                    
			                    EClass ec = (EClass) epackage.getEClassifier(oc.getLocalName());
			                    if (ec != null) {
			                        Iterator iter_er = ec.getEReferences().iterator();
			                        while (iter_er.hasNext()) {
			                            EReference er = (EReference) iter_er.next();
			                            if (er.getName().equals(rdfp.getLocalName())) {
			                                
			                                boolean isExist = false;
			                                boolean isHalfExist = false;
			                                EReference temp = null;
			                                Iterator iter_count = eclass.getEReferences().iterator();
			                                while (iter_count.hasNext()) {
			                                    EReference er_temp = (EReference) iter_count.next();
			                                    if (er_temp.getName().equals(er.getName()) && er_temp.getLowerBound() == rdfsl_value && er_temp.getUpperBound() == rdfsl_value) {
			                                        isExist = true;
			                                        break;
			                                    }
			                                    if (er_temp.getName().equals(er.getName())) {
			                                        isHalfExist = true;
			                                        temp = er_temp;
			                                        continue;
			                                    }
			                                }
			                                if (isExist == false && isHalfExist == false) {
			                                    EReference er_new = ecoreFactory.createEReference();
				                                
				                                er_new.setName(rdfp.getLocalName());
				                                er_new.setEType(er.getEReferenceType());
				                                er_new.setLowerBound(rdfsl_value);
				                                er_new.setUpperBound(rdfsl_value);
				                                createAnnotation(ecoreFactory, er_new, rdfp);
				                                
				                                eclass.getEStructuralFeatures().add(er_new);
				                                
				                                // log
				                                sb.append("cardinalityRestriction for OWLObjectProperty " + er_new.getName() + " transformed to ecore lowerBound and upperBound successfully!\n");
				                                
			                                } else if (isExist == false && isHalfExist == true) {
			                                    temp.setLowerBound(rdfsl_value);
			                                    temp.setUpperBound(rdfsl_value);
			                                    
			                                    // log
			                                    sb.append("cardinalityRestriction for OWLObjectProperty " + temp.getName() + " transformed to ecore lowerBound and upperBound successfully!\n");
			                                }
			                            }
			                        }
			                    }
			                }
		                }
		            }
		        }
		        
		        // HasValueRestriction
		        if (owlClass instanceof HasValueRestriction) {
		            
		            HasValueRestriction hvr = (HasValueRestriction) owlClass;
		            
		            RDFProperty rdfp = hvr.getOWLOnProperty();
		            // TODO
		            //RDFSResource rdfs = hvr.getOWLHasValue();
		            
		            Iterator iter_hvr_sub = reasoner.getDescendantClasses(hvr).iterator();
		            while (iter_hvr_sub.hasNext()) {
		                OWLClass oclass = (OWLClass) iter_hvr_sub.next();
		                
		                boolean isdSuper = false;
		                Iterator iter_d_super = oclass.getRDFSSubClassOf().iterator();
		                while (iter_d_super.hasNext()) {
		                    OWLClass osuper = (OWLClass) iter_d_super.next();
		                    if (osuper.getLocalName().equals(hvr.getLocalName())) {
		                        isdSuper = true;
		                        break;
		                    }
		                }
		                if (!isdSuper)
		                    continue;
		                
		                EClass eclass = (EClass) epackage.getEClassifier(oclass.getLocalName());
		                if (eclass != null) {
		                    Iterator iter_hvr_sub_supers = reasoner.getAncestorClasses(oclass).iterator();
		                    while (iter_hvr_sub_supers.hasNext()) {
		                        OWLClass oc = (OWLClass) iter_hvr_sub_supers.next();
		                        //if (oc.getLocalName().equals(oclass.getLocalName()))
			                        //continue;
		                        
		                        EClass ec = (EClass) epackage.getEClassifier(oc.getLocalName());
		                        if (ec != null) {
		                            Iterator iter_ea = ec.getEAttributes().iterator();
		                            while (iter_ea.hasNext()) {
		                                EAttribute ea = (EAttribute) iter_ea.next();
		                                
		                                if (ea.getName().equals(rdfp.getLocalName())) {
		                                    
		                                    boolean isExist = false;
		                                    boolean isHalfExist = false;
			                                EAttribute temp = null;
			                                Iterator iter_count = eclass.getEAttributes().iterator();
			                                while (iter_count.hasNext()) {
			                                    EAttribute ea_temp = (EAttribute) iter_count.next();
			                                    
			                                    if (ea_temp.getName().equals(ea.getName()) && ea_temp.getDefaultValueLiteral().equals(ea.getDefaultValueLiteral())) {
			                                        isExist = true;
			                                        break;
			                                    }
			                                    if (ea_temp.getName().equals(ea.getName())) {
			                                        isHalfExist = true;
			                                        temp = ea_temp;
			                                        continue;
			                                    }
			                                }
			                                if (isExist == false && isHalfExist == false) {
			                                    EAttribute ea_new = ecoreFactory.createEAttribute();
				                                
				                                ea_new.setName(ea.getName());
				                                ea_new.setEType(ea.getEAttributeType());
				                                ea_new.setDefaultValueLiteral(ea.getDefaultValueLiteral());
				                                createAnnotation(ecoreFactory, ea_new, rdfp);
				                                
				                                eclass.getEStructuralFeatures().add(ea_new);
				                                
				                                // log
				                                sb.append("hasValueRestriction for OWLDatatypeProperty " + ea_new.getName() + " transformed to ecore defaultValueLiteral successfully!\n");
				                                
			                                } else if (isExist == false && isHalfExist == true) {
			                                    temp.setDefaultValueLiteral(ea.getDefaultValueLiteral());
			                                    
			                                    // log
			                                    sb.append("hasValueRestriction for OWLDatatypeProperty " + temp.getName() + " transformed to ecore defaultValueLiteral successfully!\n");
			                                }
			                                
		                                }
		                            }
		                        }
		                    }
		                }
		            }
		        }
		        
		    }
		}
		
		// log
		sb.append("\n");
		
	}
	
	/*
	 * record log for transformer
	 * 
	 * @param s
	 */
	private void log(String s) {
        if ( bw != null )
        {
	    try {
            bw.write(s);
            bw.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
        }
	}
	
	/*
	 * preprocess ontology
	 * 
	 * @param ontology
	 */
	private void preprocessOntology(OWLOntology ontology) {
	    // initialize reasoner
	    //OWLTaxonomyReasoner reasoner = StructuralReasonerFactory.instance().createOWLTaxonmyReasoner();
	    reasoner.initialize(ontology);
	    
	    List classList = ontology.getResourceList(OWLClass.class);
		Iterator ri = classList.iterator();
		
		while (ri.hasNext()) {
		    OWLClass oc = (OWLClass) ri.next();
		    
		    // filter out Restriction anonymous owlclass
			if (oc.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
				continue;
		    
			// record the direct super classes
		    Set tempSuperClass = new HashSet();
			
			List directSuperClass = oc.getRDFSSubClassOf();
		    List indirectSuperClass = reasoner.getAncestorClasses(oc);
		    
		    //Iterator iter_dc = directSuperClass.iterator();
		    Iterator iter_idc = indirectSuperClass.iterator();
		    
		    while (iter_idc.hasNext()) {
		        OWLClass superClass = (OWLClass) iter_idc.next();
		        
		        // ?
		        // filter out Restriction anonymous owlclass
				if (superClass.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
					continue;
		        
		        // filter out oc itself
		        if (superClass.getLocalName().equals(oc.getLocalName()))
		            continue;
		        else {
		            Iterator i = reasoner.getAncestorClasses(superClass).iterator();
		            while (i.hasNext()) {
		                OWLClass c_i = (OWLClass) i.next();
		                
		                // ?
				        // filter out Restriction anonymous owlclass
						if (c_i.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
							continue;
						
		                if (c_i.getLocalName().equals(superClass.getLocalName()))
		                    continue;
		                else {
		                    if (!tempSuperClass.contains((Object)c_i))
		                        tempSuperClass.add(c_i);
		                }
		            }// end of "while (i.hasNext())"
		        }
		    }// end of "while (iter_idc.hasNext())"
		    
		    // remove the owl class whose localname equals oc's
		    indirectSuperClass.remove((Object)oc);
		    
		    // remove the owl class which exists in tempSuperClass
		    indirectSuperClass.removeAll(tempSuperClass);
		    
		    // add subClassOf through indirectSuperClass
		    Iterator iter_isc = indirectSuperClass.iterator();
		    while (iter_isc.hasNext()) {
		        Object object = iter_isc.next();
		        if (directSuperClass.contains(object))
		            continue;
		        else {
		            // add subClassOf
		            oc.getRDFSSubClassOf().add((OWLClass)object);
		        }
		    }
		}
		
		// log
		sb.append("Ontology preprocessed successfully!\n\n");
	}
	
	/*
	 * merge the import ontology to this ontology
	 */
	private void mergeOnto() {
	    
		List ontologies = ontology.getOWLImports();
		Iterator oi = ontologies.iterator();
		while (oi.hasNext()) {
			OWLOntology owlontology = (OWLOntology) oi.next();
			//owl2Ecore(owlontology);
			ontology.getContains().addAll(owlontology.getContains());
			
			// log
			sb.append("Ontology " + owlontology.getLocalName() + " imported successfully!\n");
		}
		// log
		sb.append("\n");
	}
	
	/*
	 * save ecore model
	 */
	private void saveEPackage() throws OWLTransformationException {
	    
		ResourceSet resourceSet = new ResourceSetImpl();
		resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap()
				.put(Resource.Factory.Registry.DEFAULT_EXTENSION,
						new EcoreResourceFactoryImpl());

		Resource resource = resourceSet.createResource(URI
				.createFileURI(new File(outputEcoreFile).getAbsolutePath()));
		resource.getContents().add(epackage);
		try {
			resource.save(Collections.EMPTY_MAP);
			
			// log
			sb.append("ecore model saved successfully!\n\n");
		} catch (Exception e) {
			OWLTransformationException te = new OWLTransformationException(
					"Error during serialize into XMI file: " + e.toString());
			
			// log
			sb.append("Error during serialize into XMI file: " + e.toString() + "\n\n");
			log(sb.toString());
			
			throw te;
		}
	}
	
	/*
	 * 
	 */
	private boolean containSameName(EClass ec, OWLClass oc) {
	    boolean flag = false;
	    List list = ec.getESuperTypes();
	    Iterator iter = list.iterator();
	    //System.out.println(list.size());
	    while (iter.hasNext()) {
	        EClass eclass = (EClass) iter.next();
	        //System.out.println("SuperEClass --- " + eclass.getName());
	        if (eclass.getName().equals(oc.getLocalName())) {
	            flag = true;
	            break;
	        }
	    }
	    
	    return flag;
	}
	
	/*
	 * check inter-subclassof
	 */
	private boolean checkInterSubClassOf(EClass ec1, EClass ec2) {
	    boolean flag = false;
	    Iterator iter_ec2 = ec2.getESuperTypes().iterator();
	    while (iter_ec2.hasNext()) {
	        EClass superClass = (EClass) iter_ec2.next();
	        if (superClass.getName().equals(ec1.getName())) {
	            flag = true;
	            break;
	        }
	    }
	    
	    return flag;
	}
	
	/**
	 * Check input and output files
	 * 
	 * @param input
	 *            input owl file
	 * @param output
	 *            target ecore file
	 * @throws OWLTransformationException
	 */
	private void checkFiles(String input, String output)
			throws OWLTransformationException {
		//check input params
		if (input == null) {
			OWLTransformationException e = new OWLTransformationException(
					"Initate mapping failed : input OWL model is NULL.");
			throw e;
		}

		//check and set input and output files
		if (input != null)
			if (input.equals("") == false)
				setInputOWLFile(input);
		if (output != null)
			if (output.equals("") == false)
				setOutputEcoreFile(output);
	}

	/**
	 * load OWL file and create a OWL ontology model
	 * 
	 * @return
	 * @throws OWLTransformationException
	 */
	private boolean loadOWLModel() throws OWLTransformationException {
		if (inputOWLFile == null) {
			OWLTransformationException e = new OWLTransformationException(
					"Load OWLModel failed : inputfile not found.");
			
			// log
			sb.append("Load OWLModel failed : inputfile not found.\n\n");
			log(sb.toString());
			
			throw e;
		}

		ResourceSet resourceSet = null;
		URI inFileURI = null;
		Resource inputResource = null;

		try {
			resourceSet = new ResourceSetImpl();
			resourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap()
					.put(Resource.Factory.Registry.DEFAULT_EXTENSION,
							new OWLXMLResourceFactoryImpl());
			inFileURI = URI.createFileURI(new File(inputOWLFile)
					.getAbsolutePath());
			inputResource = resourceSet.createResource(inFileURI);
			inputResource.load(Collections.EMPTY_MAP);
			ontology = (inputResource.getContents().get(0) == null ? null
					: (OWLOntology) inputResource.getContents().get(0));
			if (ontology == null) {
				OWLTransformationException e = new OWLTransformationException(
						"Create OWLModel failed : no RDFDescription in the input file.");
				
				// log
				sb.append("Create OWLModel failed : no RDFDescription in the input file.\n\n");
				log(sb.toString());
				
				throw e;
			}
			
			// log
			sb.append("OWLModel created successfully!\n\n");
			
		} catch (IOException e) {
			OWLTransformationException te = new OWLTransformationException(
					"Create OWLModel failed: " + e.getMessage());
			
			// log
			sb.append("Create OWLModel failed: " + e.getMessage() + "\n\n");
			log(sb.toString());
			
			throw te;
			// return false;
		} finally {
			resourceSet = null;
			inFileURI = null;
			inputResource = null;
		}

		return true;
	}

	/**
	 * initiate the ecore model
	 * 
	 * @return
	 */
	private boolean createEcoreModel() {
		if (outputEcoreFile == null || ontology == null)
			return false;
        
        String nsPrefix = null;
        String nsURI = null;
    
        Namespace ns = ontology.getNamespace();
        
        if(ns!=null) {
            nsPrefix = ns.getName();
            nsURI = ns.getURI();
        }
        
		// set up EPackage nsPrefix
		epackage.setName(ontology.getLocalName());
		epackage.setNsPrefix(nsPrefix);
		epackage.setNsURI(nsURI);
		EPackage.Registry.INSTANCE.put(nsURI, epackage);
		
		// log
		sb.append("Ecore model created successfully!\n\n");

		return true;
	}

	/**
	 * create eAnnotations from comment, isDefinedBy and Label in RDFSResource
	 * 
	 * @param fact
	 * @param ec
	 * @param rs
	 */
	private void createAnnotation(EcoreFactory fact, ENamedElement ec,
			RDFSResource rs) {

		if (fact == null || ec == null || rs == null)
			return;

		// check input params
		if (rs == null || ec == null || fact == null)
			return;

		// create comments as annotations
		Iterator it = rs.getRDFSComment().iterator();
		while (it.hasNext()) {
			if (ec instanceof EDataType) {
				String comment = ((RDFSLiteral) it.next()).getLexicalForm();
				if (comment.startsWith(XSD.ECORE_INSTANCE_CLASS_NAME)) {
					((EDataType) ec).setInstanceClassName(comment
							.substring(XSD.ECORE_INSTANCE_CLASS_NAME.length()));
					continue;
				}
			}
			EAnnotation anno = fact.createEAnnotation();
			anno.setSource("comment - "
					+ ((RDFSLiteral) it.next()).getLexicalForm());
			ec.getEAnnotations().add(anno);
		}

		// create isDefinedBy as annotations
		it = rs.getRDFSIsDefinedBy().iterator();
		while (it.hasNext()) {
			EAnnotation anno = fact.createEAnnotation();
			anno.setSource("isDefinedBy - "
					+ ((RDFSResource) it.next()).getURI());
			ec.getEAnnotations().add(anno);
		}

		// create Label as annotations
		it = rs.getRDFSLabel().iterator();
		while (it.hasNext()) {
			EAnnotation anno = fact.createEAnnotation();
			anno.setSource("label - "
					+ ((RDFSLiteral) it.next()).getLexicalForm());
			ec.getEAnnotations().add(anno);
		}

	}

	/**
	 * Create EEnum from OWL EnumeratedClass
	 * 
	 * @param enumclass
	 * @return
	 */
	private EEnum createEEnum(EnumeratedClass enumclass) {
		if (enumclass == null)
			return null;
		Iterator memberList = enumclass.getOWLOneOf().iterator();
		EEnum eenum = null;

		// if memberList.size > 0, create EEnum
		if (memberList.hasNext()) {
			eenum = ecoreFactory.createEEnum();
			//eenum.setName(enumclass.getLocalName());
			eenum.setName("EEnum");
			createAnnotation(ecoreFactory, eenum, enumclass);
			
			// log
			//sb.append("EnumeratedClass " + eenum.getName() + " transformed to EEnum successfully!\n");
			sb.append("EnumeratedClass transformed to EEnum successfully!\n");
		}

		// create enumliterals from enumclass members
		int intValue = 0;
		//System.out.println("-- Enum Start --");
		while (memberList.hasNext()) {
			RDFSResource r = (RDFSResource) (memberList.next());
			// test whether contains label "EEnum"
			String enumeration = (r instanceof RDFSLiteral ? ((RDFSLiteral) r)
					.getLexicalForm() : r.getLocalName());

			// create EEnumLiteral
			EEnumLiteral eliteral = ecoreFactory.createEEnumLiteral();
			createAnnotation(ecoreFactory, eliteral, r);
			eliteral.setName(enumeration);
			eliteral.setValue(intValue);
			eenum.getELiterals().add(eliteral);
			intValue++;
			
			//System.out.println("Individual " + enumeration + " lost.");
			
			// log
			sb.append("  Individual " + enumeration + " lost!\n");
		}
		//System.out.println("-- Enum End --");
		
		epackage.getEClassifiers().add(eenum);

		return eenum;
	}
	
	/*
	 * Create EEnum from OWL DataRange
	 * 
	 * @param dr
	 * @return
	 */
	private EEnum createEEnum_OWLDataRange(OWLDataRange dr) {
	    if (dr == null)
			return null;
		Iterator memberList = dr.getOWLOneOf().iterator();
		EEnum eenum = null;
		
		// if memberList.size > 0, create EEnum
		if (memberList.hasNext()) {
			eenum = ecoreFactory.createEEnum();
			//eenum.setName(enumclass.getLocalName());
			eenum.setName("EEnum");
			createAnnotation(ecoreFactory, eenum, dr);
		}
		
		// create enumliterals from dr members
		int intValue = 0;
		//System.out.println("-- Enum Start --");
		while (memberList.hasNext()) {
			RDFSResource r = (RDFSResource) (memberList.next());
			// test whether contains label "EEnum"
			String enumeration = (r instanceof RDFSLiteral ? ((RDFSLiteral) r)
					.getLexicalForm() : r.getLocalName());

			// create EEnumLiteral
			EEnumLiteral eliteral = ecoreFactory.createEEnumLiteral();
			createAnnotation(ecoreFactory, eliteral, r);
			eliteral.setName(enumeration);
			eliteral.setValue(intValue);
			eenum.getELiterals().add(eliteral);
			intValue++;
			
			//System.out.println("Individual " + enumeration + " lost.");
		}
		//System.out.println("-- Enum End --");
		epackage.getEClassifiers().add(eenum);
		
	    return eenum;
	}

	/**
	 * set input owl file's URL. If outputfile's URL is "null", automaticly
	 * generate a .ecore file in the same folder as default output file.
	 * 
	 * @param in
	 *            owl file's URL
	 */
	public void setInputOWLFile(String in) {
		if (in == null) {
			inputOWLFile = null;
			return;
		}
		inputOWLFile = in;

		if (outputEcoreFile == null) {
			int p = inputOWLFile.lastIndexOf(".");
			outputEcoreFile = (p != -1 ? inputOWLFile.substring(0, p + 1)
					+ "ecore" : inputOWLFile + ".ecore");
		}
	}

	/**
	 * set output file URL
	 * 
	 * @param out
	 *            ecore file's URL
	 */
	public void setOutputEcoreFile(String out) {
		if (out == null)
			return;
		outputEcoreFile = out;
	}

	/**
	 * @return
	 */
	public String getInputOWLFile() {
		return inputOWLFile;
	}

	/**
	 * @return
	 */
	public String getOutputEcoreFile() {
		return outputEcoreFile;
	}

    private static String OWL_THING = "http://www.w3.org/2002/07/owl#Thing";

    
	public static void main(String[] args) {
		OWL2ECoreTransformer owl2ecore = new OWL2ECoreTransformer();
		try {
			owl2ecore.convertOWL2Ecore("testcases/publication.owl");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
