/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved.   This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 *   
 * $Id: RDFXMLSaver.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.rdf.resource;

import java.io.BufferedWriter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Iterator;
import java.util.Random;

import org.eclipse.emf.common.util.EList;
import org.eclipse.eodm.rdf.resource.parser.wrapper.RDFNamespaceMap;
import org.eclipse.eodm.rdfs.Namespace;
import org.eclipse.eodm.rdfs.Ontology;
import org.eclipse.eodm.rdfs.RDFSFactory;
import org.eclipse.eodm.rdfs.RDFSLiteral;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.vocabulary.RDF;
import org.eclipse.eodm.rdfs.vocabulary.RDFS;

/**
 * Save a EODM-based RDF memory model into a file or output stream in RDF/XML
 * syntax.
 *  
 */
public class RDFXMLSaver {

    /**
     * A system-generated namespace for those resources that do not have
     * namespaces
     */
    protected static Namespace sysGenNamespace = null;

    /**
     * Save a RDF ontology into a file in RDF/XML syntax using the system
     * default encoding.
     * 
     * @param ontology
     *            The EODM-based RDF ontology to be saved.
     * @param fileName
     *            The name of the file to save to.
     * @throws IOException
     */
    public static void saveToFile(Ontology ontology, String fileName)
            throws IOException {
        saveToStream(ontology, new FileOutputStream(fileName));
    }

    /**
     * Save a RDF ontology to a file in RDF/XML syntax using a specified
     * encoding/charset.
     * 
     * @param ontology
     *            The EODM-based RDF ontology
     * @param fileName
     *            The name of the file to save to
     * @param charsetName
     *            The name of the charset that will be used for saving.
     * @throws IOException
     */
    public static void saveToFile(Ontology ontology, String fileName,
            String charsetName) throws IOException {
        saveToStream(ontology, new FileOutputStream(fileName), charsetName);
    }

    /**
     * Save a RDF ontology into an output steam in RDF/XML syntax using the
     * system default encoding.
     * 
     * @param ontology
     *            The EODM-based RDF ontology
     * @param os
     *            The output steam to save to
     * @throws IOException
     */
    public static void saveToStream(Ontology ontology, OutputStream os)
            throws IOException {
        OutputStreamWriter osWriter = new OutputStreamWriter(os);
        saveToStream(ontology, new BufferedWriter(osWriter), osWriter
                .getEncoding());
    }

    /**
     * Save a RDF ontology into an output steam in RDF/XML syntax using a
     * specified encoding/charset.
     * 
     * @param ontology
     *            The RDF ontology to save
     * @param os
     *            The output stream to write to
     * @param charsetName
     *            The charset that will be used to write to the stream
     * @throws IOException
     */
    public static void saveToStream(Ontology ontology, OutputStream os,
            String charsetName) throws IOException {
        saveToStream(ontology, new BufferedWriter(new OutputStreamWriter(os,
                charsetName)), charsetName);
    }

    /**
     * Save a RDF ontology into a output stream writer in RDF/XML syntax.
     * 
     * @param ontology
     *            The RDF ontology to save
     * @param writer
     *            The writer for saving
     * @param encoding
     *            The encoding used for saving, it should be the same as the
     *            encoding used by the writer.
     * @throws IOException
     */
    protected static void saveToStream(Ontology ontology, Writer writer,
            String encoding) throws IOException {
        
        //Replace all invalid characters
        if (ontology.getLocalName() !=null){
            ontology.setLocalName(replaceKeywords(ontology.getLocalName()));
            ontology.setLocalName(ontology.getLocalName().replaceAll(" ", "%20"));
        }
        EList list=ontology.getOwnedNamespace();
        for (int i=0; i<list.size();i++){
            ((Namespace)list.get(i)).setURI(replaceKeywords(((Namespace)list.get(i)).getURI()));
        }
        
        EList listOnto=ontology.getContains();
        int size =ontology.getContains().size();
        for (int i=0; i<size;i++){
            
            if (listOnto.get(i) instanceof RDFSLiteral)
            {
                if (((RDFSLiteral)listOnto.get(i)).getLexicalForm()!=null)
                    ((RDFSLiteral)listOnto.get(i)).setLexicalForm(replaceKeywords(((RDFSLiteral)listOnto.get(i)).getLexicalForm()));
                continue;
            }
            if ( listOnto.get(i) instanceof RDFSResource){
                if (((RDFSResource)listOnto.get(i)).getLocalName()!=null){
                ((RDFSResource)listOnto.get(i)).setLocalName(replaceKeywords(((RDFSResource)listOnto.get(i)).getLocalName()));
                ((RDFSResource)listOnto.get(i)).setLocalName((((RDFSResource)listOnto.get(i)).getLocalName()).replaceAll(" ", "%20"));
                }
            }
        }
        
        processErrorNamespaces(ontology);

        writeRDFXMLHeader(ontology, writer, encoding);

        RDFXMLSaverImpl saverImpl = new RDFXMLSaverImpl(writer, ontology);
        saverImpl.save();

        writeRDFXMLEnding(writer);

        writer.flush();
    }

    /**
     * Replace resources that have incorrect namespaces with the system
     * generated one
     * 
     * @param ontology
     */
    protected static void processErrorNamespaces(Ontology ontology) {
        boolean errorNamespaceFound = false;
        Random randomGen = new Random();

        for (Iterator iterator = ontology.getContains().iterator(); iterator
                .hasNext();) {
            RDFSResource res = (RDFSResource) iterator.next();

            // ignore literal and blank node
            if (res instanceof RDFSLiteral)
                continue;
            if (res.getNamespace() != null
                && res.getNamespace().getURI() != null
                && res.getNamespace().getURI().equals(RDFS.BNODE_NAMESPACE))
                continue;

            if (res.getNamespace() == null
                || res.getNamespace().getName() == null
                || res.getNamespace().getURI() == null
                || res.getNamespace().getName().length() == 0
                || res.getNamespace().getURI().length() == 0) {
                res.setNamespace(getSysGenNamespace());
                errorNamespaceFound = true;
            }
            if (res.getLocalName() == null) {
                res.setLocalName(Integer.toString(randomGen.nextInt()));
            }
        }

        if (errorNamespaceFound)
            if (!ontology.getOwnedNamespace().contains(sysGenNamespace))
                ontology.getOwnedNamespace().add(sysGenNamespace);
    }

    /**
     * Get the system-generated random namespace
     * 
     * @return the system-generated random namespace
     */
    protected static Namespace getSysGenNamespace() {
        if (sysGenNamespace == null) {
            Random randomGen = new Random();
            sysGenNamespace = RDFSFactory.eINSTANCE.createNamespace();
            String randomStr = Integer.toString(randomGen.nextInt());
            sysGenNamespace.setName("SYSGEN_NS_" + randomStr);
            sysGenNamespace.setURI("uri:sysgen_ns_" + randomStr + "#");
        }
        return sysGenNamespace;
    }

    /**
     * Write the ending of a RDF/XML document.
     * 
     * @param writer
     * @throws IOException
     */
    protected static void writeRDFXMLEnding(Writer writer) throws IOException {
        writer.write("</rdf:RDF>\n");
    }

    /**
     * Write the head part of a RDF/XML document, including the XML head and all
     * the namespace definitions.
     * 
     * @param namespaceList
     *            the list of namespaces to write
     * @param writer
     *            the writer to use
     * @param encoding
     *            the encoding to use
     * @throws IOException
     *             If I/O error occurs
     */
    protected static void writeRDFXMLHeader(Ontology ontology, Writer writer,
            String encoding) throws IOException {
        
        //Firstly, re-name the abbreviation of namespaces to aviod conflic
        EList list=ontology.getOwnedNamespace();
        RDFNamespaceMap nsMap=new RDFNamespaceMap();
        for (int i=0; i<list.size()-1;i++){        
            for (int j=i+1; j<list.size() ;j++){
                if (((Namespace)(list.get(i))).getName().equals(((Namespace)(list.get(j))).getName())){
                    ((Namespace)(list.get(j))).setName(nsMap.getKey());
                }                    
            }
        }
        
        boolean foundRDFNamespace = false;
        boolean foundRDFSNamespace = false;

        writer.write("<?xml version=\"1.0\" encoding=\"" + encoding + "\"?>\n");
        writer.write("<!DOCTYPE rdf:RDF [\n");

        for (Iterator iterator = ontology.getOwnedNamespace().iterator(); iterator
                .hasNext();) {
            Namespace ns = (Namespace) iterator.next();
            if (!ns.getURI().equals(RDFS.BNODE_NAMESPACE) ) {
                writer.write("\t<!ENTITY "
                             + ns.getName() + " \"" + ns.getURI() + "\">\n");
                if (ns.getURI().equals(RDF.NAMESPACE))
                    foundRDFNamespace = true;
                if (ns.getURI().equals(RDFS.NAMESPACE))
                    foundRDFSNamespace = true;
                if (ns == sysGenNamespace)
                    writer
                            .write("\t<!-- some resource has incorrect namespace, system generated one for it -->\n");
            }
        }

        if (!foundRDFNamespace) {
            writer.write("\t<!ENTITY rdf \"" + RDF.NAMESPACE + "\">\n");
        }

        if (!foundRDFSNamespace) {
            writer.write("\t<!ENTITY rdfs \"" + RDFS.NAMESPACE + "\">\n");
        }

        writer.write("]>\n");

        writer.write("<rdf:RDF\n");

        for (Iterator iterator = ontology.getOwnedNamespace().iterator(); iterator
                .hasNext();) {
            Namespace ns = (Namespace) iterator.next();
            if (!ns.getURI().equals(RDFS.BNODE_NAMESPACE)){        
                    writer.write("         xmlns:"
                             + ns.getName() + "=\"&" + ns.getName() + ";\"\n");
            }
        }

        if (!foundRDFNamespace) {
            writer.write("         xmlns:rdf=\"&rdf;\"\n");
        }

        if (!foundRDFSNamespace) {
            writer.write("         xmlns:rdfs=\"&rdfs;\"\n");
        }

        writer.write(">\n");
    }
    
    private static String replaceKeywords(String str) {
        str = str.replaceAll("&", "&amp;");
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("'", "&apos;");
        str = str.replaceAll("\"", "&quot;");
        return str;
    }
}