/**
 * <copyright> 
 * 
 * Copyright (c) 2004-2005 IBM Corporation and others. 
 * All rights reserved.   This program and the accompanying materials 
 * are made available under the terms of the Eclipse Public License - v 1.0 
 * which accompanies this distribution, and is available at 
 * http://opensource.org/licenses/eclipse-1.0.txt 
 * 
 * Contributors: 
 *   IBM - Initial API and implementation 
 * 
 * </copyright> 
 *   
 * $Id: RDFXMLSaverImpl.java,v 1.1.1.1 2006/01/31 09:13:26 gxie Exp $
 */

package org.eclipse.eodm.rdf.resource;

import java.io.IOException;
import java.io.Writer;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.eodm.rdfs.Ontology;
import org.eclipse.eodm.rdfs.PlainLiteral;
import org.eclipse.eodm.rdfs.RDFAlt;
import org.eclipse.eodm.rdfs.RDFBag;
import org.eclipse.eodm.rdfs.RDFList;
import org.eclipse.eodm.rdfs.RDFProperty;
import org.eclipse.eodm.rdfs.RDFSClass;
import org.eclipse.eodm.rdfs.RDFSContainer;
import org.eclipse.eodm.rdfs.RDFSContainerMembershipProperty;
import org.eclipse.eodm.rdfs.RDFSDatatype;
import org.eclipse.eodm.rdfs.RDFSLiteral;
import org.eclipse.eodm.rdfs.RDFSPackage;
import org.eclipse.eodm.rdfs.RDFSResource;
import org.eclipse.eodm.rdfs.RDFSeq;
import org.eclipse.eodm.rdfs.RDFStatement;
import org.eclipse.eodm.rdfs.RDFXMLLiteral;
import org.eclipse.eodm.rdfs.TypedLiteral;
import org.eclipse.eodm.rdfs.util.RDFSSwitch;

/**
 * Save a EODM RDFS ontology model into a RDF/XML syntax resource. We re-use
 * {@link org.eclipse.eodm.rdfs.util.RDFSSwitch RDFSSwitch}to traverse the Java
 * types of the RDF resource.
 *  
 */
class RDFXMLSaverImpl extends RDFSSwitch {

    /**
     * The {@link java.io.Writer writer}to use for output
     */
    protected Writer writer;

    /**
     * Resources in the RDFS ontology that are not RDFStatement.
     */
    protected Vector rdfResources;

    /**
     * Resources in the RDFS ontology that are RDFStatement.
     */
    protected Vector instanceStatements;

    /**
     * Any I/O exception caught during the type traversing process. If it's not
     * <code>null</code>, the traversing process will be ended.
     */
    protected IOException exceptionCaught;

    /**
     * The level of output indentation.
     */
    protected int indentLevel;

    /**
     * A flag that signals the starting of a XML element, also the starting of a
     * type traversing process.
     */
    protected boolean isElementStarting;

    /**
     * The type string of the current XML element.
     */
    protected String elementType;

    /**
     * Get the corrsponding XML element type string for a RDFResource object.
     * 
     * @param object
     *            the RDFResource object
     * @return the XML element type string for the type of the RDFResource
     *         object.
     */
    protected String getElementType(RDFSResource object) {
        switch (object.eClass().getClassifierID()) {
        case RDFSPackage.RDFS_RESOURCE:
            return "rdfs:Resource";
        case RDFSPackage.RDFS_CLASS:
            return "rdfs:Class";
        case RDFSPackage.RDFS_LITERAL:
            return "rdfs:Literal";
        case RDFSPackage.RDFS_DATATYPE:
            return "rdfs:Datatype";
        case RDFSPackage.RDFXML_LITERAL:
            return "rdfs:Literal";
        case RDFSPackage.RDF_PROPERTY:
            return "rdf:Property";
        case RDFSPackage.RDF_LIST:
            return "rdf:List";
        case RDFSPackage.RDFS_CONTAINER:
            return "rdfs:Container";
        case RDFSPackage.RDF_ALT:
            return "rdf:Alt";
        case RDFSPackage.RDF_BAG:
            return "rdf:Bag";
        case RDFSPackage.RDFS_EQ:
            return "rdf:Seq";
        case RDFSPackage.RDFS_CONTAINER_MEMBERSHIP_PROPERTY:
            return "rdfs:ContainerMembershipProperty";
        case RDFSPackage.RDFS_TATEMENT:
            return "rdf:Statement";
        case RDFSPackage.ONTOLOGY:
            return "rdf:Description";
        case RDFSPackage.PLAIN_LITERAL:
            return "rdf:Literal";
        case RDFSPackage.TYPED_LITERAL:
            return "rdf:Literal";
        default:
            return "";
        }
    }

    /**
     * Catch the first encountered I/O exception during the type traversing
     * process.
     * 
     * @param e
     *            the current exception
     */
    protected void catchException(IOException e) {
        if (exceptionCaught == null)
            exceptionCaught = e;
    }

    /**
     * Write a number of tabs according to the current indentation level
     *  
     */
    protected void writeIndents() {
        int len = indentLevel;
        while (len-- > 0)
            try {
                writer.write("\t");
            } catch (IOException e) {
                catchException(e);
                return;
            }
    }

    /**
     * Write the start part of an XML element for a RDF resource. The RDF
     * resource along its properties will be outputed.
     * 
     * @param subject
     *            the RDF resource whose properties will be outputed
     */
    protected void writeElementStart(RDFSResource subject) {
        try {
            writeIndents();
            writer.write("<"
                         + elementType + " " + getIDAttrString(subject, false)
                         + ">\n");
            indentLevel++;
        } catch (IOException e) {
            catchException(e);
        }
    }

    /**
     * Write the closing tag of the current XML element.
     *  
     */
    protected void writeElementEnding() {
        indentLevel--;
        try {
            writeIndents();
            writer.write("</" + elementType + ">\n");
        } catch (IOException e) {
            catchException(e);
        }
    }

    /**
     * Write the start part of a XML element if the RDF resource is at the start
     * point of the type traversing.
     * 
     * @param object
     *            the RDF resource to be outputed
     */
    protected void possiblelyStartElement(RDFSResource object) {
        if (isElementStarting) {
            isElementStarting = false;
            elementType = getElementType(object);
            writeElementStart(object);
        }
    }

    /**
     * Write a property of the current RDF resource in XML.
     * 
     * @param propertyName
     *            the name of the property that can be directly written as a XML
     *            tag name
     * @param propValue
     *            the value of the property
     */
    protected void writeProperty(String propertyName, RDFSResource propValue) {
        try {
            writeIndents();
            writer.write("<" + propertyName);

            int classifierID = propValue.eClass().getClassifierID();
            switch (classifierID) {
            case RDFSPackage.RDFXML_LITERAL:
                RDFXMLLiteral xmlLiteral = (RDFXMLLiteral) propValue;
                writer.write(" rdf:parseType=\"Literal\">");
                writer.write(replaceKeywords(xmlLiteral.getLexicalForm()));
                writer.write("\n");
                writeIndents();
                writer.write("</" + propertyName + ">\n");
                return;

            case RDFSPackage.TYPED_LITERAL:
                TypedLiteral typedLiteral = (TypedLiteral) propValue;
                String datatypeURI = typedLiteral.getDatatype().getURI();
                writer.write(" rdf:datatype=\"" + datatypeURI + "\">");
                writer.write(typedLiteral.getLexicalForm());
                writer.write("</" + propertyName + ">\n");
                return;

            case RDFSPackage.RDFS_LITERAL:
            case RDFSPackage.PLAIN_LITERAL:
                RDFSLiteral rdfsLiteral = (RDFSLiteral) propValue;
                if (rdfsLiteral instanceof PlainLiteral) {
                    PlainLiteral plainLiteral = (PlainLiteral) rdfsLiteral;
                    String lang = plainLiteral.getLanguage();
                    if (lang != null)
                        if (lang.length() > 0) {
                            writer.write(" xml:lang=\"" + lang + "\"");
                        }
                }
                writer.write(">"
                             + rdfsLiteral.getLexicalForm());
                writer.write("</" + propertyName + ">\n");
                return;

            default: // the value is not a literal
                writer.write(" " + getIDAttrString(propValue, true) + "/>\n");
            }
        } catch (IOException e) {
            catchException(e);
        }
    }

    /**
     * Write out all values of a given property of the current RDF resource in
     * XML
     * 
     * @param propertyName
     *            the name of the property that can be directly written as a XML
     *            tag name
     * @param propValues
     *            a list of values of the property
     */
    protected void writeProperties(String propertyName, EList propValues) {
        for (Iterator iterator = propValues.iterator(); iterator.hasNext();) {
            writeProperty(propertyName, (RDFSResource) iterator.next());
        }
    }

    /**
     * Get the XML QName representation of a RDF resource's URI. Call this only
     * for URI reference RDF resource. Do not call the method for blanks node
     * and literals.
     * 
     * @param res
     *            the RDF resource
     * @param isEntityRef
     *            whether the returned representation is in the form of XML
     *            entity reference
     * @return the XML QName representation of a RDF resource's URI
     */
    protected String getQName(RDFSResource res, boolean isEntityRef) {
        // assuming correct namespace and local name is set
        if (isEntityRef)
            return "&"
                   + res.getNamespace().getName() + ";" + res.getLocalName();
        else
            return res.getNamespace().getName() + ":" + res.getLocalName();
    }

    /**
     * Get the XML attribute string representation of a RDF resource's URI or
     * blank node ID. It is in the form of "rdf:nodeID=..." for blank node and
     * "rdf:resource=..." or "rdf:about=..." for URI resources.
     * 
     * @param res
     *            the RDF resource whose URI/ID is to be obtained
     * @param isPropAttr
     *            for a RDF URI resource, whether to obtain as a XML attribute
     *            on a property XML element ("rdf:resource=...") or as a normal
     *            "rdf:about=...".
     * @return the XML attribute string of the resource's URI or blank node ID.
     */
    protected String getIDAttrString(RDFSResource res, boolean isPropAttr) {
        if (RDFXMLSaverHelper.isBlankNode(res))
            return "rdf:nodeID=\"" + res.getLocalName() + "\"";

        String leftSide = isPropAttr ? "rdf:resource=" : "rdf:about=";

        return leftSide + "\"" + getQName(res, true) + "\"";
    }

    /**
     * Construct a RDFXMLSaverImpl object with a writer and a RDFS ontology
     * 
     * @param writer
     *            the writer to use
     * @param ontology
     *            the RDFS ontology to save
     */
    public RDFXMLSaverImpl(Writer writer, Ontology ontology) {
        this.writer = writer;
        isElementStarting = false;
        indentLevel = 0;
        elementType = "";

        // break down the ontology resources into two parts

        rdfResources = new Vector(ontology.getContains().size());
        instanceStatements = new Vector();

        for (Iterator iterator = ontology.getContains().iterator(); iterator
                .hasNext();) {
            RDFSResource res = (RDFSResource) iterator.next();
            if (res instanceof RDFStatement) {
                RDFStatement stmt = (RDFStatement) res;
                if (stmt.isReified())
                    rdfResources.add(stmt);
                else
                    instanceStatements.add(stmt);
            } else if (!(res instanceof RDFSLiteral))
                rdfResources.add(res);
        }
    }

    /**
     * Save the RDFS ontology that is used to construct the RDFXMLSaverImpl
     * object.
     * 
     * @throws IOException
     *             If I/O error occurs
     */
    public void save() throws IOException {
        // save non-statement resources
        for (Iterator iterator = rdfResources.iterator(); iterator.hasNext();) {
            saveResource((RDFSResource) iterator.next());
        }

        // collect the subjects left in the instance statements and save them
        HashSet subjects = new HashSet();
        for (Iterator iterator = instanceStatements.iterator(); iterator
                .hasNext();) {
            subjects.add(iterator.next());
        }
        for (Iterator iterator = subjects.iterator(); iterator.hasNext();) {
            saveResource((RDFSResource) iterator.next());
        }
    }

    /**
     * Start saving a RDF resource and its properties.
     * 
     * @param subject
     *            the RDF resource to save
     * @throws IOException
     *             If I/O exception occurs
     */
    protected void saveResource(RDFSResource subject) throws IOException {
        indentLevel = 0;
        isElementStarting = true;
        exceptionCaught = null;

        // Re-using RDFSSwitch's type traversing code
        doSwitch(subject);

        if (exceptionCaught != null)
            throw exceptionCaught;
    }

    /**
     * This method is called at the end of the type traversing process by
     * {@link org.eclipse.eodm.rdfs.util.RDFSSwitch RDFSSwitch}. We need write out
     * instance statements associated with the resource and close the XML
     * element.
     * 
     * @param obj
     *            the RDF resource in the type traversing process
     * @see org.eclipse.eodm.rdfs.util.RDFSSwitch RDFSSwitch
     * @see org.eclipse.eodm.rdfs.util.RDFSSwitch#defaultCase(EObject)
     *      RDFSSwitch.defaultCase(EObject)
     */
    public Object defaultCase(EObject obj) {
        // write out instance statements that begins with the obj as subject
        for (Iterator iterator = instanceStatements.iterator(); iterator
                .hasNext();) {
            RDFStatement stmt = (RDFStatement) iterator.next();
            if (stmt.getRDFSubject() == obj) {
                RDFSResource property = stmt.getRDFPredicate();
                RDFSResource object = stmt.getRDFObject();
                if (property != null && object != null) {
                    writeProperty(getQName(property, false), object);
                    iterator.remove();
                }
            }
        }

        writeElementEnding();
        return exceptionCaught;
    }

    /**
     * For rdf:Alt resource, output all its properties.
     */
    public Object caseRDFAlt(RDFAlt object) {
        possiblelyStartElement(object);

        // no properties to write

        return exceptionCaught;
    }

    /**
     * For rdf:Bag resource, output all its properties.
     */
    public Object caseRDFBag(RDFBag object) {
        possiblelyStartElement(object);

        // no properties to write

        return exceptionCaught;
    }

    /**
     * For rdf:List resource, output all its properties.
     */
    public Object caseRDFList(RDFList object) {
        possiblelyStartElement(object);

        RDFSResource first = object.getRDFFirst();
        RDFList rest = object.getRDFRest();
        if (first != null)
            writeProperty("rdf:first", first);
        if (rest != null)
            writeProperty("rdf:rest", rest);

        return exceptionCaught;
    }

    /**
     * For rdf:Property resource, output all its properties.
     */
    public Object caseRDFProperty(RDFProperty object) {
        possiblelyStartElement(object);

        writeProperties("rdfs:subPropertyOf", object.getRDFSSubPropertyOf());
        writeProperties("rdfs:domain", object.getRDFSDomain());
        writeProperties("rdfs:range", object.getRDFSRange());

        return exceptionCaught;
    }

    /**
     * For rdfs:Class resource, output all its properties.
     */
    public Object caseRDFSClass(RDFSClass object) {
        possiblelyStartElement(object);

        writeProperties("rdfs:subClassOf", object.getRDFSSubClassOf());

        return exceptionCaught;
    }

    /**
     * For for rdfs:Container resource, output all its properties.
     */
    public Object caseRDFSContainer(RDFSContainer object) {
        possiblelyStartElement(object);

        // no properties to write

        return exceptionCaught;
    }

    /**
     * For rdfs:ContainerMembershipProperty resource, output all its properties.
     */
    public Object caseRDFSContainerMembershipProperty(
            RDFSContainerMembershipProperty object) {
        possiblelyStartElement(object);

        // no properties to write

        return exceptionCaught;
    }

    /**
     * For rdfs:Datatype resource, output all its properties.
     */
    public Object caseRDFSDatatype(RDFSDatatype object) {
        possiblelyStartElement(object);

        // no propertie to write

        return exceptionCaught;
    }

    /**
     * For rdf:Seq resource, output all its properties.
     */
    public Object caseRDFSeq(RDFSeq object) {
        possiblelyStartElement(object);

        // no properties to write

        return exceptionCaught;
    }

    /**
     * For rdfs:Resource resource, output all its properties.
     */
    public Object caseRDFSResource(RDFSResource object) {
        possiblelyStartElement(object);

        writeProperties("rdfs:label", object.getRDFSLabel());
        writeProperties("rdfs:comment", object.getRDFSComment());
        writeProperties("rdfs:seeAlso", object.getRDFSSeeAlso());
        writeProperties("rdfs:isDefinedBy", object.getRDFSIsDefinedBy());
        writeProperties("rdf:value", object.getRDFValue());
        writeProperties("rdfs:member", object.getRDFSMember());
        writeProperties("rdf:type", object.getRDFType());

        return exceptionCaught;
    }

    /**
     * For rdf:Statement resource, if it is a reified statment, output it.
     */
    public Object caseRDFStatement(RDFStatement stmt) {
        RDFSResource subject = stmt.getRDFSubject();
        RDFSResource predicate = stmt.getRDFPredicate();
        RDFSResource object = stmt.getRDFObject();

        if (isElementStarting) {
            isElementStarting = false;

            if (stmt.isReified()) {
                elementType = getElementType(stmt);
                writeElementStart(stmt);
            }
        }

        // writing properties
        if (stmt.isReified()) {
            if (subject != null)
                writeProperty("rdf:subject", subject);
            if (predicate != null)
                writeProperty("rdf:predicate", predicate);
            if (object != null)
                writeProperty("rdf:object", object);
        }

        return exceptionCaught;
    }

    private String replaceKeywords(String str) {
        str = str.replaceAll("&", "&amp;");
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("'", "&apos;");
        str = str.replaceAll("\"", "&quot;");
        return str;
    }
}
